#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections;
using System.Diagnostics;
using System.Threading;
using SerialPort;

namespace DnEcuDiag.Devices
{
    public abstract class BaseSerialDevice : IDisposable
    {
        protected const int InitreadbackTimeout = 5000;

        #region Fields

        private readonly Boolean _disposePort;
        private readonly object _lockObject = new object();
        private ManualResetEvent _comPortRxDone;
        private ManualResetEvent _comPortTxDone;
        private Timer _highResolutionTimer;
        private ManualResetEvent _timerSignal;
        private Int32 _waiting;
        protected IPort Port;

        #endregion

        #region Constructors

        protected BaseSerialDevice(IPort port, Boolean disposePort)
        {
            Port = port;
            _disposePort = disposePort;
            _comPortTxDone = new ManualResetEvent(false);
            _comPortRxDone = new ManualResetEvent(false);
            _timerSignal = new ManualResetEvent(false);

            port.DataReceived += ComPortDataReceived;

            _highResolutionTimer = new Timer(OnTimer, null, Timer.Capabilities.PeriodMin, 0);
            _highResolutionTimer.Start();
        }

        #endregion

        #region Protected Methods

        /// <summary>
        /// Wait for a specified amount of time in milliseconds.
        /// </summary>
        /// <param name="ms">The number of milliseconds to wait.</param>
        protected internal void Wait(int ms)
        {
            for (int i = 0; i < ms; i++)
            {
                _timerSignal.Reset();
                Interlocked.Increment(ref _waiting);
                _timerSignal.WaitOne();
            }
        }

        /// <summary>
        /// Handles the tick of the timer.
        /// </summary>
        /// <param name="sender"></param>
        protected internal void OnTimer(object sender)
        {
            if (_waiting == 1)
            {
                Interlocked.Decrement(ref _waiting);
                _timerSignal.Set();
            }
        }

        /// <summary>
        /// Read bytes from the serial port.
        /// </summary>
        /// <param name="numberBytesToRead">The number of bytes to read.</param>
        /// <param name="waitTimeout">The timeout in milliseconds to wait for the byte to be recieved.</param>
        /// <returns>Bytes received from serial port or null if timeout occurs.</returns>
        protected Byte[] WaitForBytes(int numberBytesToRead, int waitTimeout)
        {
            byte[] buffer = null;
            Double totalWaitTime = 0;
            Double timeToWait = waitTimeout;
            DateTime startTime = DateTime.Now;

            while (timeToWait > 0)
            {
                lock (_lockObject)
                {
                    if (Port.InBufferCount < numberBytesToRead)
                    {
                        _comPortRxDone.Reset();
                    }
                    else
                    {
                        _comPortRxDone.Set();
                    }
                }

                _comPortRxDone.WaitOne((Int32) timeToWait, false);

                if (Port.InBufferCount >= numberBytesToRead)
                {
                    Port.InputLen = numberBytesToRead;
                    buffer = Port.Input;
                    Port.InputLen = 0;
                    timeToWait = 0;
                }
                else
                {
                    DateTime endTime = DateTime.Now;

                    double timeElapsed = endTime.Subtract(startTime).TotalMilliseconds;
                    totalWaitTime += timeElapsed;

                    timeToWait = timeToWait - timeElapsed;
                }
            }

            Int32 bytesReceived = 0;

            if (buffer != null)
            {
                bytesReceived = buffer.Length;
            }

            if (EnableTrace)
            {
                Trace.WriteLine(String.Format("Expected Bytes: {0} Received Bytes: {1} Total wait time: {2} ms", numberBytesToRead, bytesReceived, totalWaitTime));
            }

            return buffer;
        }

        protected void InitaliseEcu(byte ecuId)
        {
            Port.Purge();
            byte[] buffer = Port.Input;

            var bitArray = new bool[10];
            int mask = 1;
            bitArray[0] = false; //start

            for (int i = 1; i <= 8; i++)
            {
                bitArray[i] = ((ecuId & mask) == mask); //Address byte
                mask = mask << 1;
            }

            bitArray[9] = true; //stop

            foreach (bool bit in bitArray)
            {
                Port.Break = !bit;

                Wait(200);
            }

            Port.Purge();
        }

        #endregion

        #region Protected Properties

        public bool EnableTrace { get; set; }

        #endregion

        #region Private Methods

        private void ComPortDataReceived(object sender, EventArgs e)
        {
            lock (_lockObject)
            {
                _comPortRxDone.Set();
            }
        }

        #endregion

        /// <summary>
        /// Collates the responses.
        /// </summary>
        /// <param name="responses">The responses.</param>
        /// <returns></returns>
        protected internal static byte[] CollateResponses(ArrayList responses)
        {
            if (responses == null)
            {
                throw new ArgumentNullException("responses");
            }

            int lenght = 0;
            int index = 0;

            foreach (byte[] arr in responses)
            {
                lenght += arr.Length;
            }

            var buffer = new byte[lenght];

            foreach (byte[] arr in responses)
            {
                Buffer.BlockCopy(arr, 0, buffer, index, arr.Length);
                index += arr.Length;
            }

            return buffer;
        }

        #region Dispose Pattern

        private bool _disposed;

        /// <summary>
        /// Dispose the underlying resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Finalizer.
        /// </summary>
        ~BaseSerialDevice()
        {
            Dispose(false);
        }

        /// <summary>
        /// Close the device.
        /// </summary>
        public void Close()
        {
            Dispose(true);
        }

        /// <summary>
        /// Dispose of the objects resources.
        /// </summary>
        /// <param name="disposing">True if explicity disposing object else GC finalizing object.</param>
        private void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    if (Port != null)
                    {
                        Port.Close();
                        if (_disposePort)
                        {
                            var disposable = Port as IDisposable;
                            if (disposable != null)
                            {
                                disposable.Dispose();
                            }
                        }
                        Port = null;
                    }

                    // safe to access other objects.
                    if (_highResolutionTimer != null)
                    {
                        _highResolutionTimer.Stop();
                        _highResolutionTimer.Dispose();
                        _highResolutionTimer = null;
                    }

                    if (_timerSignal != null)
                    {
                        _timerSignal.Close();
                        _timerSignal = null;
                    }

                    if (_comPortTxDone != null)
                    {
                        _comPortTxDone.Close();
                        _comPortTxDone = null;
                    }

                    if (_comPortRxDone != null)
                    {
                        _comPortRxDone.Close();
                        _comPortRxDone = null;
                    }
                }
            }
            _disposed = true;
        }

        #endregion
    }
}