#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using DnEcuDiag.Devices.Enums;
using DnEcuDiag.Devices.Exceptions;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Devices.Responses;
using SerialPort;

namespace DnEcuDiag.Devices
{
    /// <summary>
    /// Serial Device. This Class
    /// </summary>
    public sealed class Kwp2000SerialDevice : BaseSerialDevice, IDevice
    {
        #region Constants

        public const Int32 DefaultEcuResponseTimeout = 1500;
        public const Int32 DefaultRequestInterbyteTimeout = 5;
        public const Int32 DefaultEcuResponseInterbyteTimeout = 100;

        private const int ReadbackTimeout = 1000;
        private const int Isobaudrate = 10400;
        private const int Headerlenght = 5;
        private const byte Header = 0x80;

        #endregion

        #region Fields

        private int _requestInterbyteTime = DefaultRequestInterbyteTimeout;
        private int _ecuResponseTimeout = DefaultEcuResponseTimeout;
        private int _ecuResponseInterByteTimeout = DefaultEcuResponseInterbyteTimeout;

        #endregion

        #region Constructors

        /// <summary>
        /// The serial device constructor.
        /// </summary>
        /// <param name="port">The com port to open.</param>
        /// <param name="disposePort"></param>
        public Kwp2000SerialDevice(IPort port, Boolean disposePort)
            : base(port, disposePort)
        {
            port.DtrEnable = true;
            port.RtsEnable = false;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Sends a command to an ecu with initialisation before send.
        /// </summary>
        /// <param name="ecuId">The ECU ID.</param>
        /// <param name="testerId">The tester ID.</param>
        /// <param name="commandToSend">The command to send.</param>
        /// <param name="type">The type of initialisation to perform.</param>
        /// <returns>The response from the ecu.</returns>
        public Response SendCommandWithInit(byte ecuId, byte testerId, byte[] commandToSend, DeviceInitType type)
        {
            if (type == DeviceInitType.Extended)
            {
                return SendCommandWithInitFast(ecuId, testerId, commandToSend);
            }
            
            return SendCommandWithInitSlow(ecuId, testerId, commandToSend);
        }

        /// <summary>
        /// Send command in hardware device.
        /// </summary>
        /// <param name="ecuId">The ECU to send the message too.</param>
        /// <param name="testerId">The ID of the tester.</param>
        /// <param name="commandToSend">Command bytes to send to ECU.</param>
        /// <returns>The response to the command.</returns>
        public Response SendCommand(byte ecuId, byte testerId, byte[] commandToSend)
        {
            SendRequest(PrepareRequestMessage(ecuId, testerId, commandToSend));

            return GetResponse();
        }

        #endregion

        #region Private Methods

        private Response GetResponse()
        {
            // read header
            Byte[] response = WaitForBytes(1, _ecuResponseTimeout);
            if (response == null)
            {
                throw new DeviceTimeoutException("Failed to read response HEADER byte.", _ecuResponseTimeout);
            }

            Int32 lenght = (response[0] & 0x3F);

            // Read Target ID
            if (WaitForBytes(1, _ecuResponseInterByteTimeout) == null)
            {
                throw new DeviceTimeoutException("Failed to read response SOURCE byte.", _ecuResponseInterByteTimeout);
            }

            if (WaitForBytes(1, _ecuResponseInterByteTimeout) == null)
            {
                throw new DeviceTimeoutException("Failed to read response TARGET byte.", _ecuResponseInterByteTimeout);
            }

            // read lenght
            if (lenght == 0)
            {
                response = WaitForBytes(1, _ecuResponseInterByteTimeout);

                if (response == null)
                {
                    throw new DeviceTimeoutException("Failed to read response DATA LENGHT byte.",
                                                     _ecuResponseInterByteTimeout);
                }

                lenght = response[0];
            }

            response = new byte[lenght];

            for (int i = 0; i < lenght; i++)
            {
                Byte[] buffer = WaitForBytes(1, _ecuResponseInterByteTimeout);

                if (buffer == null)
                {
                    throw new DeviceTimeoutException("Failed to read response DATA.", _ecuResponseInterByteTimeout);
                }

                response[i] = buffer[0];
            }

            if (WaitForBytes(1, _ecuResponseInterByteTimeout) == null)
            {
                throw new DeviceTimeoutException("Failed to read response CRC byte.", _ecuResponseInterByteTimeout);
            }
            return new Kwp2000Response(response);
        }

        private void SendRequest(ICollection<byte> request)
        {
            Wait(60);

            foreach (byte b in request)
            {
                Port.SendImmediate(b);
                Wait(5);
            }

            if (WaitForBytes(request.Count, ReadbackTimeout) == null)
            {
                throw new DeviceTimeoutException("Failed to readback REQUEST message.", _ecuResponseTimeout);
            }
        }

        /// <summary>
        /// Send command on hardware device. With fast init before transmitting data.
        /// </summary>
        /// <param name="ecuId">The ECU to send the message too.</param>
        /// <param name="testerId">The ID of the tester.</param>
        /// <param name="commandToSend">Command bytes to send to ECU.</param>
        /// <returns>The response to the command.</returns>
        private Response SendCommandWithInitFast(byte ecuId, byte testerId, byte[] commandToSend)
        {
            var request = new byte[Headerlenght + commandToSend.Length];

            Byte crc = 0;
            crc += Header;
            crc += ecuId;
            crc += testerId;
            crc += (Byte) commandToSend.Length;

            for (var index = 0; index < commandToSend.Length; index++)
            {
                crc += commandToSend[index];
            }


            // Construct the request.
            request[0] = Header;
            request[1] = ecuId;
            request[2] = testerId;
            request[3] = (byte) commandToSend.Length;

            for (int index = 0; index < commandToSend.Length; index++)
            {
                request[4 + index] = commandToSend[index];
            }

            request[Headerlenght + commandToSend.Length - 1] = crc;

            // Wait 300ms.
            Wait(300);
            //Flush the buffer. 1.0.0.6
            Port.Purge();
            Byte[] buffer = Port.Input;
            // Wait for send of Init.
            Port.SendImmediate(0xF0);

            if (WaitForBytes(1, ReadbackTimeout) == null)
            {
                throw new DeviceTimeoutException("Timeout reading INIT byte.", _ecuResponseTimeout);
            }

            // reset the baudrate to 10400.
            Port.SetBaudRate(Isobaudrate);

            Port.DtrEnable = true;
            Port.RtsEnable = false;

            // DG 1.0.0.7
            Port.Purge();

            for (int index = 0; index < request.Length; index++)
            {
                Port.SendImmediate(request[index]);
                Wait(_requestInterbyteTime);
            }

            Byte[] response = WaitForBytes(request.Length, ReadbackTimeout);
            if (response == null)
            {
                throw new DeviceTimeoutException("Failed to read back request message.", _ecuResponseTimeout);
            }

            int syncIndexA, syncIndexB;
            if (!SyncArrays(request, response, out syncIndexA, out syncIndexB))
            {
                throw new SerialDeviceException("Failed to read back request message correctly.");
            }

            if (Math.Abs(syncIndexA - syncIndexB) != 0)
            {
                response = WaitForBytes(Math.Abs(syncIndexA - syncIndexB), ReadbackTimeout);
                if (response == null)
                {
                    throw new DeviceTimeoutException("Failed to clear extra bytes.", _ecuResponseTimeout);
                }
            }
            // read header
            response = WaitForBytes(1, _ecuResponseTimeout);
            if (response == null)
            {
                throw new DeviceTimeoutException("Failed to read response HEADER byte.", _ecuResponseTimeout);
            }

            Debug.WriteLine("RESPONSE HEADER: " + response[0].ToString("x"));

            Int32 lenght = (response[0] & 0x3F);

            // Read Target ID
            if (WaitForBytes(1, _ecuResponseInterByteTimeout) == null)
            {
                throw new DeviceTimeoutException("Failed to read response TARGET byte.", _ecuResponseInterByteTimeout);
            }

            // Read Source ID
            if (WaitForBytes(1, _ecuResponseInterByteTimeout) == null)
            {
                throw new DeviceTimeoutException("Failed to read response SOURCE byte.", _ecuResponseInterByteTimeout);
            }

            // read lenght
            if (lenght == 0)
            {
                response = WaitForBytes(1, _ecuResponseInterByteTimeout);
                if (response == null)
                {
                    throw new DeviceTimeoutException("Failed to read response DATA LENGHT byte.",
                                                     _ecuResponseInterByteTimeout);
                }

                lenght = response[0];
            }

            buffer = new byte[lenght];

            for (int i = 0; i < lenght; i++)
            {
                response = WaitForBytes(1, _ecuResponseInterByteTimeout);
                if (response == null)
                {
                    throw new DeviceTimeoutException("Failed to read response DATA.", _ecuResponseInterByteTimeout);
                }
                buffer[i] = response[0];
            }

            // Read CRC
            if (WaitForBytes(1, _ecuResponseInterByteTimeout) == null)
            {
                throw new DeviceTimeoutException("Failed to read response CRC byte.", _ecuResponseInterByteTimeout);
            }

            // get the response.
            return new Kwp2000Response(buffer);
        }

        /// <summary>
        /// Send command on hardware device. With slow init before transmitting data.
        /// </summary>
        /// <param name="ecuId">The ECU to send the message too.</param>
        /// <param name="testerId">The ID of the tester.</param>
        /// <param name="commandToSend">Command bytes to send to ECU.</param>
        /// <returns>The response to the command.</returns>
        private Response SendCommandWithInitSlow(byte ecuId, byte testerId, byte[] commandToSend)
        {
            byte[] response;

            Port.SetBaudRate(Isobaudrate);

            Port.DtrEnable = true;
            Port.RtsEnable = false;

            Wait(300);

            InitaliseEcu(ecuId);

            while (true)
            {
                response = WaitForBytes(1, InitreadbackTimeout);

                if (response == null)
                {
                    throw new DeviceTimeoutException("KWP01: Timeout reading INIT byte [1]", InitreadbackTimeout);
                }

                if (EnableTrace)
                {
                    Trace.WriteLine(Response.DumpPacket(response));
                }

                if (response[0] == 0x55)
                {
                    break;
                }
            }

            for (int i = 0; i < 2; i++)
            {
                response = WaitForBytes(1, InitreadbackTimeout);
                if (response == null)
                {
                    throw new DeviceTimeoutException("KWP03: Timeout reading KeyByes", InitreadbackTimeout);
                }
            }

            Wait(25);
            Port.SendImmediate((byte) ~response[0]);

            for (int i = 0; i < 2; i++)
            {
                response = WaitForBytes(1, InitreadbackTimeout);
                if (response == null)
                {
                    throw new DeviceTimeoutException("KWP04: Timeout reading Inverted Address byte",
                                                     InitreadbackTimeout);
                }
            }

            return SendCommand(ecuId, testerId, commandToSend);
        }

        private static byte[] PrepareRequestMessage(byte ecuId, byte testerId, byte[] requestMessage)
        {
            var request = new byte[Headerlenght + requestMessage.Length];
            Byte crc = 0;
            crc += Header;
            crc += ecuId;
            crc += testerId;
            crc += (byte) requestMessage.Length;

            for (int index = 0; index < requestMessage.Length; index++)
            {
                crc += requestMessage[index];
            }

            // Construct the request.
            request[0] = Header;
            request[1] = ecuId;
            request[2] = testerId;
            request[3] = (byte) requestMessage.Length;

            for (int index = 0; index < requestMessage.Length; index++)
            {
                request[4 + index] = requestMessage[index];
            }

            request[Headerlenght + requestMessage.Length - 1] = crc;

            return request;
        }

        private static bool SyncArrays(byte[] arrayA, byte[] arrayB, out int indexA, out int indexB)
        {
            bool syncronised = false;
            indexB = 0;

            for (indexA = 0; indexA < arrayA.Length; indexA++)
            {
                for (indexB = 0; indexB < arrayB.Length; indexB++)
                {
                    if (arrayA[indexA] == arrayB[indexB])
                    {
                        syncronised = true;
                        break;
                    }
                }

                if (syncronised)
                    break;
            }

            return syncronised;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Return the key bytes that where returned from the device.
        /// </summary>
        public byte[] KeyBytes
        {
            get { return new byte[2]; }
        }

        public ReadOnlyCollection<byte[]> IdentificationData
        {
            get { return new List<byte[]>().AsReadOnly(); }
        }

        /// <summary>
        /// The number of milliseconds between each consecutive byte of a command 
        /// to be sent.
        /// </summary>
        public int RequestInterByteTime
        {
            get { return _requestInterbyteTime; }
            set { _requestInterbyteTime = value; }
        }

        /// <summary>
        /// The number of milliseconds between each request.
        /// </summary>
        public int EcuResponseTimeout
        {
            get { return _ecuResponseTimeout; }
            set { _ecuResponseTimeout = value; }
        }

        /// <summary>
        /// The number of milliseconds between each request.
        /// </summary>
        public int EcuResponseInterbyteTimeout
        {
            get { return _ecuResponseInterByteTimeout; }
            set { _ecuResponseInterByteTimeout = value; }
        }

        #endregion
    }
}