#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using DnEcuDiag.Devices.Enums;
using DnEcuDiag.Devices.Exceptions;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Devices.Responses;
using SerialPort;

namespace DnEcuDiag.Devices
{
    /// <summary>
    /// ISO5Interface. This interface implements the iso 5 protolcol.
    /// It handles the initalisation and protocol timing.
    /// </summary>
    public class Kwp81SerialDevice : BaseSerialDevice, IDevice
    {
        #region Constants

        public const Int32 DefaultEcuResponseInterbyteTimeout = 150;
        public const Int32 DefaultEcuResponseTimeout = 2000;
        public const Int32 DefaultRequestInterbyteTimeout = 5;
        protected const int Headerlenght = 2;
        protected const int ReadbackTimeout = 5000;

        #endregion

        #region Fields

        protected Int32 FrameTitleIndex;
        protected readonly byte[] Keybytes = new byte[6];
        private int _requestInterbyteTime = DefaultRequestInterbyteTimeout;
        protected Byte[] LastCommand;
        protected List<Byte[]> identificationData;

        #endregion

        #region Constructors

        /// <summary>
        /// The serial device constructor.
        /// </summary>
        /// <param name="port"></param>
        /// <param name="disposePort"></param>
        public Kwp81SerialDevice(IPort port, Boolean disposePort) : base(port, disposePort)
        {
            EcuResponseInterbyteTimeout = DefaultEcuResponseInterbyteTimeout;
            EcuResponseTimeout = DefaultEcuResponseTimeout;
            FrameTitleIndex = 1;
            port.DtrEnable = true;
            port.RtsEnable = false;
            port.SetBaudRate(4800);
            identificationData = new List<Byte[]>();
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Send command on hardware device. With init before transmitting data.
        /// </summary>
        /// <param name="ecuId">The ECU to send the message too.</param>
        /// <param name="testerId">The ID of the tester.</param>
        /// <param name="commandToSend">Command bytes to send to ECU.</param>
        /// <param name="type">This parameter is ignored.</param>
        /// <returns>The response to the command.</returns>
        public virtual Response SendCommandWithInit(byte ecuId, byte testerId, byte[] commandToSend, DeviceInitType type)
        {
            InitaliseEcu(ecuId);

            var response = new byte[6];
            Byte[] buffer;

            while (true)
            {
                buffer = WaitForBytes(1, InitreadbackTimeout);

                if (buffer == null)
                {
                    throw new DeviceTimeoutException("ISO00: Timeout reading INIT byte [1]", InitreadbackTimeout);
                }

                if (EnableTrace)
                {
                    Trace.WriteLine(Response.DumpPacket(buffer));
                }

                if (buffer[0] == 0x55)
                {
                    response[0] = buffer[0];
                    break;
                }
            }

            for (int i = 1; i < 6; i++)
            {
                buffer = WaitForBytes(1, EcuResponseInterbyteTimeout);
                if (buffer == null)
                {
                    throw new DeviceTimeoutException("ISO1: Failed to read ACK bytes from ecu.", EcuResponseInterbyteTimeout);
                }

                if (EnableTrace)
                {
                    Trace.WriteLine(Response.DumpPacket(buffer));
                }

                response[i] = buffer[0];
            }

            if (EnableTrace)
            {
                Trace.WriteLine(Response.DumpPacket(response));
            }

            Buffer.BlockCopy(response, 0, Keybytes, 0, 6);

            Port.SendImmediate((byte) (~response[2]));
            response = WaitForBytes(1, EcuResponseTimeout);
            if (response == null)
            {
                throw new DeviceTimeoutException("ISO2: Failed to read invert key byte.", EcuResponseTimeout);
            }

            // Read back response
            response = ReceiveResponse();
            if (response == null)
            {
                throw new DeviceTimeoutException("ISO3: Failed to read Ecu response packet.", EcuResponseTimeout);
            }

            if (EnableTrace)
            {
                Trace.WriteLine(Response.DumpPacket(response));
            }

            ValidateResponse(response[FrameTitleIndex]);

            return SendCommand(ecuId, testerId, commandToSend);
        }

        /// <summary>
        /// Send command in hardware device.
        /// </summary>
        /// <param name="ecuId">The ECU to send the message too.</param>
        /// <param name="testerId">The ID of the tester.</param>
        /// <param name="commandToSend">Command bytes to send to ECU.</param>
        /// <returns>The response to the command.</returns>
        public Response SendCommand(byte ecuId, byte testerId, byte[] commandToSend)
        {
            LastCommand = commandToSend;

            var responses = new List<Byte[]>();
            SendCommand(commandToSend, responses);

            return new KwpResponse(responses);
        }

        #endregion

        #region Protected Methods

        private void SendCommand(byte[] commandToSend, IList responses)
        {
            TransmitCommand(commandToSend);

            byte[] buffer = ReceiveResponse();

            if (buffer[FrameTitleIndex] != 0x09)
            {
                responses.Add(buffer);
                SendCommand(new byte[] {0x09}, responses);
            }
            else
            {
                return;
            }
        }

        /// <summary>
        /// Sends the command.
        /// </summary>
        /// <param name="commandToSend">The command to send.</param>
        protected virtual void TransmitCommand(byte[] commandToSend)
        {
            Int32 readback = Headerlenght + commandToSend.Length;
            var request = new byte[commandToSend.Length + 2];
            request[0] = (byte) (commandToSend.Length + 1);
            
            Byte crc = 0;
            crc += request[0];

            for (int index = 0; index < commandToSend.Length; index++)
            {
                crc += commandToSend[index];
            }

            Buffer.BlockCopy(commandToSend, 0, request, 1, commandToSend.Length);
            request[request.Length - 1] = crc;

            Port.Output = request;

            byte[] response = WaitForBytes(readback, ReadbackTimeout);
            if (response == null)
            {
                throw new DeviceTimeoutException("ISO5: Failed to readback REQUEST message.", EcuResponseTimeout);
            }
        }

        /// <summary>
        /// Receives the response.
        /// </summary>
        /// <returns></returns>
        protected virtual byte[] ReceiveResponse()
        {
            byte[] response = WaitForBytes(1, EcuResponseTimeout);
            if (response == null)
            {
                throw new DeviceTimeoutException("ISO6: Failed to read response byte.", EcuResponseTimeout);
            }

            Byte len = response[0];
            var buffer = new byte[len + 1];
            buffer[0] = len;

            for (int i = 0; i < len; i++)
            {
                response = WaitForBytes(1, EcuResponseInterbyteTimeout);
                if (response == null)
                {
                    throw new DeviceTimeoutException("ISO8: Failed to read response data bytes.",
                                                     EcuResponseInterbyteTimeout);
                }

                buffer[i + 1] = response[0];
            }

            return buffer;
        }

        protected void ValidateResponse(byte response)
        {
            switch (response)
            {
                case 0x0A:
                    throw new DeviceTimeoutException("ISO4: NACK Ecu response packet.", EcuResponseTimeout);
                case 0x0B:
                    throw new DeviceTimeoutException("ISO5: Unhandled Ecu response packet.", EcuResponseTimeout);
                case 0x09:
                    Trace.WriteLine("Received ACK.");
                    break;
                case 0xF6:
                    Trace.WriteLine("Received identification data.");
                    break;
                default:
                    throw new DeviceTimeoutException("ISO6: Unhandled Ecu response packet.", EcuResponseTimeout);
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Return the key bytes that where returned from the device.
        /// </summary>
        public byte[] KeyBytes
        {
            get { return Keybytes; }
        }

        /// <summary>
        /// The number of milliseconds between each consecutive byte of a command 
        /// to be sent.
        /// </summary>
        public int RequestInterByteTime
        {
            get { return _requestInterbyteTime; }
            set { _requestInterbyteTime = value; }
        }

        /// <summary>
        /// The number of milliseconds between end of request and response from ecu.
        /// </summary>
        public int EcuResponseTimeout { get; set; }

        /// <summary>
        /// The number of milliseconds between each request.
        /// </summary>
        public int EcuResponseInterbyteTimeout { get; set; }

        public ReadOnlyCollection<byte[]> IdentificationData
        {
            get { return identificationData.AsReadOnly(); }
        }

        #endregion
    }
}