#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Runtime.InteropServices;
using DnEcuDiag.Devices.Exceptions;

namespace DnEcuDiag.Devices
{
    /// <summary>
    /// Represents the method that handles calls from a multimedia timer.
    /// </summary>
    /// <param name="state">
    /// An object containing application-specific information relevant to the 
    /// method invoked by this delegate, or a null reference.
    /// </param>
    /// <remarks>
    /// Use a TimerCallback delegate to specify the method that is called by a 
    /// multimedia timer.
    /// </remarks>
    public delegate void TimerCallback(object state);

    /// <summary>
    /// Represents a multimedia timer.
    /// </summary>
    /// <remarks>
    /// <para>The multimedia timer allows applications to schedule timer events 
    /// with the greatest resolution (or accuracy) possible for the hardware 
    /// platform. The multimedia timer allows you to schedule timer events at a 
    /// higher resolution than other timer services.</para>
    /// 
    /// <para>The multimedia timer is useful for applications that demand 
    /// high-resolution timing. For example, a MIDI sequencer requires a 
    /// high-resolution timer because it must maintain the pace of MIDI events 
    /// within a resolution of 1 millisecond.</para>
    /// </remarks>
    public sealed class Timer : IDisposable
    {
        // Timer identifier.
        private int _timerId;

        // Timer mode.
        private TimerMode _mode;

        // Period between timer events in milliseconds.
        private int _period;

        // Timer resolution in milliseconds.
        private int _resolution;

        // Indicates whether or not the timer is running.
        private bool _running;

        // Called by Windows when a timer periodic event occurs.
        private readonly NativeMethods.TimeProc _timeProcPeriodic;

        // Called by Windows when a timer one shot event occurs.
        private readonly NativeMethods.TimeProc _timeProcOneShot;

        // User supplied callback that is called when a timer event occurs.
        private readonly TimerCallback _callback;

        // User supplied state information.
        private readonly object _state;

        // Multimedia timer capabilities.
        private static TimerCaps _caps;

        /// <summary>
        /// Initialize class.
        /// </summary>
        static Timer()
        {
            // Get multimedia timer capabilities.
            NativeMethods.TimeGetDevCaps(ref _caps, Marshal.SizeOf(_caps));
        }

        /// <summary>
        /// Initializes a new instance of the Timer class with the user 
        /// supplied callback and state information, timer period, and 
        /// resolution.
        /// </summary>
        /// <param name="callback">
        /// A TimerCallback delegate representing a method to be executed when 
        /// a timer event occurs. 
        /// </param>
        /// <param name="state">
        /// An object containing information to be used by the callback method, 
        /// or a null reference.
        /// </param>
        /// <param name="period">
        /// The time between timer events in milliseconds.
        /// </param>
        /// <param name="resolution">
        /// The timer resolution in milliseconds.
        /// </param>
        public Timer(TimerCallback callback, object state, int period, int resolution)
        {
            _callback = callback;
            _state = state;
            _timeProcPeriodic = new NativeMethods.TimeProc(OnTimerPeriodicEvent);
            _timeProcOneShot = new NativeMethods.TimeProc(OnTimerOneShotEvent);
            Mode = TimerMode.Periodic;
            Period = period;
            Resolution = resolution;
        }

        /// <summary>
        /// Initializes a new instance of the Timer class with the user 
        /// supplied callback and state information, timer period, resolution,
        /// and timer mode.
        /// </summary>
        /// <param name="callback">
        /// A TimerCallback delegate representing a method to be executed when 
        /// a timer event occurs. 
        /// </param>
        /// <param name="state">
        /// An object containing information to be used by the callback method, 
        /// or a null reference.
        /// </param>
        /// <param name="period">
        /// The time between timer events in milliseconds.
        /// </param>
        /// <param name="resolution">
        /// The timer resolution in milliseconds.
        /// </param>
        /// <param name="mode">
        /// The timer mode.
        /// </param>
        public Timer(TimerCallback callback, object state, int period, int resolution, TimerMode mode)
        {
            _callback = callback;
            _state = state;
            _timeProcPeriodic = new NativeMethods.TimeProc(OnTimerPeriodicEvent);
            _timeProcOneShot = new NativeMethods.TimeProc(OnTimerOneShotEvent);
            Mode = mode;
            Period = period;
            Resolution = resolution;
        }

        /// <summary>
        /// Destructor.
        /// </summary>
        ~Timer()
        {
            if (_running)
            {
                Stop();
            }
        }

        /// <summary>
        /// Starts the timer.
        /// </summary>
        /// <exception cref="TimerStartException">
        /// Thrown if unable to start timer.
        /// </exception>
        public void Start()
        {
            // If the timer is already running.
            if (_running)
            {
                // Stop timer.
                Stop();
            }

            // If the periodic event callback should be used.
            _timerId = _mode == TimerMode.Periodic ? NativeMethods.TimeSetEvent(Period, Resolution, _timeProcPeriodic, 0, (int)Mode) : NativeMethods.TimeSetEvent(Period, Resolution, _timeProcOneShot, 0, (int)Mode);

            // If the timer was created successfully.
            if (_timerId != 0)
            {
                _running = true;
            }
            else
            {
                throw new TimerStartException("TimerStartFailed");
            }
        }

        /// <summary>
        /// Stops timer.
        /// </summary>
        public void Stop()
        {
            // If the timer is running.
            if (_running)
            {
                // Stop and destroy timer.
                NativeMethods.TimeKillEvent(_timerId);

                // Indicate that the timer is not running.
                _running = false;
            }
        }

        /// <summary>
        /// Indicates whether or not the timer is running.
        /// </summary>
        /// <returns>
        /// Returns true if the timer is running; otherwise, false.
        /// </returns>
        public bool IsRunning()
        {
            return _running;
        }

        /// <summary>
        /// Callback method called by the Win32 multimedia timer when a timer
        /// periodic event occurs.
        /// </summary>
        private void OnTimerPeriodicEvent(int id, int msg, int user, int param1, int param2)
        {
            // Call user supplied callback.
            _callback(_state);
        }

        /// <summary>
        /// Callback method called by the Win32 multimedia timer when a timer
        /// one shot event occurs.
        /// </summary>
        private void OnTimerOneShotEvent(int id, int msg, int user, int param1, int param2)
        {
            // Call user supplied callback.
            _callback(_state);

            // Stop timer.
            Stop();
        }

        /// <summary>
        /// Gets or sets the timer mode.
        /// </summary>
        /// <remarks>
        /// If the timer is running when the mode is changed, the timer is 
        /// stopped and restarted with the new mode value.
        /// </remarks>
        public TimerMode Mode
        {
            get
            {
                return _mode;
            }
            set
            {
                _mode = value;

                // If the timer is running.
                if (IsRunning())
                {
                    // Stop and restart timer.
                    Stop();
                    Start();
                }
            }
        }

        /// <summary>
        /// Gets or sets the timer period.
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException">
        /// Thrown if the period is set to a value out of range for the 
        /// multimedia timer.
        /// </exception>
        /// <remarks>
        /// <para>The multimedia timer period is the time between timer events 
        /// in milliseconds.</para>
        /// 
        /// <para>If the timer is running when the period is changed, the timer 
        /// is stopped and restarted with the new period value.</para>   
        /// </remarks>     
        public int Period
        {
            get
            {
                return _period;
            }
            set
            {
                // If period is in range.
                if (value >= _caps.PeriodMin && value <= _caps.PeriodMax)
                {
                    // Assign period.
                    _period = value;

                    // If the timer is running.
                    if (IsRunning())
                    {
                        // Stop and restart timer.
                        Stop();
                        Start();
                    }
                }
                // Else period is out of range.
                else
                {
                    // Stop timer before throwing exception (has no effect if
                    // the timer is not running).
                    Stop();

                    // Get error message and throw exception.
                    throw new ArgumentOutOfRangeException("value", value, "TimerPeriodOutOfRange");
                }
            }
        }

        /// <summary>
        /// Gets or sets the timer resolution.
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException">
        /// Thrown if the resolution is set to a value less than zero.
        /// </exception>
        /// <remarks>
        /// <para>The resolution is in milliseconds. The resolution increases 
        /// with smaller values; a resolution of 0 indicates periodic events 
        /// should occur with the greatest possible accuracy. To reduce system 
        /// overhead, however, you should use the maximum value appropriate 
        /// for your application.</para>
        /// 
        /// <para>If the timer is running when the resolution is changed, the 
        /// timer is stopped and restarted with the new resolution value.
        /// </para>        
        /// </remarks>
        public int Resolution
        {
            get
            {
                return _resolution;
            }
            set
            {
                // If resolution is in range.
                if (value >= 0)
                {
                    // Assign resolution.
                    _resolution = value;

                    // If the timer is running.
                    if (IsRunning())
                    {
                        // Stop and restart timer.
                        Stop();
                        Start();
                    }
                }
                // Else resolution is out of range.
                else
                {
                    // Stop timer before throwing exception (has no effect if the timer is not running).
                    Stop();
                    throw new ArgumentOutOfRangeException("value", value, "TimerResolutionOutOfRange");
                }
            }
        }

        /// <summary>
        /// Gets the timer capabilities.
        /// </summary>
        public static TimerCaps Capabilities
        {
            get
            {
                return _caps;
            }
        }

        #region IDisposable Members

        /// <summary>
        /// Frees timer resources.
        /// </summary>
        public void Dispose()
        {
            GC.SuppressFinalize(this);
            // If the timer is still running.
            if (_running)
            {
                // Stop timer.
                Stop();
            }
        }

        #endregion
    }
}
