#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using DnEcuDiag.Ecus.Kwp;
using NUnit.Framework;
using DnEcuDiag.Factories;
using DnEcuDiag.Ecus.Generic;
using DnEcuDiag.Ecus.Enums;

namespace DnEcuDiag.Tests
{
    [TestFixture]
    public class CFC208Tests
    {
        private KwpEcu ecu;
        private KwpTestDevice device;

        [TestFixtureSetUp]
        public void Init()
        {
            try
            {
                device = new KwpTestDevice();

                ecu = EcuFactory.CreateEcu(EcuFactory.EcuList.Find(obj => obj.Name.Contains("SeleSpeed CFC208F"))) as KwpEcu;
                ecu.DiagnosticInterface = device;
                ecu.Connect();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }

        [Test]
        public void TestConnect()
        {
            Assert.AreEqual("SeleSpeed CFC208F", ecu.Model);
            Assert.AreEqual("0123456789 ", ecu.DrawingNumber);
            Assert.AreEqual("0123456789 ", ecu.HardwareNumber);
            Assert.AreEqual("0x10", ecu.HardwareVersionNumber);
            Assert.AreEqual("0123456789 ", ecu.SoftwareNumber);
            Assert.AreEqual("0x85170", ecu.SoftwareVersionNumber);
            Assert.AreEqual("01234 ", ecu.OmologationNumber);
            Assert.AreEqual("55 AA 55 AA 55 ", ecu.IsoCode);
            Assert.AreEqual("19 97 03 31 ", ecu.ProductionDate);
            Assert.AreEqual("19 97 03 31 ", ecu.ProgrammingDate);
            Assert.IsNull(ecu.RpmSensor);
            Assert.IsNull(ecu.SpeedSensor);
        }

        [Test]
        public void TestSensors_SSCFC208F()
        {
            Assert.AreEqual(107, ecu.Sensors.Count);

            TestSensorValues_SSCFC208F(ecu);
        }

        [Test]
        public void TestFaults_SSCFC208F()
        {
            Assert.AreNotEqual(0, ecu.ErrorMemoryStructureDefinition.ErrorDefinitions.Count,
                               "m_ssCFC208F.ErrorMemoryStructureDefinition.Count == 0 for SSCFC208F");
            Assert.AreEqual(32, ecu.ErrorMemoryStructureDefinition.ErrorDefinitions.Count,
                            "m_ssCFC208F.ErrorMemoryStructureDefinition.Count != 32 for SSCFC208F");

            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Three P0715
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////			
            device.FaultcodeList = new byte[] {0x58, 01, 07, 0x15, 0xE1};
            device.Faultcodeinfo = new byte[] {0x57, 1, 0x07, 0x15, 0x08, 10, 10, 100, 0x33, 10, 10, 1};
            ecu.ReadFaultCodes();

            Assert.AreEqual(1, ecu.ActiveFaultCodes.Count, "m_ssCFC208F.ActiveFaultCodes.Count == 1 for SSCFC208F");
            Console.WriteLine(ecu.ActiveFaultCodes[0].FaultDescription);
            Assert.AreEqual("P0715", (ecu.ActiveFaultCodes[0]).FaultComponent);
            Assert.AreEqual(
                "Failed Component Code: Clutch Speed\r\n" +
                "Fault Symptom: Signal Not Plausible\r\n" +
                "Warning Lamp Status: Warning Lamp not Illuminated\r\n" +
                "Fault Test State: Test complete for this Error or not applicable.\r\n" +
                "Fault Code Storage State: No Error Detected at time of request\r\n" +
                "Event Count: 1 count\r\n" +
                "Environmental Condition One: Time Error Present: 10 minutes\r\n" +
                "Environmental Condition Two: Engine Speed : 320 rpm\r\n" +
                "Environmental Condition Three: Vehicle Speed: 141 km/h\r\n" +
                "Environmental Condition Four: Gearbox Oil Temperature: 32 C\r\n" +
                "Environmental Condition Five: Requested Gear: Third\r\n" +
                "Environmental Condition Five: Inserted Gear: Third\r\n" +
                "Environmental Condition Six: Battery Voltage: 8.99 volts\r\n",
                ecu.ActiveFaultCodes[0].FaultDescription);

            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // one P1772
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

            device.FaultcodeList = new byte[] {0x58, 0x1, 0x17, 0x72, 0xE1};
            device.Faultcodeinfo = new byte[] {0x57, 1, 0x17, 0x72, 0x08, 10, 10, 100, 0x33, 10, 10, 1};
            ecu.ReadFaultCodes();

            Assert.AreEqual(1, ecu.ActiveFaultCodes.Count, "m_ssCFC208F.ActiveFaultCodes.Count == 1 for SSCFC208F");
            Console.WriteLine(ecu.ActiveFaultCodes[0].FaultDescription);
            Assert.AreEqual("P1772", (ecu.ActiveFaultCodes[0]).FaultComponent);
            Assert.AreEqual(
                "Failed Component Code: Low Pressure Diagnosed\r\n" +
                "Fault Symptom: Signal Not Plausible\r\n" +
                "Warning Lamp Status: Warning Lamp not Illuminated\r\n" +
                "Fault Test State: Test complete for this Error or not applicable.\r\n" +
                "Fault Code Storage State: No Error Detected at time of request\r\n" +
                "Event Count: 1 count\r\n" +
                "Environmental Condition One: Time Error Present: 10 minutes\r\n" +
                "Environmental Condition Two: Engine Speed : 320 rpm\r\n" +
                "Environmental Condition Three: Vehicle Speed: 141 km/h\r\n" +
                "Environmental Condition Four: Gearbox Oil Temperature: 32 C\r\n" +
                "Environmental Condition Five: Requested Gear: Third\r\n" +
                "Environmental Condition Five: Inserted Gear: Third\r\n" +
                "Environmental Condition Six: Battery Voltage: 8.99 volts\r\n",
                ecu.ActiveFaultCodes[0].FaultDescription);

            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // one P0725
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

            device.FaultcodeList = new byte[] { 0x58, 0x1, 0x07, 0x25, 0xE1 };
            device.Faultcodeinfo = new byte[] { 0x57, 1, 0x07, 0x25, 0x08, 10, 10, 100, 0x33, 10, 10, 1 };
            ecu.ReadFaultCodes();

            Assert.AreEqual(1, ecu.ActiveFaultCodes.Count, "m_ssCFC208F.ActiveFaultCodes.Count == 1 for SSCFC208F");
            Console.WriteLine(ecu.ActiveFaultCodes[0].FaultDescription);
            Assert.AreEqual("P0725", (ecu.ActiveFaultCodes[0]).FaultComponent);
            Assert.AreEqual(
                "Failed Component Code: Engine Speed\r\n" +
                "Fault Symptom: Signal Not Plausible\r\n" +
                "Warning Lamp Status: Warning Lamp not Illuminated\r\n" +
                "Fault Test State: Test complete for this Error or not applicable.\r\n" +
                "Fault Code Storage State: No Error Detected at time of request\r\n" +
                "Event Count: 1 count\r\n" +
                "Environmental Condition One: Time Error Present: 10 minutes\r\n" +
                "Environmental Condition Two: Engine Speed : 320 rpm\r\n" +
                "Environmental Condition Three: Vehicle Speed: 141 km/h\r\n" +
                "Environmental Condition Four: Gearbox Oil Temperature: 32 C\r\n" +
                "Environmental Condition Five: Requested Gear: Third\r\n" +
                "Environmental Condition Five: Inserted Gear: Third\r\n" +
                "Environmental Condition Six: Battery Voltage: 8.99 volts\r\n",
                ecu.ActiveFaultCodes[0].FaultDescription);
        }

        [Test]
        public void TestPerformanceMeasurment_SSCFC208F()
        {
            Assert.AreEqual(null, ecu.SpeedSensor);
            Assert.AreEqual(null, ecu.RpmSensor);
        }

        [Test]
        public void TestRoutines_SSCFC208F()
        {
            Assert.AreEqual(7, ecu.Routines.Count);

            Assert.AreEqual("Bleed Clutch", ecu.Routines[0].Name);
            Assert.AreEqual(RoutineType.Calibration, ecu.Routines[0].RoutineType);
            Assert.AreEqual(0, ecu.Routines[0].Result);

            Assert.AreEqual("Depressurise Accumulator", ecu.Routines[1].Name);
            Assert.AreEqual(RoutineType.Calibration, ecu.Routines[1].RoutineType);
            Assert.AreEqual(0, ecu.Routines[1].Result);

            Assert.AreEqual("Enable AutoAdjustment of Clutch", ecu.Routines[2].Name);
            Assert.AreEqual(RoutineType.Activation, ecu.Routines[2].RoutineType);
            Assert.AreEqual(0, ecu.Routines[2].Result);
            Assert.AreEqual(0, ecu.Routines[2].Arguments.Count);

            Assert.AreEqual("AutoAdjustment EOL/Service", ecu.Routines[3].Name);
            Assert.AreEqual(RoutineType.Calibration, ecu.Routines[3].RoutineType);
            Assert.AreEqual(0, ecu.Routines[3].Result);

            Assert.AreEqual("New Actuator", ecu.Routines[4].Name);
            Assert.AreEqual(RoutineType.Activation, ecu.Routines[4].RoutineType);
            Assert.AreEqual(0, ecu.Routines[4].Result);

            Assert.AreEqual("Diagnosis of the loosening of the finger retaining screw", ecu.Routines[5].Name);
            Assert.AreEqual(RoutineType.Calibration, ecu.Routines[5].RoutineType);
            Assert.AreEqual(0, ecu.Routines[5].Result);

            Assert.AreEqual("Diagnosis of return to Neutral selection", ecu.Routines[6].Name);
            Assert.AreEqual(RoutineType.Calibration, ecu.Routines[6].RoutineType);
            Assert.AreEqual(0, ecu.Routines[6].Result);

            device.LocalidentifierData = new byte[] {0x73, 0x10, 0};

            Assert.AreEqual(0, ecu.ExecuteRoutine(ecu.Routines[0]));

            device.LocalidentifierData = new byte[] {0x73, 0x10, 0xFF};

            Assert.AreEqual(0xFF, ecu.ExecuteRoutine(ecu.Routines[0]));
        }

        private void TestSensorValues_SSCFC208F(Ecu ecu)
        {
            TimeStampedMeasurement tsm;
            LocalIdentifier li;
            int i = 0;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(0xff, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xff, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(2490.32983f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xff, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(2293.7251f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xff, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(2752.46997f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xff, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(2752.46997f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xff, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(26214, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xff, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(6291.82715f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xFF};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-0.100000001f, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xFB};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-0.5, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xFF};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(25.5, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xFF};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-1, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xF0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-16, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xFF};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(255, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(7.16550016f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(255, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(255, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(2.80500007f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(2550, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(2550, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(1989, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-1, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xF0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-16, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x0f};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(15, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-1, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0xF0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-16, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x0f};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(15, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xff};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual(true, tsm.IsStatus);
            Assert.AreEqual("No Request", tsm.Status);
            Assert.AreEqual(0xff, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x9F};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual(true, tsm.IsStatus);
            Assert.AreEqual("Request From Right Button", tsm.Status);
            Assert.AreEqual(0x9f, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xAF};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual(true, tsm.IsStatus);
            Assert.AreEqual("Request From Left Button", tsm.Status);
            Assert.AreEqual(0xAf, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xF9};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual(true, tsm.IsStatus);
            Assert.AreEqual("Lever Back", tsm.Status);
            Assert.AreEqual(0xF9, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xF3};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual(true, tsm.IsStatus);
            Assert.AreEqual("Lever Forward", tsm.Status);
            Assert.AreEqual(0xF3, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFA};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual(true, tsm.IsStatus);
            Assert.AreEqual("Request Neutral From Lever", tsm.Status);
            Assert.AreEqual(0xFA, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFC};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual(true, tsm.IsStatus);
            Assert.AreEqual("Request Reverse From Lever", tsm.Status);
            Assert.AreEqual(0xFC, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Neutral", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 6};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Reverse", tsm.Status);
            Assert.AreEqual(6, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("False", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 1};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("True", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("False", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 2};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("True", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("False", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 4};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("True", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;


            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("False", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 8};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("True", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;


            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("False", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x10};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("True", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;


            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Off", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 1};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("On", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Off", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 2};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("On", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Off", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 4};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("On", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Off", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 8};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("On", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Off", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x10};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("On", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;


            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Off", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x20};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("On", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Neutral", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 6};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Reverse", tsm.Status);
            Assert.AreEqual(6, tsm.Measurement);

            i++;

            for (int x = 0; x < 14; x++)
            {
                Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
                li = ecu.Sensors[i] as LocalIdentifier;
                device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
                tsm = ecu.ReadSensor(ecu.Sensors[i]);
                
                Assert.AreEqual(3.799999995f, tsm.Measurement);

                i++;
            }

            for (int x = 0; x < 12; x++)
            {
                Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
                li = ecu.Sensors[i] as LocalIdentifier;
                device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
                tsm = ecu.ReadSensor(ecu.Sensors[i]);
                
                Assert.AreEqual(3.5, tsm.Measurement);

                i++;
            }

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(4.19999981f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(4.2f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(780, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(10, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(10, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(20, tsm.Measurement);

            i++;

            for (int x = 0; x < 25; x++)
            {
                Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
                li = ecu.Sensors[i] as LocalIdentifier;
                device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
                tsm = ecu.ReadSensor(ecu.Sensors[i]);
                
                Assert.AreEqual(100, tsm.Measurement);

                i++;
            }

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("AutoCalibration \"OK\"", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x01};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Selecting 1a-2a", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x02};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Selecting 5a-Rm", tsm.Status);
            Assert.AreEqual(2, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x03};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Selecting 3a-4a", tsm.Status);
            Assert.AreEqual(3, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x04};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Inserting 1a", tsm.Status);
            Assert.AreEqual(4, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x05};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Inserting 2a", tsm.Status);
            Assert.AreEqual(5, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x06};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Inserting 3a", tsm.Status);
            Assert.AreEqual(6, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x07};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Inserting 4a", tsm.Status);
            Assert.AreEqual(7, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x08};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Inserting 5a", tsm.Status);
            Assert.AreEqual(8, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x09};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Inserting Rm", tsm.Status);
            Assert.AreEqual(9, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x10};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Positioning Gear Insertion Actuator", tsm.Status);
            Assert.AreEqual(16, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x11};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Interuption Due to Clutch Closing to Quickly", tsm.Status);
            Assert.AreEqual(17, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x12};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Interuption due to Generic Failure Conditions", tsm.Status);
            Assert.AreEqual(18, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x13};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error Regulating Clutch Contact", tsm.Status);
            Assert.AreEqual(19, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x14};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Error AutoCalibrating Clutch Valve", tsm.Status);
            Assert.AreEqual(20, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x15};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("Slow Fingers Error", tsm.Status);
            Assert.AreEqual(21, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("No Out Of Range Insertions", tsm.Status);
            Assert.AreEqual(0, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x01};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("1a Out of Range", tsm.Status);
            Assert.AreEqual(1, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x02};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("2a Out of Range", tsm.Status);
            Assert.AreEqual(2, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x04};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("3a Out of Range", tsm.Status);
            Assert.AreEqual(4, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x8};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("4a Out of Range", tsm.Status);
            Assert.AreEqual(8, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x10};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("5a Out of Range", tsm.Status);
            Assert.AreEqual(16, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x20};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            Assert.AreEqual("RM Out of Range", tsm.Status);
            Assert.AreEqual(32, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(100, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(60.0000038f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(60.0000038f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(3.7999999f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(3.7999999f, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0x9C};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-3.7999999f, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(3.5, tsm.Measurement);

            i++;

            Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
            li = ecu.Sensors[i] as LocalIdentifier;
            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(3.5, tsm.Measurement);

            device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0xFF, 0x9C};
            tsm = ecu.ReadSensor(ecu.Sensors[i]);
            
            Assert.AreEqual(-3.5, tsm.Measurement);

            i++;

            for (int x = 0; x < 9; x++)
            {
                Console.WriteLine(string.Format("Testing Sensor: [{0}] {1}", i, ecu.Sensors[i].Name));
                li = ecu.Sensors[i] as LocalIdentifier;
                device.LocalidentifierData = new byte[] {0x61, (Byte) li.Id, 0x00, 0x64};
                tsm = ecu.ReadSensor(ecu.Sensors[i]);
                
                Assert.AreEqual(100, tsm.Measurement);

                i++;
            }
        }

        [TestFixtureTearDown]
        public void Dispose()
        {
            if (ecu != null)
            {
                ecu.Close();
                ecu.Dispose();
            }
        }
    }
}