#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Diagnostics;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;
using DnEcuDiag.Forms;
using DnEcuDiag.Properties;

namespace DnEcuDiag
{
    ///	<summary>
    ///	The	main application entry point.
    ///	</summary>
    internal static class Boot
    {
        private static Mutex _mutex;
        private static Boolean _owned;

        [STAThread]
        private static void Main()
        {
            Thread.CurrentThread.CurrentUICulture = Thread.CurrentThread.CurrentCulture;

            AppDomain.CurrentDomain.UnhandledException += CurrentDomain_UnhandledException;
            Application.ThreadException += UiThreadException;

            MainForm form = null;
            if (Settings.Default.EnableVisualStyles)
            {
                Application.EnableVisualStyles();
                Application.SetCompatibleTextRenderingDefault(false);
            }

            try
            {
                _mutex = new Mutex(true, Assembly.GetExecutingAssembly().GetName().Name, out _owned);

                if (!_owned)
                {
                    RaiseOtherProcess();
                    return;
                }
                else
                {
                    form = new MainForm();
                    Application.Run(form);
                }
            }
            catch (Exception exp)
            {
                Common.SaveException(exp);

                if (Common.ShowYesNoError(WinFormStrings.AnUnexpectedErrorHasOccuredWouldYouLikeToViewInformationAboutTheError) == DialogResult.Yes)
                {
                    TraceExceptionForm.Display(exp);
                }
            }
            finally
            {
                if ((_mutex != null) & (_owned))
                {
                    _mutex.ReleaseMutex();
                }

                if (form != null)
                {
                    form.Dispose();
                }
            }
        }

        private static void UiThreadException(object sender, ThreadExceptionEventArgs e)
        {
            HandleException(e.Exception);
        }

        private static void CurrentDomain_UnhandledException(object sender, UnhandledExceptionEventArgs e)
        {
            HandleException(e.ExceptionObject as Exception);
        }

        private static void HandleException(Exception exception)
        {
            try
            {
                if (exception != null)
                {
                    Common.DumpException(exception);

                    if (!EventLog.SourceExists("ThreadException"))
                    {
                        EventLog.CreateEventSource("ThreadException", Common.ApplicationName);
                    }

                    var myLog = new EventLog {Source = "ThreadException"};
                    myLog.WriteEntry("An unexpected error has occured " + exception.Message + "\n\nStack Trace:\n" + exception.StackTrace);

                    if (Common.ShowYesNoError(WinFormStrings.AnUnexpectedErrorHasOccuredWouldYouLikeToViewInformationAboutTheError) == DialogResult.Yes)
                    {
                        TraceExceptionForm.Display(exception);
                    }
                }
            }
            catch (Exception exp)
            {
                Common.DumpException(exception);

                try
                {
                    MessageBox.Show(String.Format(WinFormStrings.CouldNotWriteTheErrorToTheEventLogReason, exp.Message), WinFormStrings.FatalError, MessageBoxButtons.OK,
                                    MessageBoxIcon.Stop);
                }
                finally
                {
                    Application.Exit();
                }
            }
        }

        private static void RaiseOtherProcess()
        {
            Process proc = Process.GetCurrentProcess();
            // Using Process.ProcessName does not function properly	when
            // the actual name exceeds 15 characters. Using	the	assembly 
            // name	takes care of this quirk and is	more accruate than 
            // other work arounds.
            string assemblyName = Assembly.GetExecutingAssembly().GetName().Name;

            foreach (Process otherProc in
                Process.GetProcessesByName(assemblyName))
            {
                //ignore "this"	process
                if (proc.Id != otherProc.Id)
                {
                    // Found a "same named process".
                    // Assume it is	the	one	we want	brought	to the foreground.
                    // Use the Win32 API to	bring it to	the	foreground.
                    IntPtr hWnd = otherProc.MainWindowHandle;
                    if (NativeMethods.IsIconic(hWnd))
                    {
                        NativeMethods.ShowWindowAsync(hWnd, NativeMethods.SwRestore);
                    }
                    NativeMethods.SetForegroundWindow(hWnd);
                    break;
                }
            }
        }
    }
}