﻿#region License
// Copyright © 2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Windows.Forms;
using DnEcuDiag.DataBinding;
using DnEcuDiag.Ecus.Generic;
using DnEcuDiag.Forms;

namespace DnEcuDiag
{
    class Common
    {
        public const string ApplicationName = "DnEcuDiag";
        
        private Common()
        { }

        public static void ExecuteExceptionHandledDelegate(ExceptionHandledDelegate method)
        {
            try
            {
                method();
            }
            catch (Exception exp)
            {
                SaveException(exp);

                if (ShowYesNoError(WinFormStrings.AnUnexpectedErrorHasOccuredWouldYouLikeToViewInformationAboutTheError) == DialogResult.Yes)
                {
                    TraceExceptionForm.Display(exp);
                }
            }
        }

        public delegate void ExceptionHandledDelegate();

        public static void SaveException(Exception exception)
        {
            Assembly assembly = Assembly.GetExecutingAssembly();

            string file = Model.Instance.ApplicationPath + Path.DirectorySeparatorChar + "Error.txt";
            file = file.Replace("file:\\", "");

            using (var sw = new StreamWriter(file, true))
            {
                sw.WriteLine("".PadRight(100, '='));
                sw.WriteLine(WinFormStrings.VersionTimestamp, assembly.GetName().Version, DateTime.Now);
                sw.WriteLine("".PadRight(100, '='));
                sw.WriteLine(exception.ToString());
            }
        }

        /// <summary>
        /// Centralised yes no dialog routine.
        /// </summary>
        /// <param name="msg">The message to display.</param>
        /// <param name="args">Additional information.</param>
        /// <returns>The dialog result.</returns>
        public static DialogResult ShowYesNoError(String msg, params object[] args)
        {
            string message = String.Format(CultureInfo.InvariantCulture, msg, args);
            return
                MessageBox.Show(message, ApplicationName, MessageBoxButtons.YesNo, MessageBoxIcon.Error,
                                MessageBoxDefaultButton.Button1);
        }

        /// <summary>
        /// Centralised message box routine.
        /// </summary>
        /// <param name="msg">The message to display.</param>
        /// <param name="args">Additional information.</param>
        /// <returns>The dialog result.</returns>
        public static DialogResult ShowMessageExclamation(string msg, params object[] args)
        {
            string message = String.Format(CultureInfo.InvariantCulture, msg, args);
            return
                MessageBox.Show(message, ApplicationName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation,
                                MessageBoxDefaultButton.Button1);
        }

        /// <summary>
        /// Centralised message box routine.
        /// </summary>
        /// <param name="msg">The message to display.</param>
        /// <param name="args">Additional information.</param>
        /// <returns>The dialog result.</returns>
        public static DialogResult ShowMessageError(string msg, params object[] args)
        {
            string message = String.Format(CultureInfo.InvariantCulture, msg, args);
            return
                MessageBox.Show(message, ApplicationName, MessageBoxButtons.OK, MessageBoxIcon.Error,
                                MessageBoxDefaultButton.Button1);
        }

        /// <summary>
        /// Centralised message box routine.
        /// </summary>
        /// <param name="msg">The message to display.</param>
        /// <param name="args">Additional information.</param>
        /// <returns>The dialog result.</returns>
        public static DialogResult ShowMessage(string msg, params object[] args)
        {
            string message = String.Format(CultureInfo.InvariantCulture, msg, args);
            return
                MessageBox.Show(message, ApplicationName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation,
                                MessageBoxDefaultButton.Button1);
        }

        /// <summary>
        /// Centralised yes no dialog routine.
        /// </summary>
        /// <param name="msg">The message to display.</param>
        /// <param name="args">Additional information.</param>
        /// <returns>The dialog result.</returns>
        public static DialogResult ShowYesNo(string msg, params object[] args)
        {
            string message = String.Format(CultureInfo.InvariantCulture, msg, args);
            return
                MessageBox.Show(message, ApplicationName, MessageBoxButtons.YesNo, MessageBoxIcon.Question,
                                MessageBoxDefaultButton.Button1);
        }

        internal static SortableSearchableList<Sensor> ConfigureSensorBindingSource(System.Collections.ObjectModel.Collection<Sensor> collection)
        {
            var list = new SortableSearchableList<Sensor>();

            foreach (var sensor in collection)
            {
                list.Add(sensor);
            }

            return list;
        }

        internal static void DumpException(Exception exception)
        {
            using (var sw = new StreamWriter("UnhandledException.txt", true))
            {
                sw.WriteLine("".PadRight(100, '='));
                sw.WriteLine(WinFormStrings.VersionTimestamp, Assembly.GetAssembly(typeof(Common)).GetName().Version, DateTime.Now);
                sw.WriteLine("".PadRight(100, '='));
                sw.WriteLine(exception.ToString());
            }
        }
    }
}