﻿#region License
// Copyright © 2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace DnEcuDiag.CustomControls
{
    public partial class ComboBoxExtended : ComboBox
    {
        private const Int32 SwpNosize = 0x1;
        private const UInt32 WmCtlcolorlistbox = 0x0134;

        //Store the default width to perform check in UpdateDropDownWidth
        private readonly int _initialDropDownWidth;

        public ComboBoxExtended()
        {
            InitializeComponent();

            _initialDropDownWidth = DropDownWidth;

            HandleCreated += HandleCreatedExtended;
        }

        public ComboBoxExtended(IContainer container) : this()
        {
            container.Add(this);
        }

        [DllImport("user32.dll")]
        private static extern bool SetWindowPos(IntPtr hWnd, IntPtr hWndInsertAfter, int x, int y, int cx, int cy, uint uFlags);

        private void HandleCreatedExtended(object sender, EventArgs e)
        {
            UpdateDropDownWidth();
        }

        private void UpdateDropDownWidth()
        {
            //Create a GDI+ drawing surface to measure string widths
            Graphics ds = CreateGraphics();

            //Float to hold largest single item width
            float maxWidth = 0;

            //Iterate over each item, measuring the maximum width
            //of the DisplayMember strings
            foreach (object item in Items)
            {
                maxWidth = Math.Max(maxWidth, ds.MeasureString(item.ToString(), Font).Width);
            }

            //Add a buffer for some white space
            //around the text
            maxWidth += 20;

            //round maxWidth and cast to an int
            var newWidth = (int) Decimal.Round((decimal) maxWidth, 0);

            //If the width is bigger than the screen, ensure
            //we stay within the bounds of the screen
            if (newWidth > Screen.GetWorkingArea(this).Width)
            {
                newWidth = Screen.GetWorkingArea(this).Width;
            }

            //Only change the default width if it's smaller
            //than the newly calculated width
            if (newWidth > _initialDropDownWidth)
            {
                DropDownWidth = newWidth;
            }

            //Clean up the drawing surface
            ds.Dispose();
        }

        protected override void WndProc(ref Message m)
        {
            if (m.Msg == WmCtlcolorlistbox)
            {
                // Make sure we are inbounds of the screen
                int left = PointToScreen(new Point(0, 0)).X;

                //Only do this if the dropdown is going off right edge of screen
                if (DropDownWidth > Screen.PrimaryScreen.WorkingArea.Width - left)
                {
                    // Get the current combo position and size
                    Rectangle comboRect = RectangleToScreen(ClientRectangle);

                    int dropHeight = 0;
                    int topOfDropDown;

                    //Calculate dropped list height
                    for (int i = 0; (i < Items.Count && i < MaxDropDownItems); i++)
                    {
                        dropHeight += ItemHeight;
                    }

                    //Set top position of the dropped list if 
                    //it goes off the bottom of the screen
                    if (dropHeight > Screen.PrimaryScreen.WorkingArea.Height -
                                     PointToScreen(new Point(0, 0)).Y)
                    {
                        topOfDropDown = comboRect.Top - dropHeight - 2;
                    }
                    else
                    {
                        topOfDropDown = comboRect.Bottom;
                    }

                    //Calculate shifted left position
                    int leftOfDropDown = comboRect.Left - (DropDownWidth -
                                                           (Screen.PrimaryScreen.WorkingArea.Width - left));

                    // Postioning/sizing the drop-down
                    //SetWindowPos(HWND hWnd,
                    //      HWND hWndInsertAfter,
                    //      int X,
                    //      int Y,
                    //      int cx,
                    //      int cy,
                    //      UINT uFlags);
                    //when using the SwpNosize flag, cx and cy params are ignored
                    SetWindowPos(m.LParam,
                                 IntPtr.Zero,
                                 leftOfDropDown,
                                 topOfDropDown,
                                 0,
                                 0,
                                 SwpNosize);
                }
            }

            base.WndProc(ref m);
        }
    }
}