#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

namespace DnEcuDiag.CustomControls
{
    public class TabControlExtended : TabControl
    {
        private Container _components;

        public TabControlExtended()
        {
            // This call is required by the Windows.Forms Form Designer.
            InitializeComponent();
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (_components != null)
                {
                    _components.Dispose();
                }
            }
            base.Dispose(disposing);
        }

        [Description("Occurs as a tab is being changed.")]
        public event EventHandler<TabPageChangeEventArgs> SelectedIndexChanging;

        protected override void WndProc(ref Message m)
        {
            //Tab Selection via mouse button occurs before OnMouseDown()
            //so we need to catch it here instead.
            if (m.Msg == 0x201) //WM_MOUSEDOWN
            {
                TabPage tp = TestTab(new Point(m.LParam.ToInt32()));
                if (tp != null)
                {
                    if (tp.Enabled == false)
                    {
                        return;
                    }

                    var ev = new TabPageChangeEventArgs(SelectedTab, tp);
                    OnSelectedIndexChanging(ev);
                    return;
                }
            }
            base.WndProc(ref m);
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyValue >= 37 && e.KeyValue <= 40)
            {
                e.Handled = true;

                TabPage tp;
                Rectangle r = GetTabRect(SelectedIndex);

                bool foundNextTab = false;
                do
                {
                    Point pt;
                    if (e.KeyCode == Keys.Left)
                    {
                        pt = new Point(r.Left - 3, r.Top);
                    }
                    else if (e.KeyCode == Keys.Up)
                    {
                        pt = new Point(r.Left, r.Top - 3);
                    }
                    else if (e.KeyCode == Keys.Right)
                    {
                        pt = new Point(r.Right + 3, r.Top);
                    }
                    else
                    {
                        pt = new Point(r.Left, r.Bottom + 3);
                    }

                    tp = TestTab(pt);
                    if (tp != null)
                    {
                        foundNextTab = tp.Enabled;
                        r = GetTabRect(TabPages.IndexOf(tp));
                    }
                } while (tp != null && !foundNextTab);

                if (tp == null)
                    tp = GetNextEnabledTab(e.KeyValue > 38, false);

                if (tp != null && tp != SelectedTab)
                {
                    var ev = new TabPageChangeEventArgs(SelectedTab, tp);
                    OnSelectedIndexChanging(ev);
                }
            }

            base.OnKeyDown(e);
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            if ((keyData == (Keys.Tab | Keys.Control)) || (keyData == (Keys.Tab | Keys.Shift | Keys.Control)))
            {
                TabPage tp = GetNextEnabledTab((keyData & Keys.Shift) == Keys.None, true);
                if (tp != null)
                    if (tp != SelectedTab)
                    {
                        var ev = new TabPageChangeEventArgs(SelectedTab, tp);
                        OnSelectedIndexChanging(ev);
                    }
                return true;
            }
            return base.ProcessCmdKey(ref msg, keyData);
        }

        protected virtual void OnSelectedIndexChanging(TabPageChangeEventArgs e)
        {
            if (SelectedIndexChanging != null)
                SelectedIndexChanging(this, e);
            if (!e.Cancel)
                SelectedTab = e.NextTab;
        }

        private TabPage TestTab(Point pt)
        {
            for (int index = 0; index <= TabCount - 1; index++)
                if (GetTabRect(index).Contains(pt.X, pt.Y))
                    return TabPages[index];
            return null;
        }

        private TabPage GetNextEnabledTab(bool forward, bool wrap)
        {
            if (forward)
            {
                for (int index = SelectedIndex + 1; index < TabCount; index++)
                {
                    if (TabPages[index].Enabled)
                    {
                        return TabPages[index];
                    }
                }

                if (wrap)
                {
                    for (int index = 0; index < SelectedIndex; index++)
                    {
                        if (TabPages[index].Enabled)
                        {
                            return TabPages[index];
                        }
                    }
                }
            }
            else
            {
                for (int index = SelectedIndex - 1; index >= 0; index--)
                {
                    if (TabPages[index].Enabled)
                    {
                        return TabPages[index];
                    }
                }

                if (wrap)
                {
                    for (int index = TabCount - 1; index > SelectedIndex; index--)
                    {
                        if (TabPages[index].Enabled)
                        {
                            return TabPages[index];
                        }
                    }
                }
            }

            return null;
        }

        #region Component Designer generated code

        /// <summary> 
        /// Required method for Designer support - do not modify 
        /// the contents of this method with the code editor.
        /// </summary>
        private void InitializeComponent()
        {
            _components = new System.ComponentModel.Container();
        }

        #endregion
    }
}