#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Globalization;
using System.Text;
using System.Threading;
using System.Xml.Serialization;
using DnEcuDiag.Conversion;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Ecus.EventArgTypes;
using DnEcuDiag.Ecus.Exceptions;
using DnEcuDiag.Ecus.Iso;
using DnEcuDiag.Ecus.Kwp;

namespace DnEcuDiag.Ecus.Generic
{
    [Serializable]
    public abstract class Ecu : IDisposable
    {
        #region	Events

        ///	<summary>
        ///	Event used to signal a disconnection from the ecu.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<EcuDisconnectEventArgs> EcuDisconnected { get; set; }

        ///	<summary>
        ///	Occurs after Sensor	One	updated.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorUpdatedEventArgs> SensorOneUpdated { get; set; }

        ///	<summary>
        ///	Occurs after Sensor	two updated.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorUpdatedEventArgs> SensorTwoUpdated { get; set; }

        ///	<summary>
        ///	Occurs after Sensor	three updated.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorUpdatedEventArgs> SensorThreeUpdated { get; set; }

        ///	<summary>
        ///	Occurs after Sensor	One	updated.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorUpdatedEventArgs> SensorFourUpdated { get; set; }

        ///	<summary>
        ///	Occurs if the Sensor One read operation	is not suported.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorReadErrorEventArgs> SensorOneReadError { get; set; }

        ///	<summary>
        ///	Occurs if the Sensor One read operation	is not suported.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorReadErrorEventArgs> SensorTwoReadError { get; set; }

        ///	<summary>
        ///	Occurs if the Sensor One read operation	is not suported.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorReadErrorEventArgs> SensorThreeReadError { get; set; }

        ///	<summary>
        ///	Occurs if the Sensor One read operation	is not suported.
        ///	</summary>
        [XmlIgnore]
        public EventHandler<SensorReadErrorEventArgs> SensorFourReadError { get; set; }

        #endregion

        #region Fields

        private readonly Object _lockObject = new Object();
        [NonSerialized]
        private Thread _dataCaptureThread;
        [NonSerialized]
        private Sensor _sensorMonitorFour;
        private Boolean _sensorMonitorFourEnable;
        [NonSerialized]
        private Sensor _sensorMonitorOne;
        private Boolean _sensorMonitorOneEnable;
        [NonSerialized]
        private Sensor _sensorMonitorThree;
        private Boolean _sensorMonitorThreeEnable;
        [NonSerialized]
        private Sensor _sensorMonitorTwo;
        private Boolean _sensorMonitorTwoEnable;
        [NonSerialized]
        private ManualResetEvent _signalCaptureData;
        private Boolean _stopDataCapture;

        #endregion

        #region Constructors

        protected Ecu()
        {
            HardwareNumber = String.Empty;
            ActiveFaultCodes = new Collection<FaultCode>();
            Actuators = new Collection<Actuator>();
            Sensors = new Collection<Sensor>();
            SnapshotDefinitions = new Collection<SnapshotDefinition>();
            Routines = new Collection<Routine>();
            SerialNumber = String.Empty;
            ProgrammingDate = String.Empty;
            ProductionDate = String.Empty;
            IsoCode = String.Empty;
            OmologationNumber = String.Empty;
            SoftwareVersionNumber = String.Empty;
            SoftwareNumber = String.Empty;
            HardwareVersionNumber = String.Empty;
            EcuWarning = String.Empty;
            DrawingNumber = String.Empty;
            Model = String.Empty;
            CreateAndStartDataCaptureThread();
        }

        #endregion

        #region Dispose Pattern

        private bool _disposed;

        /// <summary>
        /// Dispose the underlying resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Close the device.
        /// </summary>
        public void Close()
        {
            Dispose(true);
        }

        /// <summary>
        /// Dispose of the objects resources.
        /// </summary>
        /// <param name="disposing">True if explicity disposing object else GC finalizing object.</param>
        private void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    if (_dataCaptureThread != null)
                    {
                        _stopDataCapture = true;

                        if (!_dataCaptureThread.Join(1000))
                        {
                            _dataCaptureThread.Abort();
                        }

                        _dataCaptureThread = null;
                    }

                    if (_signalCaptureData != null)
                    {
                        _signalCaptureData.Close();
                        _signalCaptureData = null;
                    }
                }
            }

            _disposed = true;
        }

        #endregion

        #region Properties

        ///	<summary>
        ///	The	Ecu	model name.
        ///	</summary>
        public String Model { get; set; }

        ///	<summary>
        ///	The	type of <see cref="IConvertor"/> to use for conversions.
        ///	</summary>
        public String ConvertorType { get; set; }

        /// <summary>
        /// Is ecu connected.
        /// </summary>
        [XmlIgnore]
        public bool Connected { get; protected set; }

        ///	<summary>
        ///	The	Ecu	drawing	number.
        ///	</summary>
        [XmlIgnore]
        public string DrawingNumber { get; protected set; }

        /// <summary>
        /// The Error memory Structure for this ecu.
        /// </summary>
        public ErrorMemoryDefinition ErrorMemoryStructureDefinition { get; set; }

        ///	<summary>
        ///	The	ecu	address.
        ///	</summary>
        public Int32 EcuAddress { get; set; }

        ///	<summary>
        ///	A warning message for this ecu.	The	UI should prompt the user with this	warning	message.
        ///	</summary>
        public string EcuWarning { get; set; }

        ///	<summary>
        ///	The	hardware number.
        ///	</summary>
        [XmlIgnore]
        public string HardwareNumber { get; protected set; }

        ///	<summary>
        ///	The	Ecu	hardware number.
        ///	</summary>
        [XmlIgnore]
        public string HardwareVersionNumber { get; protected set; }

        ///	<summary>
        ///	The	software number.
        ///	</summary>
        [XmlIgnore]
        public string SoftwareNumber { get; protected set; }

        ///	<summary>
        ///	The	Ecu	Software Number
        ///	</summary>
        [XmlIgnore]
        public string SoftwareVersionNumber { get; protected set; }

        ///	<summary>
        ///	The	Omologation	number.
        ///	</summary>
        [XmlIgnore]
        public string OmologationNumber { get; protected set; }

        ///	<summary>
        ///	The	iso	number.
        ///	</summary>
        [XmlIgnore]
        public string IsoCode { get; protected set; }

        ///	<summary>
        ///	The	ecu	production date.
        ///	</summary>
        [XmlIgnore]
        public string ProductionDate { get; protected set; }

        ///	<summary>
        ///	The	programming	date. 
        ///	</summary>
        [XmlIgnore]
        public string ProgrammingDate { get; protected set; }

        ///	<summary>
        ///	The	ecu	serial number.
        ///	</summary>
        [XmlIgnore]
        public string SerialNumber { get; protected set; }

        ///	<summary>
        ///	The	Rpm	sensor.
        ///	</summary>
        public Sensor RpmSensor { get; set; }

        ///	<summary>
        ///	The	Speed sensor.
        ///	</summary>
        public Sensor SpeedSensor { get; set; }

        [XmlIgnore]
        public int MonitoredSensorRefreshDelay { get; set; }

        [XmlIgnore]
        public Sensor SensorMonitorOne
        {
            get { return _sensorMonitorOne; }
            set
            {
                lock (_lockObject)
                {
                    _sensorMonitorOne = value;
                }
            }
        }

        [XmlIgnore]
        public Sensor SensorMonitorTwo
        {
            get { return _sensorMonitorTwo; }
            set
            {
                lock (_lockObject)
                {
                    _sensorMonitorTwo = value;
                }
            }
        }

        [XmlIgnore]
        public Sensor SensorMonitorThree
        {
            get { return _sensorMonitorThree; }
            set
            {
                lock (_lockObject)
                {
                    _sensorMonitorThree = value;
                }
            }
        }

        [XmlIgnore]
        public Sensor SensorMonitorFour
        {
            get { return _sensorMonitorFour; }
            set
            {
                lock (_lockObject)
                {
                    _sensorMonitorFour = value;
                }
            }
        }

        [XmlIgnore]
        public Boolean SensorMonitorOneEnable
        {
            get { return _sensorMonitorOneEnable; }
            set
            {
                if (value)
                {
                    _sensorOneMinimum = null;
                    _sensorOneMaximum = null;
                }

                _sensorMonitorOneEnable = value;
            }
        }

        [XmlIgnore]
        public Boolean SensorMonitorTwoEnable
        {
            get { return _sensorMonitorTwoEnable; }
            set
            {
                if (value)
                {
                    _sensorTwoMinimum = null;
                    _sensorTwoMaximum = null;
                }

                _sensorMonitorTwoEnable = value;
            }
        }

        [XmlIgnore]
        public Boolean SensorMonitorThreeEnable
        {
            get { return _sensorMonitorThreeEnable; }
            set
            {
                if (value)
                {
                    _sensorThreeMinimum = null;
                    _sensorThreeMaximum = null;
                }

                _sensorMonitorThreeEnable = value;
            }
        }

        [XmlIgnore]
        public Boolean SensorMonitorFourEnable
        {
            get { return _sensorMonitorFourEnable; }
            set
            {
                if (value)
                {
                    _sensorFourMinimum = null;
                    _sensorFourMaximum = null;
                }

                _sensorMonitorFourEnable = value;
            }
        }

        ///	<summary>
        ///	The	collection of <see cref="Routine"/> that this ecu supports.
        ///	</summary>
        public Collection<Routine> Routines { get; set; }

        ///	<summary>
        ///	Gets the snap shot definitions.
        ///	</summary>
        ///	<value>The snap	shot definitions.</value>
        public Collection<SnapshotDefinition> SnapshotDefinitions { get; set; }

        ///	<summary>
        ///	The	collection of ecu sensors.
        ///	</summary>
        [XmlArrayItem(Type = typeof(IsoSensor)), XmlArrayItem(Type = typeof(LocalIdentifier))]
        public Collection<Sensor> Sensors { get; set; }

        ///	<summary>
        ///	The	collection of ecu actuators.
        ///	</summary>
        [XmlArrayItem(Type = typeof(InputOutputLocalIdentifier)), XmlArrayItem(Type = typeof(IsoActuator))]
        public Collection<Actuator> Actuators { get; set; }

        ///	<summary>
        ///	Gets the active	fault codes.
        ///	</summary>
        ///	<value>The active fault	codes.</value>
        [XmlIgnore]
        public Collection<FaultCode> ActiveFaultCodes { get; private set; }

        /// <summary>
        /// The convertor class to convert sensor data bytes.
        /// </summary>
        [XmlIgnore]
        public IConvertor Convertor { get; set; }

        #endregion

        #region Abstract Properties

        protected IDevice diagnosticInterface;
        private TimeStampedMeasurement _sensorFourMaximum;
        private TimeStampedMeasurement _sensorFourMinimum;
        private TimeStampedMeasurement _sensorOneMaximum;
        private TimeStampedMeasurement _sensorOneMinimum;
        private TimeStampedMeasurement _sensorThreeMaximum;
        private TimeStampedMeasurement _sensorThreeMinimum;
        private TimeStampedMeasurement _sensorTwoMaximum;
        private TimeStampedMeasurement _sensorTwoMinimum;
        private ThreadPriority _threadPriority;

        ///	<summary>
        ///	The	underlying interface device.
        ///	</summary>
        [XmlIgnore]
        public abstract IDevice DiagnosticInterface { get; set; }

        #endregion

        #region Public Methods

        ///	<summary>
        ///	Connect	to the underlying Ecu.
        ///	</summary>
        public void Connect()
        {
            Trace.WriteLine("->Connect()");

            if (diagnosticInterface == null)
            {
                throw new EcuException("The Interface has not been configured.");
            }

            try
            {
                BoostThreadPrioity();

                ActiveFaultCodes.Clear();

                ConnectImplemenation();
            }
            finally
            {
                LowerThreadPriority();
            }
            Trace.WriteLine("<-Connect()");
        }

        ///	<summary>
        ///	Stop monitoring	the	MonitoredSensors collection.
        ///	</summary>
        public void StopSensorMonitoring()
        {
            _signalCaptureData.Reset();
        }

        ///	<summary>
        ///	Start monitoring the MonitoredSensors collection.
        ///	</summary>
        public void StartSensorMonitoring()
        {
            _signalCaptureData.Set();
        }

        #endregion

        #region Protected methods

        protected virtual void OnEcuDisconnect(EcuDisconnectEventArgs e)
        {
            ClearEcuInstanceData();

            if (EcuDisconnected != null)
            {
                EcuDisconnected(this, e);
            }
        }

        protected virtual void HandleProtocolDisconnect(Object sender, ProtocolDisconnectedEventArgs e)
        {
            Connected = false;

            OnEcuDisconnect(new EcuDisconnectEventArgs(e.Exception));
        }

        protected static string DecodeSnapshot(byte[] snapshotData, SnapshotDefinition definitionToEvaluate,
                                               IConvertor convertor)
        {
            try
            {
                var sb = new StringBuilder();

                if (!String.IsNullOrEmpty(definitionToEvaluate.Name))
                {
                    sb.Append(definitionToEvaluate.Name);
                    sb.Append(Environment.NewLine);
                    sb.Append(Environment.NewLine);
                }

                foreach (Element e in definitionToEvaluate.Elements)
                {
                    if (!String.IsNullOrEmpty(e.Name))
                    {
                        sb.Append("\t");
                        sb.Append(e.Name);
                        sb.Append(Environment.NewLine);
                    }

                    foreach (ElementDefinition ed in e.ElementDefinitions)
                    {
                        foreach (
                            ElementEvaluation eval in
                                ElementEvaluator.Evaluate(snapshotData, convertor, ed))
                        {
                            if (ValidEvaluation(eval))
                            {
                                sb.Append("\t\t");

                                if (!String.IsNullOrEmpty(eval.Name))
                                {
                                    sb.Append(eval.Name);
                                    sb.Append(": ");
                                }

                                if (eval.Value.GetType() == typeof(float))
                                {
                                    sb.Append(((Single)eval.Value).ToString("0.###", CultureInfo.InvariantCulture));

                                    if (!String.IsNullOrEmpty(ed.Units))
                                    {
                                        sb.Append(" ");
                                        sb.Append(ed.Units);
                                    }

                                    sb.Append(Environment.NewLine);
                                }
                                else
                                {
                                    var maskMatchResult = (MaskMatchResult)eval.Value;
                                    if (!String.IsNullOrEmpty(maskMatchResult.Result))
                                    {
                                        sb.Append(maskMatchResult.Result);
                                        sb.Append(Environment.NewLine);
                                    }
                                }
                            }
                        }
                    }

                    sb.Append(Environment.NewLine);
                }

                return sb.ToString();
            }
            catch (Exception exp)
            {
                throw new EcuException("Failed executing DecodeSnapShot.", exp);
            }
        }


        private static bool ValidEvaluation(ElementEvaluation eval)
        {
            Boolean evaulationValid = !String.IsNullOrEmpty(eval.Name);

            var maskMatchResult = eval.Value as MaskMatchResult;

            if (maskMatchResult != null)
            {
                evaulationValid = evaulationValid | !String.IsNullOrEmpty(maskMatchResult.Result);
            }

            return evaulationValid;
        }

        protected string DecodeFaultCode(String faultCode, Byte[] faultCodeData, IConvertor convertor)
        {
            var sb = new StringBuilder();

            foreach (ErrorDefinition definition in ErrorMemoryStructureDefinition.ErrorDefinitions)
            {
                if (definition.ErrorCode == faultCode)
                {
                    sb.Append("Failed Component Code: ");
                    sb.Append(definition.ErrorDescription);
                    sb.Append(Environment.NewLine);

                    foreach (ElementDefinition errorElement in definition.ErrorElements)
                    {
                        Collection<ElementEvaluation> evaluations = ElementEvaluator.Evaluate(faultCodeData, convertor, errorElement);

                        foreach (ElementEvaluation evaluation in evaluations)
                        {
                            sb.Append(evaluation.Name);
                            sb.Append(": ");

                            if (evaluation.Value is float)
                            {
                                sb.Append(((float)evaluation.Value).ToString("0.##", CultureInfo.InvariantCulture));
                            }
                            else
                            {
                                sb.Append(((MaskMatchResult)evaluation.Value).Result);
                            }

                            if (errorElement.Units.Length != 0)
                            {
                                sb.Append(" ");
                                sb.Append(errorElement.Units);
                            }

                            sb.Append(Environment.NewLine);
                        }
                    }
                    break;
                }
            }

            return sb.ToString();
        }

        #endregion

        #region	Abstract Methods

        ///	<summary>
        ///	Disconnect from	the	underlying Ecu.
        ///	</summary>
        public abstract void Disconnect();

        ///	<summary>
        ///	Clear the fault	codes for the Ecu.
        ///	</summary>
        public abstract void ClearFaultCodes();

        ///	<summary>
        ///	Read the active	fault codes	from the Ecu.
        ///	</summary>
        public abstract void ReadFaultCodes();

        ///	<summary>
        ///	Reads the snap shot.
        ///	</summary>
        ///	<param name="snapshotToRead">The snap shot to read.</param>
        ///	<returns>The SnapShot evaluation.</returns>
        /// <exception cref="SnapshotNotSupportedException"></exception>
        public abstract String ReadSnapshot(SnapshotDefinition snapshotToRead);

        ///	<summary>
        ///	Reads the sensor.
        ///	</summary>
        ///	<param name="sensor">The sensor.</param>
        ///	<returns></returns>
        /// <exception cref="SensorNotSupportedException"></exception>
        public abstract TimeStampedMeasurement ReadSensor(Sensor sensor);

        ///	<summary>
        ///	Activates the actuator.
        ///	</summary>
        ///	<param name="actuator">The actuator.</param>
        ///	<returns></returns>
        public abstract bool ActivateActuator(Actuator actuator);

        ///	<summary>
        ///	Execute	the	selected routine.
        ///	</summary>
        ///	<param name="routine">The Routine to execute.</param>
        ///	<returns>The result	of the routine execution.</returns>
        public abstract UInt32 ExecuteRoutine(Routine routine);

        /// <summary>
        /// Cancel any executing routines.
        /// </summary>
        public abstract void ExecuteRoutineCancel();

        internal abstract void ConnectImplemenation();

        #endregion Methods

        #region Private Methods

        private void CreateAndStartDataCaptureThread()
        {
            _signalCaptureData = new ManualResetEvent(false);
            _dataCaptureThread = new Thread(DataCaptureMethod) { Name = "Capture", IsBackground = true, Priority = ThreadPriority.AboveNormal };
            _dataCaptureThread.Start();
        }

        private void DataCaptureMethod()
        {
            try
            {
                while (true)
                {
                    CaptureMonitoredSensor(ref _sensorMonitorOneEnable, ref _sensorOneMinimum, ref _sensorOneMaximum, _sensorMonitorOne, SensorOneUpdated,
                                           SensorOneReadError);

                    Thread.Sleep(5);

                    CaptureMonitoredSensor(ref _sensorMonitorTwoEnable, ref _sensorTwoMinimum, ref _sensorTwoMaximum, _sensorMonitorTwo, SensorTwoUpdated,
                                           SensorTwoReadError);

                    Thread.Sleep(5);

                    CaptureMonitoredSensor(ref _sensorMonitorThreeEnable, ref _sensorThreeMinimum, ref _sensorThreeMaximum, _sensorMonitorThree, SensorThreeUpdated,
                                           SensorThreeReadError);

                    Thread.Sleep(5);

                    CaptureMonitoredSensor(ref _sensorMonitorFourEnable, ref _sensorFourMinimum, ref _sensorFourMaximum, _sensorMonitorFour, SensorFourUpdated,
                                           SensorFourReadError);

                    Thread.Sleep(MonitoredSensorRefreshDelay + 10);

                    _signalCaptureData.WaitOne();

                    if (_stopDataCapture)
                    {
                        break;
                    }
                }
            }
            catch (ThreadAbortException)
            {
            }
            catch (Exception exp)
            {
                Trace.WriteLine(exp.ToString());
            }
        }

        private void CaptureMonitoredSensor(ref bool sensorMonitorEnable, ref TimeStampedMeasurement minimum, ref TimeStampedMeasurement maximum, Sensor monitoredSensor,
                                            EventHandler<SensorUpdatedEventArgs> updateSensor, EventHandler<SensorReadErrorEventArgs> sensorNotSupported)
        {
            if (sensorMonitorEnable)
            {
                try
                {
                    TimeStampedMeasurement current = CaptureMonitoredSensor(monitoredSensor);

                    if(maximum != null)
                    {
                        if (current.Measurement > maximum.Measurement)
                        {
                            maximum = current;
                        }    
                    }
                    else
                    {
                        maximum = current;
                    }
                    
                    if(minimum != null)
                    {
                        if (current.Measurement < minimum.Measurement)
                        {
                            minimum = current;
                        }    
                    }
                    else
                    {
                        minimum = current;
                    }
                    if (updateSensor != null)
                    {
                        updateSensor(this, new SensorUpdatedEventArgs(current, minimum, maximum));
                    }
                }
                catch (Exception e)
                {
                    sensorMonitorEnable =
                        HandleSensorReadErrorException(monitoredSensor, sensorNotSupported, e);
                }
            }
        }

        private bool HandleSensorReadErrorException(Sensor monitoredSensor, EventHandler<SensorReadErrorEventArgs> sensorNotSupported, Exception exception)
        {
            if (Connected)
            {
                if (sensorNotSupported != null)
                {
                    sensorNotSupported(this, new SensorReadErrorEventArgs(monitoredSensor, exception));
                }
            }

            Trace.WriteLine(exception.ToString());

            return false;
        }

        private TimeStampedMeasurement CaptureMonitoredSensor(Sensor sensor)
        {
            TimeStampedMeasurement measurement = null;

            lock (_lockObject)
            {
                if (sensor != null)
                {
                    measurement = ReadSensor(sensor);
                }
            }

            return measurement;
        }

        protected void ClearEcuInstanceData()
        {
            DrawingNumber = String.Empty;
            HardwareNumber = String.Empty;
            HardwareVersionNumber = String.Empty;
            SoftwareNumber = String.Empty;
            SoftwareVersionNumber = String.Empty;
            OmologationNumber = String.Empty;
            IsoCode = String.Empty;
            ProductionDate = String.Empty;
            ProgrammingDate = String.Empty;
            SerialNumber = String.Empty;
            
            ActiveFaultCodes.Clear();
        }
        #endregion

        protected void EngineRunningCheck()
        {
            if (RpmSensor != null)
            {
                if (ReadSensor(RpmSensor).Measurement != 0)
                {
                    throw new EngineRunningException();
                }
            }
        }

        protected void BoostThreadPrioity()
        {
            _threadPriority = Thread.CurrentThread.Priority;
            Thread.CurrentThread.Priority = ThreadPriority.AboveNormal;
        }

        protected void LowerThreadPriority()
        {
            Thread.CurrentThread.Priority = _threadPriority;
        }
    }
}