#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.ObjectModel;
using DnEcuDiag.Conversion;

namespace DnEcuDiag.Ecus.Generic
{
    /// <summary>
    /// A class that evaluates and converts an element to an ElementEvaluation.
    /// </summary>
    public sealed class ElementEvaluator
    {
        #region Fields

        private static ElementEvaluator _instance;

        #endregion

        #region Singlton Pattern

        /// <summary>
        /// Prevent instantiation from outside of class.
        /// </summary>
        private ElementEvaluator()
        {
        }

        /// <summary>
        /// Singleton Pattern.
        /// </summary>
        public static ElementEvaluator Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new ElementEvaluator();
                }

                return _instance;
            }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Evaluate an element definition.
        /// </summary>
        /// <param name="rawData">The raw bytes to use for the evaluation of the ElementDefinition.</param>
        /// <param name="convertor">An instance of a class that implements IConvertor.</param>
        /// <param name="elementDefinition">The element to evaluate.</param>
        /// <returns>An ElementEvaluation.</returns>
        public static Collection<ElementEvaluation> Evaluate(byte[] rawData, IConvertor convertor,
                                                       ElementDefinition elementDefinition)
        {
            var evaluations = new Collection<ElementEvaluation>();
            var bytesToConvert = new byte[elementDefinition.ByteLenght];

            if (elementDefinition.ByteIndex >= rawData.Length)
            {
                if (elementDefinition.IsBit)
                {
                    evaluations.Add(new ElementEvaluation(elementDefinition.Name, new MaskMatchResult(0, string.Empty)));
                }
                else
                {
                    evaluations.Add(new ElementEvaluation(elementDefinition.Name, 0));
                }

                return evaluations;
            }

            Buffer.BlockCopy(rawData, elementDefinition.ByteIndex, bytesToConvert, 0, elementDefinition.ByteLenght);

            uint rawdata = Convertor.ConvertByteArray(bytesToConvert, elementDefinition.LittleEndian);

            uint val = rawdata;
            val &= elementDefinition.Bitmask;
            val = val >> elementDefinition.BitShift;

            if (elementDefinition.IsBit)
            {
                foreach (MaskMatch match in elementDefinition.MaskMatches)
                {
                    if (match.MatchValue)
                    {
                        if (match.MaskValue == val)
                        {
                            evaluations.Add(new ElementEvaluation(elementDefinition.Name,
                                                                  new MaskMatchResult(val, match.MaskResultTrue)));
                        }
                    }
                    else
                    {
                        if ((val & match.MaskValue) >= 1)
                        {
                            evaluations.Add(new ElementEvaluation(elementDefinition.Name,
                                                                  new MaskMatchResult(val, match.MaskResultTrue)));
                        }
                        else
                        {    
                            evaluations.Add(new ElementEvaluation(elementDefinition.Name,
                                                                  new MaskMatchResult(val, match.MaskResultFalse)));
                        }
                    }
                }
            }
            else
            {
                evaluations.Add(new ElementEvaluation(elementDefinition.Name,
                                                      convertor.Convert(elementDefinition.ConversionId, bytesToConvert)));
            }

            return evaluations;
        }

        #endregion Methods
    }
}