#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.Text;
using System.Threading;
using System.Xml.Serialization;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Ecus.Enums;
using DnEcuDiag.Ecus.Exceptions;
using DnEcuDiag.Ecus.Generic;
using DnEcuDiag.Protocols;

namespace DnEcuDiag.Ecus.Iso
{
    /// <summary>
    /// Represents an Iso9141 compliant ecu.
    /// </summary>
    [Serializable]
    public class Kwp81Ecu : Ecu, IDisposable
    {
        #region Fields
        [NonSerialized] protected Kwp81 Protocol;

        #endregion

        public Kwp81Ecu()
        {
            HeaderEndIndex = 2;
            CrcBytesCount = 1;
        }

        #region Dispose Pattern

        private bool _disposed;

        /// <summary>
        /// Dispose the underlying resources.
        /// </summary>
        public new void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Close the device.
        /// </summary>
        public new void Close()
        {
            Dispose(true);
        }

        /// <summary>
        /// Dispose of the objects resources.
        /// </summary>
        /// <param name="disposing">True if explicity disposing object else GC finalizing object.</param>
        private void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                base.Dispose();

                if (disposing)
                {
                    if (Protocol != null)
                    {
                        Protocol.Close();
                        Protocol.Dispose();
                        Protocol = null;
                    }
                }
            }

            _disposed = true;
        }

        #endregion Dispose Pattern

        #region Private Methods

        private void DecodeFaultCodeData(Byte[] faultCodeData)
        {
            try
            {
                var faultCodes = new List<Byte[]>();
                ActiveFaultCodes.Clear();

                Int32 numberOfFaults = faultCodeData.Length/ErrorMemoryStructureDefinition.Lenght;

                for (int i = 0; i < numberOfFaults; i++)
                {
                    var bytes = new byte[ErrorMemoryStructureDefinition.Lenght];

                    Buffer.BlockCopy(faultCodeData, i*ErrorMemoryStructureDefinition.Lenght, bytes, 0,
                                     ErrorMemoryStructureDefinition.Lenght);
                    faultCodes.Add(bytes);
                }

                foreach (var faultCodeBytes in faultCodes)
                {
                    String faultCode = faultCodeBytes[0].ToString("X4", CultureInfo.InvariantCulture);
                   
                    ActiveFaultCodes.Add(new FaultCode(faultCode,
                                                       DecodeFaultCode(faultCodeBytes[0].ToString("X", CultureInfo.InvariantCulture), faultCodeBytes, Convertor)));
                }
            }
            catch (Exception exp)
            {
                throw new EcuException("Failed executing DecodeFaultCodeData.", exp);
            }
        }

        #endregion

        [XmlIgnore]
        public override IDevice DiagnosticInterface
        {
            get { return diagnosticInterface; }
            set
            {
                diagnosticInterface = value;
                Protocol = new Kwp81(diagnosticInterface, (Byte) EcuAddress, 0);
            }
        }

        [XmlIgnore]
        public int CrcBytesCount { get; internal set; }

        [XmlIgnore]
        public int HeaderEndIndex { get; internal set; }

        /// <summary>
        /// Disconnect from the underlying Ecu.
        /// </summary>
        public override void Disconnect()
        {
            if (Protocol == null) throw new EcuException("The Interface has not been configured.");
            if (Connected == false) return;

            try
            {
                Protocol.EndDiagnostics();

                Connected = false;
            }
            catch (Kwp81Exception exp)
            {
                throw new EcuException("Failed executing Disconnect.", exp);
            }
            finally
            {
                ClearEcuInstanceData();
            }
        }

        /// <summary>
        /// Clear the fault codes for the Ecu.
        /// </summary>
        public override void ClearFaultCodes()
        {
            try
            {
                Protocol.ClearErrorMemory();
                ActiveFaultCodes.Clear();
            }
            catch (Exception exp)
            {
                throw new EcuException("Failed executing ClearFaultCodes.", exp);
            }
        }

        /// <summary>
        /// Read the active fault codes from the Ecu.
        /// </summary>
        public override void ReadFaultCodes()
        {
            try
            {
                ReadOnlyCollection<Byte[]> response = Protocol.ReadErrorMemory();

                Byte[] data = RemoveHeadersAndCollateData(response);

                DecodeFaultCodeData(data);
            }
            catch (Kwp81Exception exp)
            {
                throw new EcuException("Failed executing ReadFaultCodes.", exp);
            }
        }

        /// <summary>
        /// Reads the snap shot.
        /// </summary>
        /// <param name="snapshotToRead">The snap shot to read.</param>
        /// <returns>The Snapshot evaluation.</returns>
        public override string ReadSnapshot(SnapshotDefinition snapshotToRead)
        {
            if (snapshotToRead == null)
            {
                throw new ArgumentNullException("snapshotToRead");
            }
            try
            {
                ReadOnlyCollection<byte[]> response = Protocol.RequestSnapshot(snapshotToRead.Identifier);
                byte[] snapshot = RemoveHeadersAndCollateData(response);

                return DecodeSnapshot(snapshot, snapshotToRead, Convertor);
            }
            catch (Kwp81Exception exp)
            {
                throw new EcuException("Failed executing ReadSnapShot.", exp);
            }
        }

        /// <summary>
        /// Reads the sensor.
        /// </summary>
        /// <param name="sensor">The sensor.</param>
        /// <returns></returns>
        public override TimeStampedMeasurement ReadSensor(Sensor sensor)
        {
            try
            {
                var s = (IsoSensor) sensor;

                ReadOnlyCollection<byte[]> response = null;

                switch (s.SensorType)
                {
                    case SensorType.Ram:
                        response = Protocol.ReadRamCell(s.Id, (byte) s.ByteLenght);
                        break;
                    case SensorType.Adc:
                        response = Protocol.ReadAdcChannel(s.Id);
                        break;
                    case SensorType.Sys:
                        response = Protocol.RequestSnapshot(s.Id);
                        break;
                    default:
                        break;
                }

                byte[] sensorData = RemoveHeadersAndCollateData(response);

                var result = new TimeStampedMeasurement(0, string.Empty, DateTime.Now, sensor.IsBit);

                var evals = ElementEvaluator.Evaluate(sensorData, Convertor, sensor);

                var e = evals[0];
                if (sensor.IsBit)
                {
                    result.Status = ((MaskMatchResult) e.Value).Result;
                    result.Measurement = ((MaskMatchResult) e.Value).Value;
                }
                else
                {
                    result.Measurement = Convert.ToSingle(e.Value, CultureInfo.InvariantCulture);
                }

                return result;
            }
            catch (Kwp81Exception exp)
            {
                throw new EcuException("Failed executing ReadSensor.", exp);
            }
        }

        /// <summary>
        /// Activates the actuator.
        /// </summary>
        /// <param name="actuator">The actuator.</param>
        /// <returns></returns>
        public override bool ActivateActuator(Actuator actuator)
        {
            EngineRunningCheck();

            var isoActuator = actuator as IsoActuator;

            if(isoActuator.ActivationType == ActuatorActivationType.Ecu)
            {
                Protocol.ActivateActuator(isoActuator.Id);
                Thread.Sleep(isoActuator.Duration * 1000);

                Protocol.ReadErrorMemory();
            }
            else
            {
                byte state = 0;
                
                for (int i = 0; i < isoActuator.Duration; i++)
                {
                    Protocol.StartDownloadRoutine(new Byte[] { state, isoActuator.Id,0,0,0,0,0,0,0,0 });

                    Thread.Sleep(1000);
                    state = (byte) (~state);
                }
            }

            return true;
        }

        /// <summary>
        /// Execute the selected routine.
        /// </summary>
        /// <param name="routine">The Routine to execute.</param>
        /// <returns>The result of the routine execution.</returns>
        public override uint ExecuteRoutine(Routine routine)
        {
            throw new NotImplementedException();
        }

        public override void ExecuteRoutineCancel()
        {
            throw new NotImplementedException();
        }

        internal byte[] RemoveHeadersAndCollateData(IEnumerable<Byte[]> response)
        {
            var collatedDataBytes = new List<Byte>();

            foreach (var bytes in response)
            {
                collatedDataBytes.AddRange(RemoveHeaders(bytes));
            }

            return collatedDataBytes.ToArray();
        }

        private byte[] RemoveHeaders(Byte[] response)
        {
            var bytes = new byte[response.Length - (HeaderEndIndex + CrcBytesCount)];

            Buffer.BlockCopy(response, HeaderEndIndex, bytes, 0, bytes.Length);

            return bytes;
        }

        protected virtual void DecodeIdentificationData(Byte[] identificationData)
        {
            try
            {
                var builder = new StringBuilder();

                builder.Append(((identificationData[0] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append(".");
                builder.Append((identificationData[0] & 0x0F).ToString(CultureInfo.InvariantCulture));
                HardwareVersionNumber = builder.ToString();

                builder.Remove(0, builder.Length);

                builder.Append(((identificationData[1] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append(".");
                builder.Append((identificationData[1] & 0x0F).ToString(CultureInfo.InvariantCulture));
                SoftwareVersionNumber = builder.ToString();

                builder.Remove(0, builder.Length);

                builder.Append(((identificationData[2] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[2] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(" ");
                builder.Append(((identificationData[3] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[3] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(((identificationData[4] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));

                ProductionDate = builder.ToString();
                builder.Remove(0, builder.Length);

                builder.Append((identificationData[4] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(((identificationData[5] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[5] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(((identificationData[6] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[6] & 0x0F).ToString(CultureInfo.InvariantCulture));
                SerialNumber = builder.ToString();

                builder.Remove(0, builder.Length);

                builder.Append(((identificationData[7] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[7] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(((identificationData[8] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[8] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(((identificationData[9] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[9] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(((identificationData[10] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[10] & 0x0F).ToString(CultureInfo.InvariantCulture));
                builder.Append(((identificationData[11] & 0xF0) >> 4).ToString(CultureInfo.InvariantCulture));
                builder.Append((identificationData[11] & 0x0F).ToString(CultureInfo.InvariantCulture));
                DrawingNumber = builder.ToString();

                builder.Remove(0, builder.Length);

                foreach (byte b in diagnosticInterface.KeyBytes)
                {
                    builder.Append(b.ToString("X", CultureInfo.InvariantCulture));
                    builder.Append(" ");
                }

                IsoCode = builder.ToString().Trim();
            }
            catch (Exception exp)
            {
                throw new EcuException("Failed executing DecodeIdData.", exp);
            }
        }

        internal override void ConnectImplemenation()
        {
            try
            {
                Protocol.StartCommunication();

                ReadOnlyCollection<Byte[]> response = Protocol.ReadEcuIdentification();

                Byte[] identificationData = RemoveHeadersAndCollateData(response);

                DecodeIdentificationData(identificationData);

                Protocol.ProtocolDisconnected += HandleProtocolDisconnect;
                Protocol.StartKeepAlive();

                Connected = true;
            }
            catch (Kwp81Exception exp)
            {
                throw new EcuException("Failed executing Connect.", exp);
            }
        }
    }
}