#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Globalization;
using System.Text;
using System.Threading;
using System.Xml.Serialization;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Ecus.Enums;
using DnEcuDiag.Ecus.Exceptions;
using DnEcuDiag.Ecus.Generic;
using DnEcuDiag.Protocols;

namespace DnEcuDiag.Ecus.Kwp
{
    /// <summary>
    /// Implements a programatic model of an ECU.
    /// </summary>
    [Serializable]
    public sealed class KwpEcu : Ecu, IDisposable
    {
        #region Fields

        private Boolean _cancelRoutineExecution;
        private InitType _initType = InitType.Standard;
        [NonSerialized] private Kwp2000 _protocol;
        private Boolean _readFaultsByLocalId;

        #endregion

        #region Constructors

        /// <summary>
        /// Prevent access to the default constructor.
        /// </summary>
        private KwpEcu()
        {
        }

        #endregion

        #region Dispose Pattern

        private bool _disposed;

        /// <summary>
        /// Dispose the underlying resources.
        /// </summary>
        public new void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Close the device.
        /// </summary>
        public new void Close()
        {
            Dispose(true);
        }

        /// <summary>
        /// Dispose of the objects resources.
        /// </summary>
        /// <param name="disposing">True if explicity disposing object else GC finalizing object.</param>
        private void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                base.Dispose();

                if (disposing)
                {
                    if (_protocol != null)
                    {
                        _protocol.Close();
                        _protocol = null;
                    }
                }
            }

            _disposed = true;
        }

        #endregion Dispose Pattern

        #region Private Methods

        private UInt32 ExecuteRoutine(Routine routine, params Byte[] parameters)
        {
            try
            {
                Byte[] result;
                _cancelRoutineExecution = false;

                _protocol.StartRoutineByLocalIdentifier((byte) routine.Id, parameters);

                do
                {
                    if (_cancelRoutineExecution)
                    {
                        _protocol.StopRoutineByLocalIdentifier((byte) routine.Id);

                        throw new RoutineExecutionCanceledException();
                    }

                    result = _protocol.RequestRoutineResultsByLocalIdentifier((byte) routine.Id);

                    Thread.Sleep(100);

                } while (CheckRoutineContinue(result));

#warning This is not a robust check.
                return result[2];
            }
            catch (RoutineExecutionCanceledException)
            {
                throw;
            }
            catch (Kwp2000Exception kwpException)
            {
                if (kwpException.ErrorCode == ResponseCode.SubFunctionNotSupportedInvalidFormat)
                {
                    throw new RoutineNotSupportedException();
                }

                throw new EcuException(
                    String.Format(CultureInfo.InvariantCulture, WinFormStrings.ErrorExecutingRoutineResult,
                                  routine.Name, kwpException.ErrorCode), kwpException);
            }
            catch (Exception exp)
            {
                throw new EcuException(
                    String.Format(CultureInfo.InvariantCulture, WinFormStrings.ErrorExecutingRoutine, routine.Name), exp);
            }
        }

        private static bool CheckRoutineContinue(byte[] routineResult)
        {
            bool checkContinue = false;

            if (routineResult != null)
            {
                if (routineResult.Length != 0)
                {
                    if (routineResult[0] == 0x7F)
                    {
                        int resonseCode = Convert.ToInt32(routineResult[2]);

                        if ((resonseCode == (Int32) ResponseCode.BusyRepeatRequest) |
                            ((resonseCode == (Int32) ResponseCode.RoutineNotComplete)))
                        {
                            checkContinue = true;
                        }
                    }
                }
            }

            return checkContinue;
        }

        /// <summary>
        /// Decodes the DTC.
        /// </summary>
        /// <param name="dtcHigh">The DTC high.</param>
        /// <param name="dtcLow">The DTC low.</param>
        /// <returns>A string representation of the fault code.</returns>
        private static string DecodeDtc(byte dtcHigh, byte dtcLow)
        {
            var bldr = new StringBuilder();

            if ((dtcHigh & 0xC0) == 0)
                bldr.Append("P");
            if ((dtcHigh & 0xC0) == 0x40)
                bldr.Append("C");
            if ((dtcHigh & 0xC0) == 0x80)
                bldr.Append("B");
            if ((dtcHigh & 0xC0) == 0xC0)
                bldr.Append("U");

            int i = (dtcHigh & 0x30);

            i = i >> 4;
            bldr.Append(i.ToString(CultureInfo.InvariantCulture));
            i = (dtcHigh & 0x0F);
            bldr.Append(i.ToString(CultureInfo.InvariantCulture));
            i = (dtcLow & 0xF0);
            i = i >> 4;
            bldr.Append(i.ToString(CultureInfo.InvariantCulture));
            i = (dtcLow & 0x0F);
            bldr.Append(i.ToString(CultureInfo.InvariantCulture));

            return bldr.ToString();
        }

        private void DecodeIdentificationData(byte[] identificationData)
        {
            try
            {
                var builder = new StringBuilder();
                Int32 offset = 11;

                if (identificationData.Length == 59)
                {
                    offset = 10;
                }


                //ASCII
                int count = 11;
                int i;
                for (i = 0; i < count; i++)
                {
                    builder.Append(Convert.ToChar(identificationData[i]));
                }

                DrawingNumber = builder.ToString();
                builder.Remove(0, builder.Length);

                //ASCII
                count = offset + i;
                for (; i < count; i++)
                {
                    builder.Append(Convert.ToChar(identificationData[i]));
                }

                HardwareNumber = builder.ToString();
                builder.Remove(0, builder.Length);

                //hex
                HardwareVersionNumber = string.Format(CultureInfo.InvariantCulture, "0x{0,00:X}", identificationData[i]);
                i++;

                count = offset + i;
                //ASCII
                for (; i < count; i++)
                {
                    builder.Append(Convert.ToChar(identificationData[i]));
                }

                SoftwareNumber = builder.ToString();
                builder.Remove(0, builder.Length);

                //Hex
                builder.Append(identificationData[i]);
                i++;
                builder.Append(identificationData[i]);
                i++;
                SoftwareVersionNumber = string.Format(CultureInfo.InvariantCulture, "0x{0,00:X}", builder);
                builder.Remove(0, builder.Length);

                //ASCII
                count = 6 + i;
                for (; i < count; i++)
                {
                    builder.Append(Convert.ToChar(identificationData[i]));
                }

                OmologationNumber = builder.ToString();
                builder.Remove(0, builder.Length);

                count = 5 + i;
                for (; i < count; i++)
                {
                    builder.Append(identificationData[i].ToString("X", CultureInfo.InvariantCulture));
                    builder.Append(" ");
                }

                IsoCode = builder.ToString();
                builder.Remove(0, builder.Length);

                count = 10 + i;
                for (; i < count; i++)
                {
                    builder.Append(Convert.ToChar(identificationData[i]));
                }

                builder.Remove(0, builder.Length);

                int val;
                count = 4 + i;
                for (; i < count; i++)
                {
                    val = identificationData[i] & 0xF0;
                    val = val >> 4;
                    builder.Append(val.ToString(CultureInfo.InvariantCulture));

                    val = identificationData[i] & 0x0F;
                    builder.Append(val.ToString(CultureInfo.InvariantCulture));

                    builder.Append(" ");
                }

                ProductionDate = ProgrammingDate = builder.ToString();
                builder.Remove(0, builder.Length);
            }
            catch (Exception exp)
            {
                throw new EcuException("Failed executing DecodeIdData.", exp);
            }
        }

        /// <summary>
        /// Return control to the Ecu.
        /// </summary>
        private void ReturnControlToEcu(InputOutputLocalIdentifier iolocalIdentifier)
        {
            if (_protocol == null) throw new EcuException("The Interface has not been configured.");

            _protocol.InputOutputControlByLocalIdentifier(iolocalIdentifier.Id,
                                                         Kwp2000.InputOutputControlParameter.ReturnControlToEcu);
        }

        /// <summary>
        /// Reset the InputOutputLocalIdentifier to its default.
        /// </summary>
        private void ResetToDefault(InputOutputLocalIdentifier iolocalIdentifier)
        {
            if (_protocol == null) throw new EcuException("The Interface has not been configured.");

            _protocol.InputOutputControlByLocalIdentifier(iolocalIdentifier.Id,
                                                         Kwp2000.InputOutputControlParameter.ResetToDefault);
        }

        /// <summary>
        /// Adjust the output/input state of the for the short term i.e. until key of etc.
        /// </summary>
        /// <returns></returns>
        private byte[] ShortTermAdjust(InputOutputLocalIdentifier iolocalIdentifier)
        {
            if (_protocol == null) throw new EcuException("The Interface has not been configured.");

            return _protocol.InputOutputControlByLocalIdentifier(iolocalIdentifier.Id,
                                                              Kwp2000.InputOutputControlParameter.ShortTermAdjustment);
        }

        /// <summary>
        /// Adjust the output/input state of the for the short term i.e. until key of etc.
        /// </summary>
        /// <param name="iolocalIdentifier">The iolocal identifier.</param>
        /// <param name="adjustValue">The value to adjust the output too.</param>
        /// <returns></returns>
        private byte[] ShortTermAdjust(InputOutputLocalIdentifier iolocalIdentifier, byte[] adjustValue)
        {
            if (_protocol == null) throw new EcuException("The Interface has not been configured.");

            return
                (_protocol.InputOutputControlByLocalIdentifier(iolocalIdentifier.Id,
                                                              Kwp2000.InputOutputControlParameter.ShortTermAdjustment,
                                                              adjustValue));
        }

        #endregion Private Methods

        #region Properties

        /// <summary>
        /// Gets a value indicating whether to read faults by local id.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if [read faults by local id]; otherwise, <c>false</c>.
        /// </value>
        public bool ReadFaultsByLocalId
        {
            get { return _readFaultsByLocalId; }
            set { _readFaultsByLocalId = value; }
        }

        /// <summary>
        /// The type of initalisation for this ecu.
        /// </summary>
        public InitType InitType
        {
            get { return _initType; }
            set { _initType = value; }
        }

        #endregion Properties

        [XmlIgnore]
        public override IDevice DiagnosticInterface
        {
            get { return diagnosticInterface; }
            set
            {
                diagnosticInterface = value;
                _protocol = new Kwp2000(value, (byte) EcuAddress, 0xF1);
            }
        }

        /// <summary>
        /// Disconnect from the ECU.
        /// </summary>
        public override void Disconnect()
        {
            if (Connected == false) return;
            if (_protocol == null) throw new EcuException("The Interface has not been configured.");

            try
            {
                Trace.WriteLine("Disconnecting from ECU.");

                Connected = false;

                StopSensorMonitoring();

                _protocol.StopDiagnosticSession();

                _protocol.StopCommunication();
            }
            catch (Kwp2000Exception)
            {
                // DG 1.0.0.4 swallow the exception.
                //throw new EcuException("Failed executing Disconnect.",exp);
            }
            finally
            {
                ClearEcuInstanceData();
            }
        }

        /// <summary>
        /// Disconnect from the ecu.
        /// </summary>
        public override void ClearFaultCodes()
        {
            if (_protocol == null) throw new EcuException("The Interface has not been configured.");

            try
            {
                _protocol.ClearDiagnosticInformation();
                ActiveFaultCodes.Clear();
            }
            catch (Kwp2000Exception exp)
            {
                throw new EcuException("Failed to clear fault codes.", exp);
            }
        }

        /// <summary>
        /// Read the list of active fault codes.
        /// </summary>
        public override void ReadFaultCodes()
        {
            if (_protocol == null) throw new EcuException("The Interface has not been configured.");

            try
            {
                byte[] resp = _protocol.ReadDiagnosticTroubleCodesByStatus(0, 0xFF, 0x00);

                int numberFaults = resp[1];
                int offset = 2;

                ActiveFaultCodes.Clear();

                for (int i = 0; i < numberFaults; i++)
                {
                    byte dtcHigh = resp[offset];
                    byte dtcLow = resp[++offset];

                    byte[] faultDataPacket = _readFaultsByLocalId ? _protocol.ReadDataByLocalIdentifier(0x10, dtcHigh, dtcLow) : _protocol.ReadStatusOfDiagnosticTroubleCodes(dtcHigh, dtcLow);

                    if (faultDataPacket[1] == 0)
                    {
                        continue;
                    }

                    string dtcCode = DecodeDtc(dtcHigh, dtcLow);

                    var faultData = new byte[ErrorMemoryStructureDefinition.Lenght];

                    Buffer.BlockCopy(faultDataPacket, 2, faultData, 0, ErrorMemoryStructureDefinition.Lenght);

                    ActiveFaultCodes.Add(new FaultCode(dtcCode, DecodeFaultCode(dtcCode, faultData, Convertor)));

                    offset += 2;
                }
            }
            catch (Kwp2000Exception exp)
            {
                throw new EcuException("Failed executing ReadFaultCodes.", exp);
            }
        }

        /// <summary>
        /// Reads the snap shot.
        /// </summary>
        /// <param name="snapshotToRead">The snap shot to read.</param>
        /// <returns>The SnapShot evaluation.</returns>
        public override String ReadSnapshot(SnapshotDefinition snapshotToRead)
        {
            try
            {
                byte[] rawData = _protocol.ReadDataByLocalIdentifier(snapshotToRead.Identifier);
                var data = new byte[rawData.Length - 2];
                Buffer.BlockCopy(rawData, 2, data, 0, rawData.Length - 2);

                return DecodeSnapshot(data, snapshotToRead, Convertor);
            }
            catch (Kwp2000Exception exp)
            {
                if (exp.ErrorCode == ResponseCode.SubFunctionNotSupportedInvalidFormat)
                {
                    throw new SnapshotNotSupportedException("This snapshot is not supported by this ecu.", exp);
                }

                throw new EcuException("Failed executing Read.", exp);
            }
            catch (Exception exp)
            {
                throw new EcuException("Failed executing Read.", exp);
            }
        }

        /// <summary>
        /// Reads the sensor.
        /// </summary>
        /// <param name="sensor">The sensor.</param>
        /// <returns></returns>
        public override TimeStampedMeasurement ReadSensor(Sensor sensor)
        {
            try
            {

                TimeStampedMeasurement result = null;
                var localSensor = sensor as LocalIdentifier;

                if (localSensor != null)
                {
                    localSensor.RawData = _protocol.ReadDataByLocalIdentifier((Byte) localSensor.Id);
                    var data = new byte[localSensor.RawData.Length - 2];
                    Buffer.BlockCopy(localSensor.RawData, 2, data, 0, localSensor.RawData.Length - 2);

                    result =
                        new TimeStampedMeasurement(0, string.Empty, DateTime.Now, localSensor.IsBit);
                    if (localSensor.IsBit)
                    {
                        float val;
                        result.Status = ProcessMaskMatches(localSensor.MaskMatches, localSensor.Bitmask, localSensor.BitShift, data,
                                                  out val);
                        result.Measurement = val;
                    }
                    else
                    {
                        result.Measurement = Convertor.Convert(localSensor.ConversionId, data);
                    }
                }

                return result;
            }
            catch (Kwp2000Exception exp)
            {
                if (exp.ErrorCode == ResponseCode.SubFunctionNotSupportedInvalidFormat)
                {
                    throw new SensorNotSupportedException("This sensor is not supported by this ecu.", exp);
                }
                
                throw new EcuException("Failed executing Read.", exp);
            }
            catch (Exception exp)
            {
                throw new EcuException("Failed executing Read.", exp);
            }
        }

        public override bool ActivateActuator(Actuator actuator)
        {
            var ioli = actuator as InputOutputLocalIdentifier;

            try
            {
                EngineRunningCheck();

                var array = new byte[] {0xFF};
                int complement = array[0];

                switch (ioli.ActivationType)
                {
                    case ActuatorActivationType.Manual:
                        for (int i = 0; i < ioli.Duration * 2; i++)
                        {
                            ShortTermAdjust(ioli, array);

                            Thread.Sleep(500);

                            complement = ~complement;
                            array[0] = (byte)complement;
                        }
                        
                        break;
                    case ActuatorActivationType.Ecu:
                        
                        if (ioli.UseControlParameters)
                        {
                            array[0] = ioli.ControlParameter;
                            ShortTermAdjust(ioli, array);
                        }
                        else
                        {
                            ShortTermAdjust(ioli);
                        }

                        // Just spin.
                        for (int i = 0; i < ioli.Duration * 10; i++)
                        {
                            Thread.Sleep(100);
                        }

                        break;
                    case ActuatorActivationType.Reset:
                        ResetToDefault(ioli);
                        break;
                    case ActuatorActivationType.Routine:

                        for (int i = 0; i < ioli.Duration * 2; i++)
                        {
                            _protocol.StartRoutineByLocalIdentifier(ioli.ControlParameter, new[] { array[0], ioli.Id});
                            
                            Thread.Sleep(500);

                            complement = ~complement;
                            array[0] = (byte)complement;
                        }

                        break;
                    case ActuatorActivationType.RoutineCalibration:
                        break;
                    default:
                        break;
                }

                // return control to the ECU.
                if (ioli.ReturnEcuControl)
                {
                    ReturnControlToEcu(ioli);
                }

                return true;
            }
            catch (EngineRunningException)
            {
                throw;
            }
            catch (Kwp2000Exception exp)
            {
                throw new ActuatorNotSupportedException(
                    String.Format(CultureInfo.InvariantCulture, WinFormStrings.OperationOfActuatorFailedForTheFollowingReason, exp.ErrorCode), exp);
            }
            catch (Exception exp)
            {
                throw new EcuException(
                    String.Format(CultureInfo.InvariantCulture, WinFormStrings.ErrorActivatingActuator, ioli), exp);
            }
        }

        /// <summary>
        /// Execute the selected routine.
        /// </summary>
        /// <param name="routine">The Routine to execute.</param>
        /// <returns>The result of the routine execution.</returns>
        public override UInt32 ExecuteRoutine(Routine routine)
        {
            return ExecuteRoutine(routine);
        }


        public override void ExecuteRoutineCancel()
        {
            _cancelRoutineExecution = true;
        }

        internal override void ConnectImplemenation()
        {
            try
            {
                _protocol.StartCommunication(_initType);

                InitaliseTimingParameters();

                StartDiagnosticSession();
                
                GetIdentificationData();

                _protocol.ProtocolDisconnected += HandleProtocolDisconnect;
                _protocol.StartKeepAlive();

                Connected = true;
            }
            catch (Kwp2000Exception exp)
            {
                throw new EcuException("Failed executing Connect.", exp);
            }
        }

        private void GetIdentificationData()
        {
            byte[] identificationData;
            try
            {
                identificationData = _protocol.ReadDataByLocalIdentifier(0x80);
            }
            catch (Kwp2000Exception)
            {
                identificationData = _protocol.ReadEcuIdentification(0x80);
            }

            var data = new byte[identificationData.Length - 2];
            Buffer.BlockCopy(identificationData, 2, data, 0, identificationData.Length - 2);
            DecodeIdentificationData(data);
        }

        /// <summary>
        /// Start diagnostic session.
        /// </summary>
        private void StartDiagnosticSession()
        {
            try
            {
                _protocol.StartDiagnosticSession(0x81);
            }
            catch (Kwp2000Exception e)
            {
                // Exclude to cases where diagnostic cannot be started.
                switch (e.ErrorCode)
                {
                    case ResponseCode.ServiceNotSupported:
                        break;
                    case ResponseCode.SubFunctionNotSupportedInvalidFormat:
                        break;
                    default:
                        throw;
                }
            }
        }

        /// <summary>
        /// Initalise the timimg parameters.
        /// </summary>
        private void InitaliseTimingParameters()
        {
            byte p2Min = 0;
            byte p2Max = 0;
            byte p3Min = 0;
            byte p3Max = 0;
            byte p4Min = 0;

            try
            {
                _protocol.AccessTimingParameter(Kwp2000.CommsParamIdentifier.ReadPossibleLimits, ref p2Min, ref p2Max,
                                               ref p3Min, ref p3Max, ref p4Min);

                _protocol.AccessTimingParameter(Kwp2000.CommsParamIdentifier.SetValues, ref p2Min, ref p2Max, ref p3Min,
                                               ref p3Max, ref p4Min);
            }
            catch (Kwp2000Exception exp)
            {
                Trace.WriteLine(String.Format("Exception: {0}", exp));
            }
        }

        /// <summary>
        /// Overloaded conversion function that converts bit mapped 
        /// RLI to the corresponding values.
        /// </summary>
        /// <param name="masks"></param>
        /// <param name="mask">The mask to apply to the value.</param>
        /// <param name="bitShift">The number of bits to shift the bits by to normalse the result.</param>
        /// <param name="rawData"></param>
        /// <param name="result"></param>
        /// <returns></returns>
        public static string ProcessMaskMatches(Collection<MaskMatch> masks, uint mask, int bitShift, byte[] rawData, out float result)
        {
            try
            {
                string matchedResult = String.Empty;
                uint rawdata = Conversion.Convertor.ConvertByteArray(rawData);

                // and with mask.
                uint val = rawdata;
                val &= mask;

                //Bitshift the value to normalise.
                if (bitShift != 0)
                {
                    val = val >> bitShift;
                }

                // find match
                foreach (MaskMatch match in masks)
                {
                    if (match.MatchValue)
                    {
                        if (match.MaskValue == val)
                        {
                            matchedResult = match.MaskResultTrue;
                        }
                    }
                    else
                    {
                        matchedResult = (val & match.MaskValue) >= 1 ? match.MaskResultTrue : match.MaskResultFalse;
                    }
                }

                result = val;
                return matchedResult;
            }
            catch (Exception exp)
            {
                throw new ConversionException("Failed executing Conversion.", exp);
            }
        }
    }
}