#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.IO.Compression;
using System.Reflection;
using System.Text;
using System.Xml.Serialization;
using DnEcuDiag.Conversion;
using DnEcuDiag.Ecus.Enums;
using DnEcuDiag.Ecus.Generic;
using DnEcuDiag.Ecus.Iso;
using DnEcuDiag.Ecus.Kwp;

namespace DnEcuDiag.Factories
{
    /// <summary>
    /// Factory to create an instance of an IEcu.
    /// </summary>
    public static class EcuFactory
    {
        private static Ecu _lastEcu;
        private static Int32 _lastIndex;
        private static readonly EcuDescriptors _ecuDescriptors;

        static EcuFactory()
        {
            _ecuDescriptors = DeserializeEcuDescriptors(EcuDefinitions.ResourceManager.GetString("EcuDescriptors"));
        }

        /// <summary>
        /// The list of Ecu descriptors.
        /// </summary>
        public static List<EcuDescriptor> EcuList
        {
            get { return _ecuDescriptors.Descriptors; }
        }

        #region Private Methods

        private static EcuDescriptors DeserializeEcuDescriptors(String descriptors)
        {
            var serializer = new XmlSerializer(typeof(EcuDescriptors));
            var rdr = new StringReader(descriptors);

            return serializer.Deserialize(rdr) as EcuDescriptors;
        }

        private static KwpEcu DeserializeKwpEcu(String ecuDefinition)
        {
            var serializer = new XmlSerializer(typeof(KwpEcu));
            var rdr = new StringReader(ecuDefinition);

            return serializer.Deserialize(rdr) as KwpEcu;
        }

        private static Kwp81Ecu DeserializeKwp81Ecu(String ecuDefinition)
        {
            var serializer = new XmlSerializer(typeof(Kwp81Ecu));
            var rdr = new StringReader(ecuDefinition);

            return serializer.Deserialize(rdr) as Kwp81Ecu;
        }

        private static Kwp71Ecu DeserializeKwp71Ecu(String ecuDefinition)
        {
            var serializer = new XmlSerializer(typeof(Kwp71Ecu));
            var rdr = new StringReader(ecuDefinition);

            return serializer.Deserialize(rdr) as Kwp71Ecu;
        }

        private static Ecu LoadEcu(int ecuIndex)
        {
            Ecu newEcu = null;
            String ecuDefiniton =
                DecompressBytes(EcuDefinitions.ResourceManager.GetObject((ecuIndex + 1).ToString(CultureInfo.InvariantCulture)) as byte[]);

            if (_ecuDescriptors.Descriptors[ecuIndex].EcuType == EcuType.KwpEcu)
            {
                newEcu = DeserializeKwpEcu(ecuDefiniton);
            }

            if (_ecuDescriptors.Descriptors[ecuIndex].EcuType == EcuType.Kwp81Ecu)
            {
                newEcu = DeserializeKwp81Ecu(ecuDefiniton);
            }

            if (_ecuDescriptors.Descriptors[ecuIndex].EcuType == EcuType.Kwp71Ecu)
            {
                newEcu = DeserializeKwp71Ecu(ecuDefiniton);
            }

            if (newEcu != null)
            {
                newEcu.Convertor = CreateConversionClass(newEcu);
            }

            return newEcu;
        }

        private static IConvertor CreateConversionClass(Ecu ecu)
        {
            var t = Assembly.GetExecutingAssembly().GetType(ecu.ConvertorType);

            return Activator.CreateInstance(t) as IConvertor;
        }

        private static String DecompressBytes(Byte[] bytes)
        {
            List<Byte> decompressedBytes;

            using (var ms = new MemoryStream(bytes))
            {
                using (var compressedzipStream = new DeflateStream(ms, CompressionMode.Decompress))
                {
                    decompressedBytes = ReadAllBytesFromStream(compressedzipStream);
                }
            }

            return ConvertByteArrayToString(decompressedBytes.ToArray());
        }

        private static List<Byte> ReadAllBytesFromStream(Stream stream)
        {
            var buffer = new Byte[10000];
            var bytes = new List<byte>();
            Int32 readCount;

            do
            {
                readCount = stream.Read(buffer, 0, buffer.Length);

                for (int i = 0; i < readCount; i++)
                {
                    bytes.Add(buffer[i]);
                }
            } while (readCount != 0);

            return bytes;
        }

        private static string ConvertByteArrayToString(Byte[] bytesToConvert)
        {
            Encoding e = Encoding.UTF8;

            return e.GetString(bytesToConvert);
        }

        #endregion

        public static Ecu CreateEcu(EcuDescriptor ecuDescriptor)
        {
            var ecuIndex = _ecuDescriptors.Descriptors.IndexOf(ecuDescriptor);

            if ((_lastIndex != ecuIndex) | (_lastEcu == null))
            {
                _lastEcu = LoadEcu(ecuIndex);
                _lastIndex = ecuIndex;
            }

            return _lastEcu;
        }
    }
}