#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Globalization;
using DnEcuDiag.Configuration;
using DnEcuDiag.Devices;
using DnEcuDiag.Devices.Exceptions;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Ecus.Enums;
using DnEcuDiag.Ecus.Generic;
using CommStudio.Connections;
using SerialPort;
using SerialPort.Api.Enums;
using SerialPort.Exceptions;
using SerialPort.Settings;
using Parity=SerialPort.Api.Enums.Parity;

namespace DnEcuDiag.Factories
{
    /// <summary>
    /// Factory to create an instance of an IDevice.
    /// </summary>
    internal static class InterfaceDeviceFactory
    {
        /// <summary>
        /// Create an instance of an interface device.
        /// </summary>
        /// <param name="ecuDescriptor">The ecudescriptor which contains ecu type information.</param>
        /// <param name="deviceConfiguration">The device configuration to use to create the device.</param>
        /// <returns>An Interface device.</returns>
        public static IDevice CreateInterfaceDevice(EcuDescriptor ecuDescriptor, DeviceConfiguration deviceConfiguration)
        {
            if (ecuDescriptor == null)
            {
                throw new ArgumentNullException("ecuDescriptor");
            }

            IDevice newDevice = null;

            IPort port = null;
            
            try
            {
                port = CreatePort(deviceConfiguration);

                if (ecuDescriptor.EcuType == EcuType.KwpEcu)
                {
                    newDevice = new Kwp2000SerialDevice(port, true);
                    SetupDeviceTiming(newDevice, deviceConfiguration.KwpTiming);
                }

                if (ecuDescriptor.EcuType == EcuType.Kwp81Ecu)
                {
                    newDevice = new Kwp81SerialDevice(port, true);
                    SetupDeviceTiming(newDevice, deviceConfiguration.IsoTiming);
                }

                if (ecuDescriptor.EcuType == EcuType.Kwp71Ecu)
                {
                    newDevice = new Kwp71SerialDevice(port, true);
                    SetupDeviceTiming(newDevice, deviceConfiguration.IsoTiming);
                }
            }
            catch
            {
                if(port != null)
                {
                    port.Close();   
                }
                throw;
            }


#if DEBUG
            newDevice.EnableTrace = true;
#endif
            return newDevice;
        }

        public static IPort CreatePort(DeviceConfiguration deviceConfiguration)
        {
            IPort port;
            UInt32 bufferSize = 128;
            Byte latencyTimer = 2;

            if (deviceConfiguration.UseUsb)
            {
                if (deviceConfiguration.UseAlternateUsbTimings)
                {
                    bufferSize = bufferSize/2;
                    latencyTimer = 1;
                }

                var busSettings = new UsbBusSettings(bufferSize, bufferSize, latencyTimer);

                var ftd2XxSettings =
                    new Ftd2XxSettings(Parity.None, 8, 200, StopBit.One, Ftd2XxFlowControl.FtFlowNone,
                                       0, 0, busSettings);

                port = new Ftd2XxPort("FTDI Port 0", ftd2XxSettings);
            }
            else
            {
                port = new SaxPort(deviceConfiguration.SerialPort.PortName, 200, CommStudio.Connections.Parity.None, 8, CommStopBits.One);
            }

            try
            {
                port.Open();
            }
            catch (ComportException e)
            {
                throw new SerialDeviceException(String.Format(CultureInfo.InvariantCulture, WinFormStrings.UnableToOpenSerialPort, deviceConfiguration.SerialPort), e);
            }
            catch (UsbPortException e)
            {
                throw new SerialDeviceException(WinFormStrings.UnableToOpenUSBDevice, e);
            }

            return port;
        }

        public static void SetupDeviceTiming(IDevice device, TimingConfig timingConfig)
        {
            device.EcuResponseInterbyteTimeout = timingConfig.EcuResponseInterbyteTimeout;
            device.EcuResponseTimeout = timingConfig.EcuResponseTimeout;
        }
    }
}