#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Drawing;
using System.Drawing.Imaging;

namespace DnEcuDiag
{
    ///	<summary>
    ///	A class	used to	apply imagine transformations to an	image.
    ///	</summary>
    internal struct ImageProcess
    {
        private Single _bleach;
        private Single _translucence;

        ///	<summary>
        ///	The	image to apply the processing too.
        ///	</summary>
        public Image ImageToProcess { get; set; }

        ///	<summary>
        ///	 The amout of drop shadow.
        ///	</summary>
        public byte DropShadow { get; set; }

        ///	<summary>
        ///	The	amount of translucence for an image.
        ///	</summary>
        public Single Translucence
        {
            get { return _translucence; }
            set
            {
                if (value > 0)
                {
                    _translucence = value;

                    while (_translucence > 1)
                        _translucence /= 10;
                }
                else
                    _translucence = 1;
            }
        }

        ///	<summary>
        ///	The	amount of bleaching	of an image.
        ///	</summary>
        public Single Bleach
        {
            get { return _bleach; }
            set
            {
                if (value > 0)
                {
                    _bleach = value;

                    while (_bleach > 1)
                    {
                        _bleach /= 10;
                    }
                }
                else
                    _bleach = 1;
            }
        }
    }

    ///	<summary>
    ///	Allows icon	processing to be performed on an icon.
    ///	</summary>
    internal sealed class ImageProcessor
    {
        #region	Singleton Pattern

        // Singleton.
        private static ImageProcessor _instance;
        private static readonly object LockObject = new object();

        ///	<summary>
        ///	Prevent	instantiation from outside of class.
        ///	</summary>
        private ImageProcessor()
        {
        }

        ///	<summary>
        ///	Singleton Pattern.
        ///	</summary>
        ///	<returns>Singleton.</returns>
        public static ImageProcessor Instance
        {
            get
            {
                lock (LockObject)
                {
                    if (_instance == null)
                    {
                        _instance = new ImageProcessor();
                    }
                }
                return _instance;
            }
        }

        #endregion

        #region	Overlay

        ///	<summary>
        ///	Allow an icon to be	overlayed by multiple icons.
        ///	</summary>
        ///	<param name="targetImage"></param>
        ///	<param name="overlayIcons"></param>
        ///	<returns>An	icon with the overlays applied.</returns>
        public static Image Overlay(Image targetImage, ImageProcess[] overlayIcons)
        {
            if (targetImage == null)
            {
                throw new ArgumentNullException("targetImage");
            }

            if (overlayIcons == null)
            {
                throw new ArgumentNullException("overlayIcons");
            }

            var rect = new Rectangle(0, 0, targetImage.Width, targetImage.Height);
            var bmp = new Bitmap(targetImage.Width, targetImage.Height);

            using (Graphics g = Graphics.FromImage(bmp))
            {
                g.DrawImage(targetImage, rect);

                foreach (ImageProcess ip in overlayIcons)
                {
                    bool processed = false;
                    if (ip.Translucence > 0)
                    {
                        g.DrawImage(Translucent(ip, ip.Translucence), rect);
                        processed = true;
                    }

                    if (ip.Bleach > 0 & ip.Bleach < 1)
                    {
                        g.DrawImage(Bleach(ip, ip.Bleach), rect);
                        processed = true;
                    }
                    if (ip.DropShadow > 0)
                    {
                        g.DrawImage(DropShadow(ip, ip.DropShadow), rect);
                        processed = true;
                    }
                    if (!processed)
                    {
                        //no operation was performed on	the	icon so	use	default
                        g.DrawImage(ip.ImageToProcess, rect);
                    }
                }
            }

            return bmp;
        }

        #endregion Overlay

        #region	Translucent

        ///	<summary>
        ///	Make an	image translucent using	translucence value of 0.5.
        ///	</summary>
        ///	<param name="imageToProcess">The image to apply	translucence too.</param>
        ///	<returns>The image processed</returns>
        public static Image Translucent(ImageProcess imageToProcess)
        {
            return Translucent(imageToProcess, 0.5f);
        }

        ///	<summary>
        ///	Make an	image translucent.
        ///	</summary>
        ///	<param name="imageToProcess">The image to apply	translucence too.</param>
        ///	<param name="amount">The amout of translucence 0 = opaque 1	= solid.</param>
        ///	<returns>The image processed</returns>
        public static Image Translucent(ImageProcess imageToProcess, Single amount)
        {
            var ia = new ImageAttributes();
            var cm = new ColorMatrix();
            var bmp = new Bitmap(imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height);
            if (amount > 1)
            {
                while (amount < 1)
                {
                    amount /= 10;
                }
            }

            Graphics g = Graphics.FromImage(bmp);
            cm.Matrix00 = 1;
            cm.Matrix11 = 1;
            cm.Matrix22 = 1;
            cm.Matrix33 = amount;
            cm.Matrix44 = 1;

            ia.SetColorMatrix(cm);
            g.DrawImage(imageToProcess.ImageToProcess,
                        new Rectangle(0, 0, imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height),
                        0, 0, imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height,
                        GraphicsUnit.Pixel, ia);
            ia.Dispose();

            g.Dispose();

            return bmp;
        }

        #endregion Translucent

        #region	Bleach

        ///	<summary>
        ///	Bleaches the specified image to	process.
        ///	</summary>
        ///	<param name="imageToProcess">The image to process.</param>
        ///	<returns>The procesed immage.</returns>
        public static Image Bleach(ImageProcess imageToProcess)
        {
            return Bleach(imageToProcess, 0.5f);
        }

        ///	<summary>
        ///	Bleaches the specified image to	process.
        ///	</summary>
        ///	<param name="imageToProcess">The image to process.</param>
        ///	<param name="amount">The amount	to bleach the image.</param>
        ///	<returns>The procesed immage.</returns>
        public static Image Bleach(ImageProcess imageToProcess, Single amount)
        {
            var ia = new ImageAttributes();
            var cm = new ColorMatrix();
            var bmp = new Bitmap(imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height);

            if (amount > 1)
            {
                while (amount > 1)
                {
                    amount /= 10;
                }
            }

            Graphics g = Graphics.FromImage(bmp);
            cm.Matrix00 = 1;
            cm.Matrix11 = 1;
            cm.Matrix22 = 1;
            cm.Matrix33 = amount;

            ia.SetColorMatrix(cm);
            g.DrawImage(imageToProcess.ImageToProcess,
                        new Rectangle(0, 0, imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height),
                        0, 0, imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height,
                        GraphicsUnit.Pixel, ia);
            ia.Dispose();
            g.Dispose();

            return bmp;
        }

        #endregion Bleach

        #region	DropShadow

        ///	<summary>
        ///	Drops a	shadow on an image.
        ///	</summary>
        ///	<param name="imageToProcess">The image to process.</param>
        ///	<returns>The procesed immage.</returns>
        public static Image DropShadow(ImageProcess imageToProcess)
        {
            return DropShadow(imageToProcess, 2, 11);
        }

        ///	<summary>
        ///	Drops a	shadow on an image.
        ///	</summary>
        ///	<param name="imageToProcess">The image to process.</param>
        ///	<param name="offset">The offset	of the shadow.</param>
        ///	<returns>The procesed immage.</returns>
        public static Image DropShadow(ImageProcess imageToProcess, byte offset)
        {
            return DropShadow(imageToProcess, offset, 11);
        }

        ///	<summary>
        ///	Drops a	shadow on an image.
        ///	</summary>
        ///	<param name="imageToProcess">The image to process.</param>
        ///	<param name="offset">The offset	of the shadow.</param>
        ///	<param name="shadow">The shadow.</param>
        ///	<returns>The procesed immage.</returns>
        public static Image DropShadow(ImageProcess imageToProcess, byte offset, Single shadow)
        {
            var bmp = new Bitmap(imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height);

            using (Graphics g = Graphics.FromImage(bmp))
            {
                g.DrawIcon(GreyScale(imageToProcess, shadow),
                           new Rectangle(offset, offset, imageToProcess.ImageToProcess.Width,
                                         imageToProcess.ImageToProcess.Height));
                g.DrawImage(imageToProcess.ImageToProcess,
                            new Rectangle(0, 0, imageToProcess.ImageToProcess.Width,
                                          imageToProcess.ImageToProcess.Height), 0, 0,
                            imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height,
                            GraphicsUnit.Pixel);
            }
            
            return bmp;
        }

        #endregion DropShadow

        #region	GreyScale

        ///	<summary>
        ///	Greyscale an image.
        ///	</summary>
        ///	<param name="imageToProcess">The image to process.</param>
        ///	<returns></returns>
        public static Icon GreyScale(ImageProcess imageToProcess)
        {
            return GreyScale(imageToProcess, 0.5f);
        }

        ///	<summary>
        ///	Greyscale an image.
        ///	</summary>
        ///	<param name="imageToProcess">The image to process.</param>
        ///	<param name="amount">The amount.</param>
        ///	<returns></returns>
        public static Icon GreyScale(ImageProcess imageToProcess, Single amount)
        {
            var ia = new ImageAttributes();
            var cm = new ColorMatrix();

            if (amount > 1)
            {
                while (amount > 1)
                {
                    amount /= 10;
                }
            }

            var bmp = new Bitmap(imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height);
            Graphics g = Graphics.FromImage(bmp);
            cm.Matrix00 = amount;
            cm.Matrix01 = amount;
            cm.Matrix02 = amount;
            cm.Matrix10 = amount;
            cm.Matrix11 = amount;
            cm.Matrix12 = amount;
            cm.Matrix20 = amount;
            cm.Matrix21 = amount;
            cm.Matrix22 = amount;
            ia.SetColorMatrix(cm);
            g.DrawImage(imageToProcess.ImageToProcess,
                        new Rectangle(0, 0, imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height),
                        0, 0, imageToProcess.ImageToProcess.Width, imageToProcess.ImageToProcess.Height,
                        GraphicsUnit.Pixel, ia);
            Icon returnIcon = Icon.FromHandle(bmp.GetHicon());
            ia.Dispose();
            g.Dispose();

            return returnIcon;
        }

        #endregion GreyScale
    }
}