#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using DnEcuDiag.Ecus.Generic;

namespace DnEcuDiag.Logging
{

    #region	DataLogger

    ///	<summary>
    ///	A class	used to	log	data to	a log file.
    ///	</summary>
    internal class DataLogger
    {
        #region	Enums

        ///	<summary>
        ///	The	timestamp format.
        ///	</summary>
        public enum TimestampFormat
        {
            ///	<summary>
            ///	MillisecondFormat is the difference	between	the
            ///	first timestamp	and	the	current	measurement
            ///	timestamp expressed	in the format
            ///	milliseconds.microseconds i.e 34.6677
            ///	</summary>
            MillisecondFormat,
            ///	<summary>
            ///	MillisecondFormat is the difference	between	the
            ///	first timestamp	and	the	current	measurement
            ///	timestamp expressed	in the format
            ///	dd:hh:mm:ss:fff.
            ///	</summary>
            ClockFormat
        }

        #endregion Enums

        #region	Constants

        private const string LogfileExt = ".csv";
        private const string LogfileName = "SensorLogs";
        private const string TimestampHdr = "Timestamp";

        #endregion

        #region	Fields

        private readonly Hashtable m_fileTable;
        private readonly NumberFormatInfo numberFormat;
        private string logPath;
        private bool m_logdata;
        private bool m_singlelogfile;
        private DateTime m_startTime;
        private TimestampFormat m_timeMarkerFormat = TimestampFormat.ClockFormat;

        #endregion

        #region	Constructor

        ///	<summary>
        ///	Create an instance of the datalogger.
        ///	</summary>
        public DataLogger()
        {
            m_fileTable = new Hashtable();
            numberFormat = new NumberFormatInfo();
        }

        #endregion

        #region	Properties

        ///	<summary>
        ///	Set	the	timestamp format of	the	log	file
        ///	data.
        ///	</summary>
        public TimestampFormat TimeMarkerFormat
        {
            get { return m_timeMarkerFormat; }
            set { m_timeMarkerFormat = value; }
        }

        ///	<summary>
        ///	log	all	data to	a single file if true
        ///	or to separate file	if false.
        ///	</summary>
        public bool SingleFileLogging
        {
            get { return m_singlelogfile; }
            set { m_singlelogfile = value; }
        }

        ///	<summary>
        ///	The	folder to create the log files.
        ///	</summary>
        public string LogPath
        {
            get { return logPath; }
            set
            {
                if (String.IsNullOrEmpty(value))
                {
                    throw new ArgumentException("Cannot be null or empty", "value");
                }

                logPath = value;

                if (logPath.EndsWith("\\",StringComparison.OrdinalIgnoreCase))
                {
                    logPath = logPath.Substring(0, logPath.Length - 2);
                }

                if (! Directory.Exists(logPath))
                {
                    Directory.CreateDirectory(logPath);
                }
            }
        }

        #endregion Properties

        #region	Methods

        ///	<summary>
        ///	Write a	sensors	value to a log file.
        ///	</summary>
        ///	<param name="sensor">The sensor.</param>
        ///	<param name="measurement">The measurement from the sensor.</param>
        public void WriteSensorMeasurement(Sensor sensor, TimeStampedMeasurement measurement)
        {
            if (!m_logdata)
            {
                return;
            }

            if (measurement == null)
            {
                throw new ArgumentNullException("measurement");
            }

            if (sensor == null)
            {
                throw new ArgumentNullException("sensor");
            }

            TimeSpan tspan = measurement.Timestamp.Subtract(m_startTime);
            var wtr = (StreamWriter) m_fileTable[sensor.Name];

            if (m_timeMarkerFormat == TimestampFormat.ClockFormat)
            {
                wtr.Write(tspan.Days);
                wtr.Write(":");
                wtr.Write(tspan.Hours);
                wtr.Write(":");
                wtr.Write(tspan.Minutes);
                wtr.Write(":");
                wtr.Write(tspan.Seconds);
                wtr.Write(".");
                wtr.Write(tspan.Milliseconds.ToString("000", CultureInfo.InvariantCulture));
            }
            else
            {
                wtr.Write(tspan.TotalMilliseconds.ToString("#.000", numberFormat));
            }

            wtr.Write(",");

            if (!m_singlelogfile)
            {
                if (measurement.IsStatus)
                {
                    wtr.Write(measurement.Measurement);
                }
                else
                {
                    wtr.Write(measurement.Measurement.ToString("#.000", numberFormat));
                }

                wtr.WriteLine();
            }
            else
            {
                int i = 1;
                foreach (object o in	m_fileTable.Keys)
                {
                    if (o.ToString() == sensor.Name)
                    {
                        if (measurement.IsStatus)
                        {
                            wtr.Write(measurement.Status);
                        }
                        else
                        {
                            wtr.Write(measurement.Measurement.ToString("#.000", numberFormat));
                        }
                    }

                    wtr.Write(",");

                    if (i == m_fileTable.Count)
                    {
                        wtr.WriteLine();
                        break;
                    }

                    i++;
                }
            }
        }

        ///	<summary>
        ///	Start logging sensor updates to	log	files.
        ///	</summary>	
        ///	<param name="sensors">The list of sensors to log data.</param>
        public void StartLogging(List<Sensor> sensors)
        {
            if (sensors == null)
            {
                throw new ArgumentNullException("sensors");
            }

            StreamWriter wtr = null;

            CloseStreams();

            m_startTime = DateTime.Now;

            if (m_singlelogfile)
            {
                wtr =
                    new StreamWriter(
                        string.Format(CultureInfo.InvariantCulture, "{0}{1}{2}{3}{4}", logPath, Path.DirectorySeparatorChar, LogfileName,
                                      m_startTime.ToString("yyyy_MMM_ddd_HH_mm_ss", CultureInfo.InvariantCulture), LogfileExt), false);
            }

            foreach (Sensor	sensor in sensors)
            {
                if (!m_fileTable.ContainsKey(sensor.Name))
                {
                    if (!m_singlelogfile)
                    {
                        wtr =
                            new StreamWriter(
                                string.Format(CultureInfo.InvariantCulture, "{0}{1}{2}{3}{4}", logPath, Path.DirectorySeparatorChar, sensor.Name,
                                              m_startTime.ToString("yyyy_MMM_ddd_HH_mm_ss", CultureInfo.InvariantCulture), LogfileExt), false);
                    }

                    m_fileTable.Add(sensor.Name, wtr);
                }
            }

            WriteHeaders(sensors);

            m_logdata = true;
        }

        ///	<summary>
        ///	Stop logging data to log files.
        ///	</summary>
        public void StopLogging()
        {
            m_logdata = false;
            CloseStreams();
        }

        #endregion Methods

        #region	Private	Methods

        private void CloseStreams()
        {
            StreamWriter wtr;

            if (m_fileTable.Count == 0) return;

            foreach (Object obj in m_fileTable.Values)
            {
                wtr = (StreamWriter) obj;

                wtr.Flush();
                wtr.Close();

                if (m_singlelogfile) break;
            }

            m_fileTable.Clear();
        }

        private void WriteHeaders(List<Sensor> sensors)
        {
            StreamWriter wtr;

            if (m_singlelogfile)
            {
                wtr = (StreamWriter) m_fileTable[sensors[0].Name];
                wtr.Write(TimestampHdr);
            }
            else
                foreach (object o in	m_fileTable.Keys)
                {
                    wtr = (StreamWriter) m_fileTable[o.ToString()];

                    wtr.Write(TimestampHdr);
                }

            foreach (object o in	m_fileTable.Keys)
            {
                wtr = (StreamWriter) m_fileTable[o.ToString()];

                foreach (Sensor	s in sensors)
                {
                    if (s.Name == o.ToString())
                    {
                        wtr.Write("," + s.Name + " (" + s.Units + ")");
                        break;
                    }
                }
            }

            if (m_singlelogfile)
            {
                wtr = (StreamWriter) m_fileTable[sensors[0].Name];
                wtr.WriteLine();
            }
            else
                foreach (object o in	m_fileTable.Keys)
                {
                    wtr = (StreamWriter) m_fileTable[o.ToString()];

                    wtr.WriteLine();
                }
        }

        #endregion
    }

    #endregion DataLogger
}