#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;
using DnEcuDiag.Configuration;
using DnEcuDiag.Devices;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Ecus.EventArgTypes;
using DnEcuDiag.Ecus.Generic;
using DnEcuDiag.Exceptions;
using DnEcuDiag.Factories;
using DnEcuDiag.Logging;
using DnEcuDiag.Properties;
using DnEcuDiag.VehiclePerformance;
using SerialPort;

namespace DnEcuDiag
{
    ///	<summary>
    ///	The	core component of the application.
    ///	</summary>
    internal sealed class Model : IDisposable
    {
        #region	Singlton Pattern

        // Singleton.
        private static Model _instance;

        ///	<summary>
        ///	Prevent	instantiation from outside of class.
        ///	</summary>
        private Model()
        {
            LoadApplicationConfiguration();

            CreateSerialDeviceConfiguration();

            CreateAndConfigureFoldersAndPath();

            LoadDeviceConfiguration();

            LoadSensorConfiguration();

            VehiclePerformance = new VehiclePerformanceMeasure();
            DynoVehicleProfile = new VehicleProfile();
            Logger = new DataLogger {LogPath = SensorLogPath};

            LastConnectedEcu = EcuFactory.EcuList[0];

            foreach (var ecuDescriptor in EcuFactory.EcuList)
            {
                if (ecuDescriptor.Name == Settings.Default.LastConnectedEcu)
                {
                    LastConnectedEcu = ecuDescriptor;
                    break;
                }
            }
        }

        ///	<summary>
        ///	Singleton Pattern.
        ///	</summary>
        public static Model Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new Model();
                }

                return _instance;
            }
        }

        #endregion

        #region	Dispose	Pattern

        private bool _disposed;

        ///	<summary>
        ///	Dispose	the	underlying resources.
        ///	</summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ///	<summary>
        ///	Close the device.
        ///	</summary>
        public void Close()
        {
            Dispose(true);
        }

        ///	<summary>
        ///	Dispose	of the objects resources.
        ///	</summary>
        ///	<param name="disposing">True if	explicity disposing	object else	GC finalizing object.</param>
        private void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    if (SensorGroups != null)
                    {
                        SensorGroups.Dispose();
                        SensorGroups = null;
                    }

                    if (TargetEcu != null)
                    {
                        ((IDisposable) TargetEcu).Dispose();
                    }

                    if (_diagnosticInterface != null)
                    {
                        _diagnosticInterface.Close();
                        _diagnosticInterface.Dispose();
                        _diagnosticInterface = null;
                    }
                }
            }
            _disposed = true;
        }

        #endregion


        private String _configurationFilename = "SerialDevice.Config";
        private IDevice _diagnosticInterface;
        private SensorConfigurationContainer _selectedSensors;
        private String _sensorGroupsFileName = String.Empty;
        private Boolean _stopPerformanceMeasurement;

        public DataLogger Logger { get; set; }

        public String ApplicationPath { get; private set; }

        ///	<summary>
        ///	Gets the sensor	groups.
        ///	</summary>
        ///	<value>The sensor groups.</value>
        public DataSet SensorGroups { get; private set; }

        ///	<summary>
        ///	Gets the vehicles dyno profile.
        ///	</summary>
        ///	<value>The dyno	vehicle	profile.</value>
        public VehicleProfile DynoVehicleProfile { get; set; }

        ///	<summary>
        ///	Gets the vehicle performance.
        ///	</summary>
        ///	<value>The vehicle performance.</value>
        public VehiclePerformanceMeasure VehiclePerformance { get; private set; }

        ///	<summary>
        ///	Gets the last ECU the application connected	too.
        ///	</summary>
        ///	<value>The last	ECU.</value>
        public EcuDescriptor LastConnectedEcu { get; private set; }

        ///	<summary>
        ///	Gets the ECU.
        ///	</summary>
        ///	<value>The ECU.</value>
        public Ecu TargetEcu { get; private set; }

        ///	<summary>
        ///	Gets a value indicating	whether	[config	loaded].
        ///	</summary>
        ///	<value><c>true</c> if [config loaded]; otherwise, <c>false</c>.</value>
        public Boolean DeviceConfigured { get; private set; }

        public Int32 FinalDriveProgress { get; private set; }

        public float LastDynoRpm { get; private set; }

        public double LastLossMeasureSpeed { get; private set; }

        /// <summary>
        /// The last speed captured.
        /// </summary>
        public double LastSpeedMeasurement { get; private set; }

        ///	<summary>
        ///	Gets the ECU configurations.
        ///	</summary>
        ///	<value>The ECU configurations.</value>
        public static List<EcuDescriptor> EcuConfigurations
        {
            get { return EcuFactory.EcuList; }
        }

        public static List<SerialPortDescriptor> SerialPorts
        {
            get { return Port.GetSerialPorts(); }
        }

        public static ReadOnlyCollection<UsbPortDescriptor> UsbPorts
        {
            get { return Ftd2XxPort.GetUsbPorts(); }
        }

        /// <summary>
        /// Gets the serial device configuration.
        /// </summary>
        /// <value>The serial device configuration.</value>
        public DeviceConfiguration DeviceConfiguration { get; private set; }

        /// <summary>
        /// The path to where vehicle profiles ae stored.
        /// </summary>
        public String VehicleProfileFolder { get; private set; }

        /// <summary>
        /// The path to perfomance run data.
        /// </summary>
        public String PerformanceRunFolder { get; private set; }

        /// <summary>
        /// The path to sensor log data.
        /// </summary>
        public String SensorLogPath { get; private set; }

        /// <summary>
        /// The help file.
        /// </summary>
        public String HelpFile { get; private set; }

        /// <summary>
        /// The path to sensor group configuration.
        /// </summary>
        public String SensorGroupsPath { get; private set; }

        ///	<summary>
        ///	Disconnect from	the	Ecu	and	cleanup	objects.
        ///	</summary>
        public void DisconnectFromEcu()
        {
            if (TargetEcu != null)
            {
                TargetEcu.SensorOneUpdated -= OnSensorOneUpdated;
                TargetEcu.SensorTwoUpdated -= OnSensorTwoUpdated;
                TargetEcu.SensorThreeUpdated -= OnSensorThreeUpdated;
                TargetEcu.SensorFourUpdated -= OnSensorfourUpdated;
                TargetEcu.EcuDisconnected -= OnEcuDisconnected;

                if (TargetEcu.Connected)
                {
                    TargetEcu.Disconnect();
                }
            }

            if (_diagnosticInterface != null)
            {
                _diagnosticInterface.Close();
            }
        }

        ///	<summary>
        ///	Connect	to the Ecu that	was	specified by the index.
        ///	</summary>
        ///	<param name="ecuDescriptor">The descriptor for the ecu to connect too.</param>
        public bool ConnectToEcu(EcuDescriptor ecuDescriptor)
        {
            try
            {
                _diagnosticInterface =
                    InterfaceDeviceFactory.CreateInterfaceDevice(ecuDescriptor, DeviceConfiguration);

                TargetEcu.MonitoredSensorRefreshDelay = (int) Settings.Default.SampleDelay;
                TargetEcu.DiagnosticInterface = _diagnosticInterface;

                if (!string.IsNullOrEmpty(TargetEcu.EcuWarning))
                {
                    if (Common.ShowYesNo(Instance.TargetEcu.EcuWarning) == DialogResult.No)
                    {
                        ReleaseInterfaceDevice();
                        return false;
                    }
                }

                TargetEcu.Connect();
            }
            catch
            {
                ReleaseInterfaceDevice();
                throw;
            }

            TargetEcu.EcuDisconnected += OnEcuDisconnected;
            TargetEcu.SensorOneUpdated += OnSensorOneUpdated;
            TargetEcu.SensorTwoUpdated += OnSensorTwoUpdated;
            TargetEcu.SensorThreeUpdated += OnSensorThreeUpdated;
            TargetEcu.SensorFourUpdated += OnSensorfourUpdated;

            Settings.Default.LastConnectedEcu = ecuDescriptor.Name;

            PrepareSensorGroups(TargetEcu);

            return true;
        }

        private void ConfigureLastSensors()
        {
            if (TargetEcu.Sensors.Count != 0)
            {
                TargetEcu.SensorMonitorOne = TargetEcu.Sensors[0];
                TargetEcu.SensorMonitorTwo = TargetEcu.Sensors[0];
                TargetEcu.SensorMonitorThree = TargetEcu.Sensors[0];
                TargetEcu.SensorMonitorFour = TargetEcu.Sensors[0];
            }

            foreach (var sensorConfiguration in _selectedSensors.SensorConfigurations)
            {
                if (TargetEcu.Model == sensorConfiguration.EcuModel)
                {
                    TargetEcu.SensorMonitorOne = FindSensor(sensorConfiguration.ChannelOneIndex);
                    TargetEcu.SensorMonitorTwo = FindSensor(sensorConfiguration.ChannelTwoIndex);
                    TargetEcu.SensorMonitorThree = FindSensor(sensorConfiguration.ChannelThreeIndex);
                    TargetEcu.SensorMonitorFour = FindSensor(sensorConfiguration.ChannelFourIndex);
                }
            }
        }

        private Sensor FindSensor(int idenifier)
        {
            Sensor s = TargetEcu.Sensors[0];
            foreach (var sensor in TargetEcu.Sensors)
            {
                if (sensor.Id == idenifier)
                {
                    s = sensor;
                    break;
                }
            }

            return s;
        }

        ///	<summary>
        ///	Save the changes to	the	sensor groups.
        ///	</summary>
        public void SaveSensorGroups()
        {
            SensorGroups.WriteXml(_sensorGroupsFileName, XmlWriteMode.WriteSchema);
            SensorGroups.AcceptChanges();
        }

        ///	<summary>
        ///	Disgard	changes	made to	the	sensor groups.
        ///	</summary>
        public void DisgardSensorGroupsChanges()
        {
            SensorGroups.AcceptChanges();
            SensorGroups.WriteXml(_sensorGroupsFileName, XmlWriteMode.WriteSchema);
        }

        public static void LoadApplicationConfiguration()
        {
            Settings.Default.Reload();
        }

        private void ReleaseInterfaceDevice()
        {
            if (_diagnosticInterface != null)
            {
                _diagnosticInterface.Close();
                _diagnosticInterface.Dispose();
            }
        }

        private void PrepareSensorGroups(Ecu ecu)
        {
            _sensorGroupsFileName = SensorGroupsPath + Path.DirectorySeparatorChar + ecu.Model + ".xml";

            if (!File.Exists(_sensorGroupsFileName))
            {
                CreateDatasetSchema();
                SensorGroups.WriteXml(_sensorGroupsFileName, XmlWriteMode.WriteSchema);
            }
            else
            {
                SensorGroups = new DataSet {Locale = CultureInfo.InvariantCulture};
                SensorGroups.ReadXml(_sensorGroupsFileName, XmlReadMode.ReadSchema);
                SensorGroups.AcceptChanges();
            }
        }

        private void CreateDatasetSchema()
        {
            SensorGroups = new DataSet {Locale = CultureInfo.InvariantCulture};

            DataTable sensorGroupTable = SensorGroups.Tables.Add("SensorGroups");

            DataColumn pkCol = sensorGroupTable.Columns.Add("GroupId", typeof (Int32));
            pkCol.AutoIncrement = true;

            var col = new DataColumn("GroupName", typeof (string)) {DefaultValue = String.Empty};
            sensorGroupTable.Columns.Add(col);

            col = new DataColumn("GroupDescription", typeof (string)) {DefaultValue = String.Empty};
            sensorGroupTable.Columns.Add(col);

            col = new DataColumn("SensorIndexOne", typeof (Int32)) {DefaultValue = 0};
            sensorGroupTable.Columns.Add(col);

            col = new DataColumn("SensorIndexTwo", typeof (Int32)) {DefaultValue = 0};
            sensorGroupTable.Columns.Add(col);

            col = new DataColumn("SensorIndexThree", typeof (Int32)) {DefaultValue = 0};
            sensorGroupTable.Columns.Add(col);

            col = new DataColumn("SensorIndexFour", typeof (Int32)) {DefaultValue = 0};
            sensorGroupTable.Columns.Add(col);

            sensorGroupTable.PrimaryKey = new[] {pkCol};
        }

        private void CreateAndConfigureFoldersAndPath()
        {
            ApplicationPath =
                Path.GetDirectoryName(Assembly.GetExecutingAssembly().GetName().CodeBase).Replace("file:\\", "");

            VehicleProfileFolder = ApplicationPath + Path.DirectorySeparatorChar + "VehicleProfiles";

            if (!Directory.Exists(VehicleProfileFolder))
            {
                Directory.CreateDirectory(VehicleProfileFolder);
            }

            PerformanceRunFolder = ApplicationPath + Path.DirectorySeparatorChar + "PerformanceRuns";

            if (!Directory.Exists(PerformanceRunFolder))
            {
                Directory.CreateDirectory(PerformanceRunFolder);
            }

            SensorLogPath = ApplicationPath + Path.DirectorySeparatorChar + "SensorLogs";

            if (!Directory.Exists(SensorLogPath))
            {
                Directory.CreateDirectory(SensorLogPath);
            }

            SensorGroupsPath = ApplicationPath + Path.DirectorySeparatorChar + "SensorGroups";

            if (!Directory.Exists(SensorGroupsPath))
            {
                Directory.CreateDirectory(SensorGroupsPath);
            }

            HelpFile = ApplicationPath + Path.DirectorySeparatorChar + "DnEcuDiagHelp.chm";

            _configurationFilename = ApplicationPath + Path.DirectorySeparatorChar + _configurationFilename;
        }

        private void LoadSensorConfiguration()
        {
            _selectedSensors = !String.IsNullOrEmpty(Settings.Default.SelectedSensors)
                                   ? ConfigurationSerializer.DeserializeSensorConfigurations(Settings.Default.SelectedSensors)
                                   : new SensorConfigurationContainer();
        }

        private void CreateSerialDeviceConfiguration()
        {
            DeviceConfiguration = new DeviceConfiguration
                                      {
                                          KwpTiming =
                                              {
                                                  EcuResponseInterbyteTimeout =
                                                      Kwp2000SerialDevice.DefaultEcuResponseInterbyteTimeout,
                                                  EcuResponseTimeout = Kwp2000SerialDevice.DefaultEcuResponseTimeout,
                                                  RequestInterByteTime =
                                                      Kwp2000SerialDevice.DefaultRequestInterbyteTimeout
                                              },
                                          IsoTiming =
                                              {
                                                  EcuResponseInterbyteTimeout =
                                                      Kwp81SerialDevice.DefaultEcuResponseInterbyteTimeout,
                                                  EcuResponseTimeout = Kwp81SerialDevice.DefaultEcuResponseTimeout,
                                                  RequestInterByteTime =
                                                      Kwp81SerialDevice.DefaultRequestInterbyteTimeout
                                              }
                                      };
        }

        private SensorConfiguration RetrieveSensorConfiguration()
        {
            SensorConfiguration sensorConfiguration = null;

            lock (_selectedSensors)
            {
                foreach (var configuration in _selectedSensors.SensorConfigurations)
                {
                    if (configuration.EcuModel == Instance.TargetEcu.Model)
                    {
                        sensorConfiguration = configuration;
                        break;
                    }
                }

                if (sensorConfiguration == null)
                {
                    sensorConfiguration = new SensorConfiguration(Instance.TargetEcu.Model);
                    _selectedSensors.SensorConfigurations.Add(sensorConfiguration);
                }
            }

            return sensorConfiguration;
        }

        private void OnEcuDisconnected(Object sender, EcuDisconnectEventArgs e)
        {
            Logger.StopLogging();

            ReleaseInterfaceDevice();
        }

        private static void OnSensorOneUpdated(object sender, SensorUpdatedEventArgs e)
        {
            if (_instance.TargetEcu.SensorMonitorOneEnable)
            {
                Instance.Logger.WriteSensorMeasurement(Instance.TargetEcu.SensorMonitorOne, e.Current);
            }
        }

        private static void OnSensorTwoUpdated(object sender, SensorUpdatedEventArgs e)
        {
            if (_instance.TargetEcu.SensorMonitorTwoEnable)
            {
                Instance.Logger.WriteSensorMeasurement(Instance.TargetEcu.SensorMonitorTwo, e.Current);
            }
        }

        private static void OnSensorThreeUpdated(object sender, SensorUpdatedEventArgs e)
        {
            if (_instance.TargetEcu.SensorMonitorThreeEnable)
            {
                Instance.Logger.WriteSensorMeasurement(Instance.TargetEcu.SensorMonitorThree, e.Current);
            }
        }

        private static void OnSensorfourUpdated(object sender, SensorUpdatedEventArgs e)
        {
            if (_instance.TargetEcu.SensorMonitorFourEnable)
            {
                Instance.Logger.WriteSensorMeasurement(Instance.TargetEcu.SensorMonitorFour, e.Current);
            }
        }

        internal void SaveSelectedSensorSetting(SensorChannel channel, int sensorIndex)
        {
            SensorConfiguration sensorConfiguration = RetrieveSensorConfiguration();

            switch (channel)
            {
                case SensorChannel.ChannelOne:
                    sensorConfiguration.ChannelOneIndex = sensorIndex;
                    break;
                case SensorChannel.ChannelTwo:
                    sensorConfiguration.ChannelTwoIndex = sensorIndex;
                    break;
                case SensorChannel.ChannelThree:
                    sensorConfiguration.ChannelThreeIndex = sensorIndex;
                    break;
                case SensorChannel.ChannelFour:
                    sensorConfiguration.ChannelFourIndex = sensorIndex;
                    break;
                default:
                    break;
            }
        }

        internal int GetSelectedSensorSetting(SensorChannel channel)
        {
            SensorConfiguration sensorConfiguration = RetrieveSensorConfiguration();
            Int32 index = 0;

            switch (channel)
            {
                case SensorChannel.ChannelOne:
                    index = sensorConfiguration.ChannelOneIndex;
                    break;
                case SensorChannel.ChannelTwo:
                    index = sensorConfiguration.ChannelTwoIndex;
                    break;
                case SensorChannel.ChannelThree:
                    index = sensorConfiguration.ChannelThreeIndex;
                    break;
                case SensorChannel.ChannelFour:
                    index = sensorConfiguration.ChannelFourIndex;
                    break;
                default:
                    break;
            }
            return index;
        }

        internal void SaveSelectedSensorConfig()
        {
            Settings.Default.SelectedSensors = ConfigurationSerializer.SerializeSensorConfigurations(_selectedSensors);
            Settings.Default.Save();
        }


        internal void SaveDeviceConfiguration()
        {
            ConfigurationSerializer.SerializeSerialDeviceConfiguration(_configurationFilename, DeviceConfiguration);
            DeviceConfigured = true;
        }

        internal void LoadDeviceConfiguration()
        {
            if (File.Exists(_configurationFilename))
            {
                DeviceConfiguration =
                    ConfigurationSerializer.DeserializeDeviceConfigurationFromFile(_configurationFilename);
                DeviceConfigured = true;
            }
            else
            {
                if (SerialPorts.Count > 0)
                {
                    DeviceConfiguration.SerialPort = SerialPorts[0];
                }
            }
        }

        internal void SaveVehiclePerformanceRun(String vehiclePerformanceRunFile)
        {
            if (!vehiclePerformanceRunFile.Contains(".vpr"))
            {
                vehiclePerformanceRunFile += ".vpr";
            }

            ConfigurationSerializer.SerializeVehiclePerformanceRun(vehiclePerformanceRunFile, VehiclePerformance.PerformanceRun);
        }

        internal void LoadPerformanceRun(String vehiclePerformanceRunFile)
        {
            VehiclePerformance.PerformanceRun = ConfigurationSerializer.DeserializeVehiclePerformanceRun(vehiclePerformanceRunFile);
        }

        internal void SaveApplicationConfiguration()
        {
            if (TargetEcu != null)
            {
                TargetEcu.MonitoredSensorRefreshDelay = (int) Settings.Default.SampleDelay;
            }

            Logger.SingleFileLogging = Settings.Default.SingleLogFile;

            Logger.TimeMarkerFormat = Settings.Default.MillisecondFormatTimestamp ? DataLogger.TimestampFormat.MillisecondFormat : DataLogger.TimestampFormat.ClockFormat;

            Settings.Default.Save();
        }

        internal void SaveVehicleProfile(String vehicleProfileFile)
        {
            String file = ValidateProfileFileName(vehicleProfileFile);

            ConfigurationSerializer.SerializeVehicleProfile(file, DynoVehicleProfile);
            Settings.Default.LastDynoVehicleProfile = vehicleProfileFile;
        }

        internal void HandleLoadVehicleProfile(String vehicleProfileFile)
        {
            String file = ValidateProfileFileName(vehicleProfileFile);

            DynoVehicleProfile = ConfigurationSerializer.DeserializeVehicleProfile(file);

            if (Instance.DynoVehicleProfile != null)
            {
                Settings.Default.LastDynoVehicleProfile = file;
            }
        }

        private string ValidateProfileFileName(String vehicleProfileFile)
        {
            if (String.IsNullOrEmpty(_sensorGroupsFileName))
            {
                throw new ArgumentException("Cannot be null or empty", "vehicleProfileFile");
            }

            string file;

            if (vehicleProfileFile.IndexOf(".vprof", 0, StringComparison.OrdinalIgnoreCase) != -1)
            {
                file = vehicleProfileFile;
            }
            else
            {
                file = vehicleProfileFile + ".vprof";
            }
            return file;
        }

        internal void StartPerformanceRun()
        {
            _stopPerformanceMeasurement = false;

            if (Instance.TargetEcu.SpeedSensor == null)
            {
                throw new PerformanceRunNotSupportedException(WinFormStrings.ThisEcuDoesNotSupportPerformanceMeasurements);
            }

            // Wait for speed to increase from zero.
            var speed = new TimeStampedMeasurement(0, string.Empty, DateTime.Now, false);
            Instance.VehiclePerformance.Reset();
            LastSpeedMeasurement = speed.Measurement;

            while (speed.Measurement == 0)
            {
                speed = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.SpeedSensor);

                Thread.Sleep(10);

                if (_stopPerformanceMeasurement)
                {
                    return;
                }
            }

            LastSpeedMeasurement = speed.Measurement;

            while (!_stopPerformanceMeasurement)
            {
                Thread.Sleep(5);

                speed = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.SpeedSensor);

                Instance.VehiclePerformance.AddSpeedMeasurement(speed);

                // While the measurements not complete continue.
                if (Instance.VehiclePerformance.Complete)
                {
                    break;
                }
            }
        }

        internal void StopPerformanceMeasurement()
        {
            _stopPerformanceMeasurement = true;
        }

        /// <summary>
        /// Capture loss measurements so that frictional and aerodynamic loss can be measeured.
        /// </summary>
        /// <returns></returns>
        /// <exception cref="DynoRunNotSupportedException"></exception>
        /// <exception cref="ArgumentNullException"></exception>
        internal bool StartLossMeasure()
        {
            if ((Instance.TargetEcu.SpeedSensor == null) | (Instance.TargetEcu.RpmSensor == null))
            {
                throw new DynoRunNotSupportedException("This ecu does not support Dyno measurements.");
            }

            ThreadPriority currentPriority = Thread.CurrentThread.Priority;

            try
            {
                Thread.CurrentThread.Priority = ThreadPriority.Highest;

                Instance.VehiclePerformance.DynoMass = Instance.DynoVehicleProfile.VehicleMass;
                TimeStampedMeasurement speed = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.SpeedSensor);
                LastSpeedMeasurement = speed.Measurement;

                while (true)
                {
                    // Read speed measurement.
                    speed = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.SpeedSensor);

                    if (speed.Measurement < LastSpeedMeasurement)
                    {
                        Instance.VehiclePerformance.AddLossMeasurement(speed);

                        LastSpeedMeasurement = speed.Measurement;
                    }

                    if (speed.Measurement == 0)
                    {
                        break;
                    }

                    if (_stopPerformanceMeasurement)
                    {
                        break;
                    }

                    Thread.Sleep(300);
                }


                Instance.VehiclePerformance.DynoLossCalculate();

                Instance.DynoVehicleProfile.FrictionLoss = Instance.VehiclePerformance.FrictionLossFactor;
                Instance.DynoVehicleProfile.DynamicLoss = Instance.VehiclePerformance.AerodynamicLossFactor;

                return true;
            }
            finally
            {
                Thread.CurrentThread.Priority = currentPriority;
                _stopPerformanceMeasurement = false;
            }
        }

        /// <summary>
        /// Capture rpm and speed measurements to calculate a final drive ratio.
        /// </summary>
        internal void CaptureFinalDriveRatio(Int32 selectedGearIndex)
        {
            if (selectedGearIndex == -1)
            {
                throw new PerformanceRunNotSupportedException("Selected gear invalid.");
            }

            if ((Instance.TargetEcu.SpeedSensor == null) | (Instance.TargetEcu.RpmSensor == null))
            {
                throw new PerformanceRunNotSupportedException(WinFormStrings.ThisEcuDoesNotSupportPerformanceMeasurements);
            }

            ThreadPriority currentPriority = Thread.CurrentThread.Priority;

            try
            {
                Thread.CurrentThread.Priority = ThreadPriority.Highest;

                float rpmSum = 0;
                float speedSum = 0;
                TimeStampedMeasurement rpm = null;
                TimeStampedMeasurement speed = null;

                FinalDriveProgress = 0;

                for (int i = 0; i < 25; i++)
                {
                    speed = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.SpeedSensor);
                    rpm = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.RpmSensor);

                    speedSum += speed.Measurement;
                    rpmSum += rpm.Measurement;

                    FinalDriveProgress++;

                    Thread.Sleep(100);
                }

                // Calculate final drive ratio
                speed.Measurement = speedSum/25;
                rpm.Measurement = rpmSum/25;

                // Prevent division by zero.
                if (!((rpm.Measurement == 0) | (speed.Measurement == 0)))
                {
                    Instance.VehiclePerformance.DynoSetGear(
                        VehiclePerformanceMeasure.ConvertSpeedToMetersSecond((int) speed.Measurement),
                        (int) rpm.Measurement);
                }

                Instance.DynoVehicleProfile.FinalDriveRatios[selectedGearIndex] = Instance.VehiclePerformance.DynoGear;
            }
            finally
            {
                Thread.CurrentThread.Priority = currentPriority;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="startRpm"></param>
        /// <param name="stopRpm"></param>
        /// <param name="gearIndex"></param>
        /// <param name="vehicleMass"></param>
        /// <returns></returns>
        public bool StartDynoMeasure(Int32 startRpm, Int32 stopRpm, Int32 gearIndex, Double vehicleMass)
        {
            if (startRpm == stopRpm)
            {
                throw new PerformanceRunNotSupportedException("The Start and Stop Rpm are the same. The Dyno run cannot be performed.");
            }

            if (gearIndex == -1)
            {
                throw new PerformanceRunNotSupportedException("Please select a gear to perform the Dyno run.");
            }

            if (Instance.DynoVehicleProfile.FinalDriveRatios[gearIndex] == 0)
            {
                throw new PerformanceRunNotSupportedException(
                    "A Final Drive Ratio must be configured for the selected gear before a Dyno measurment can be performed.");
            }

            if ((Instance.TargetEcu.SpeedSensor == null) | (Instance.TargetEcu.RpmSensor == null))
            {
                throw new PerformanceRunNotSupportedException("This ecu does not support Dyno measurements.");
            }

            ThreadPriority currentPriority = Thread.CurrentThread.Priority;
            try
            {
                Thread.CurrentThread.Priority = ThreadPriority.Highest;

                TimeStampedMeasurement rpm = null;
                bool started = false;
                _stopPerformanceMeasurement = false;
                LastDynoRpm = 0;


                Instance.VehiclePerformance.Reset();
                Instance.VehiclePerformance.DynoMass = vehicleMass;
                Instance.DynoVehicleProfile.VehicleMass = vehicleMass;
                Instance.VehiclePerformance.DynoGear = (int) Instance.DynoVehicleProfile.FinalDriveRatios[gearIndex];

                while (!started)
                {
                    rpm = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.RpmSensor);

                    Thread.Sleep(5);

                    if (rpm.Measurement >= startRpm)
                    {
                        started = true;
                    }

                    if (_stopPerformanceMeasurement)
                    {
                        return false;
                    }
                }

                LastDynoRpm = rpm.Measurement;

                Instance.VehiclePerformance.AddRpmMeasurement(rpm);

                while (true)
                {
                    // Read Rpm measurement.
                    rpm = Instance.TargetEcu.ReadSensor(Instance.TargetEcu.RpmSensor);

                    if (rpm.Measurement >= stopRpm)
                    {
                        Instance.VehiclePerformance.AddRpmMeasurement(rpm);
                        break;
                    }

                    if (rpm.Measurement < LastDynoRpm)
                    {
                        return false;
                    }

                    if (_stopPerformanceMeasurement)
                    {
                        return false;
                    }

                    if (LastDynoRpm == rpm.Measurement)
                    {
                        continue;
                    }

                    Instance.VehiclePerformance.AddRpmMeasurement(rpm);
                    LastDynoRpm = rpm.Measurement;

                    Thread.Sleep(200);
                }

                return true;
            }
            finally
            {
                Thread.CurrentThread.Priority = currentPriority;
            }
        }

        internal void ShutDown()
        {
            SaveSelectedSensorConfig();
            _instance.DisconnectFromEcu();
        }

        internal void SelectTargetEcu(EcuDescriptor ecuDescriptor)
        {
            TargetEcu = EcuFactory.CreateEcu(ecuDescriptor);
            ConfigureLastSensors();
        }
    }
}