#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Diagnostics;
using System.Threading;
using DnEcuDiag.Devices.Enums;
using DnEcuDiag.Devices.Exceptions;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Devices.Responses;
using DnEcuDiag.Ecus.Enums;
using DnEcuDiag.Ecus.EventArgTypes;
using DnEcuDiag.Ecus.Exceptions;
using DnEcuDiag.Ecus.Generic;
using DnEcuDiag.Ecus.Kwp;
using ThreadState=System.Threading.ThreadState;

namespace DnEcuDiag.Protocols
{
    ///	<summary>
    ///	Class that implements the KWP2000 protocol.
    ///	</summary>
    internal sealed class Kwp2000 : Protocol, IDisposable
    {
        private const int KeepAliveInterval = 1;

        #region	Fields

        private readonly SyncCounter _msgSentCount = new SyncCounter(0);
        private readonly byte _ecuId;

        private readonly IDevice _interfaceDevice;
        private readonly Object _lockObject = new Object();
        private readonly byte _testerId;
        private Boolean _keepAliveActive;
        private Thread _keepAliveThread;

        private Boolean _protocolConnected;
        private ManualResetEvent _signalKeepAlive = new ManualResetEvent(false);

        #endregion

        #region	Constructors

        ///	<summary>
        ///	Constructor	for	Kwp.
        ///	</summary>
        ///	<param name="interfaceDevice">The hardware interface device	to use for communication.</param>
        ///	<param name="ecuId">The	EcuID</param>
        ///	<param name="testerId">The TesterID</param>
        public Kwp2000(IDevice interfaceDevice, byte ecuId, byte testerId)
        {
            _ecuId = ecuId;
            _interfaceDevice = interfaceDevice;
            _testerId = testerId;
            _interfaceDevice = interfaceDevice;
            _testerId = testerId;
            _ecuId = ecuId;
        }

        #endregion

        #region	Dispose	Pattern

        private bool _disposed;

        ///	<summary>
        ///	Dispose	the	underlying resources.
        ///	</summary>
        public void Dispose()
        {
            GC.SuppressFinalize(this);

            Dispose(true);
        }

        ///	<summary>
        ///	Finalizer.
        ///	</summary>
        ~Kwp2000()
        {
            Dispose(false);
        }

        ///	<summary>
        ///	Close the device.
        ///	</summary>
        public void Close()
        {
            Dispose(true);
        }

        ///	<summary>
        ///	Dispose	of the objects resources.
        ///	</summary>
        ///	<param name="disposing">True if	explicity disposing	object else	GC finalizing object.</param>
        private void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    StopKeepAlive();

                    if (_signalKeepAlive != null)
                    {
                        _signalKeepAlive.Close();
                        _signalKeepAlive = null;
                    }

                    _protocolConnected = false;
                }
            }

            _disposed = true;
        }

        #endregion Dispose Pattern

        #region	Properties

        ///	<summary>
        ///	The	interface device.
        ///	</summary>
        public IDevice InterfaceDevice
        {
            get { return _interfaceDevice; }
        }

        #endregion Properties

        #region	Methods

        ///	<summary>
        ///	Start the keep alove thread.
        ///	</summary>
        public void StartKeepAlive()
        {
            _signalKeepAlive.Set();
        }

        ///	<summary>
        ///	Send start Communication message to	the	target ECU.
        ///	Starts the keep	alive thread.
        ///	</summary>
        ///	<param name="type">The initalisation type that the ecu expects.</param>
        /// <exception cref="Kwp2000Exception"></exception>
        public void StartCommunication(InitType type)
        {
            if (_interfaceDevice == null)
            {
                throw new Kwp2000Exception("The Interface device has not be set.");
            }

            byte[] command = {(byte) KwpCommands.StartCommunication};

            try
            {
                _keepAliveThread = new Thread(KeepAliveMethod) { Name = "KeepAliveThread", IsBackground = true, Priority = ThreadPriority.AboveNormal };
                _keepAliveThread.Start();
            }
            catch (Exception e)
            {
                throw new Kwp2000Exception("Unable to start the keep alive thread.", e);
            }

            // Wait	for	the	thread to start.
            if (!_signalKeepAlive.WaitOne(1000, false))
            {
                throw new Kwp2000Exception("Failed to start keep alive thread.");
            }

            try
            {
                _interfaceDevice.SendCommandWithInit(_ecuId, _testerId, command, (DeviceInitType) type);
            }
            catch (Exception e)
            {
                throw new Kwp2000Exception("Error sending command to Ecu.", e);
            }

            _protocolConnected = true;
        }

        ///	<summary>
        ///	Send stop command to ECU. Stops	the	keep alive thread.
        ///	</summary>
        public void StopCommunication()
        {
            byte[] command = {(byte) KwpCommands.StopCommunication};

            StopKeepAlive();

            if (_protocolConnected)
            {
                SendKwpCommand(command);
            }

            _protocolConnected = false;
        }

        ///	<summary>
        ///	Access the timing parameters for the connection.
        ///	</summary>
        ///	<param name="parameter"></param>
        ///	<param name="p2Min"></param>
        ///	<param name="p2Max"></param>
        ///	<param name="p3Min"></param>
        ///	<param name="p3Max"></param>
        ///	<param name="p4Min"></param>
        public void AccessTimingParameter(CommsParamIdentifier parameter, ref byte p2Min, ref byte p2Max, ref byte p3Min,
                                          ref byte p3Max, ref byte p4Min)
        {
            byte[] command = parameter == CommsParamIdentifier.SetValues
                                 ? new[]
                                       {
                                           (byte) KwpCommands.AccessTimingParameter, (byte) parameter, p2Min, p2Max, p3Min, p3Max,
                                           p4Min
                                       }
                                 : new[] {(byte) KwpCommands.AccessTimingParameter, (byte) parameter};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("AccessTimingParameter failed.", (ResponseCode) response[2]);
            }

            if ((parameter == CommsParamIdentifier.ReadCurrent) | (parameter == CommsParamIdentifier.ReadPossibleLimits))
            {
                p2Min = response[2];
                p2Max = response[3];
                p3Min = response[4];
                p3Max = response[5];
                p4Min = response[6];
            }
        }

        ///	<summary>
        ///	Send StartDiagnosticSession
        ///	</summary>
        public void StartDiagnosticSession()
        {
            byte[] command = {(byte) KwpCommands.StartDiagnosticSession};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("StartDiagnosticSession failed.", (ResponseCode) response[2]);
            }

            _protocolConnected = true;
        }

        ///	<summary>
        ///	Send StartDiagnosticSession
        ///	</summary>
        ///	<param name="diagnosticMode">The diagnostic	mode to	start.</param>
        public void StartDiagnosticSession(byte diagnosticMode)
        {
            byte[] command = {(byte) KwpCommands.StartDiagnosticSession, diagnosticMode};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("StartDiagnosticSession failed.", (ResponseCode) response[2]);
            }

            _protocolConnected = true;
        }

        ///	<summary>
        ///	Send StopDiagnosticSession
        ///	</summary>
        public void StopDiagnosticSession()
        {
            byte[] command = {(byte) KwpCommands.StopDiagnosticSession};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("StopDiagnosticSession failed.", (ResponseCode) response[2]);
            }
        }

        ///	<summary>
        ///	Send the ECU tester	present	message. 
        ///	</summary>
        public void TesterPresent()
        {
            byte[] command = {(byte) KwpCommands.TesterPresent};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("TesterPresent failed.", (ResponseCode) response[2]);
            }
        }

        ///	<summary>
        ///	Read the identification	data from the ECU.
        ///	</summary>
        ///	<param name="identificationOption">The identification option to	read.</param>
        ///	<returns>The identification	data.</returns>
        public Byte[] ReadEcuIdentification(byte identificationOption)
        {
            byte[] command = {(byte) KwpCommands.ReadEcuIdentification, identificationOption};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("readEcuIdentification failed.", (ResponseCode) response[2]);
            }

            return response.Bytes;
        }

        ///	<summary>
        ///	Read a specified local identifier.
        ///	</summary>
        ///	<param name="recordLocalIdentifier">The	recordlocalIdentifier to read.</param>
        ///	<returns>The data retrived from	the	ecu	for	the	recordlocalIdentifier.</returns>
        ///	<exception cref="Kwp2000Exception">Thrown when protocol error detected.</exception>
        public byte[] ReadDataByLocalIdentifier(byte recordLocalIdentifier)
        {
            byte[] command = {(byte) KwpCommands.ReadDataByLocalIdentifier, recordLocalIdentifier};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("ReadDataByLocalIdentifier failed.", (ResponseCode) response[2]);
            }

            return response.Bytes;
        }

        ///	<summary>
        ///	Read a specified local identifier. Special version for retrieval of	fault data.
        ///	</summary>
        ///	<param name="recordLocalIdentifier">The	recordlocalIdentifier to read.</param>
        ///	<param name="dtcHigh"></param>
        ///	<param name="dtcLow"></param>
        ///	<returns>The data retrived from	the	ecu	for	the	recordlocalIdentifier.</returns>
        public byte[] ReadDataByLocalIdentifier(byte recordLocalIdentifier, byte dtcHigh, byte dtcLow)
        {
            byte[] command = {(byte) KwpCommands.ReadDataByLocalIdentifier, recordLocalIdentifier, dtcHigh, dtcLow};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("ReadDataByLocalIdentifier failed.", (ResponseCode) response[2]);
            }

            return response.Bytes;
        }

        ///	<summary>
        ///	Read the active	fault codes.
        ///	</summary>
        ///	<param name="statusOfDtc">The status of	DTCs to	read.</param>
        ///	<param name="groupOfDtcHigh">Upper byte	of DTC group to	read.</param>
        ///	<param name="groupOfDtcLow"></param>
        ///	<returns></returns>
        public Byte[] ReadDiagnosticTroubleCodesByStatus(byte statusOfDtc, byte groupOfDtcHigh, byte groupOfDtcLow)
        {
            byte[] command = {
                                 (byte) KwpCommands.ReadDiagnosticTroubleCodesByStatus, statusOfDtc, groupOfDtcHigh,
                                 groupOfDtcLow
                             };

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("ReadDiagnosticTroubleCodesByStatus failed.", (ResponseCode) response[2]);
            }

            return response.Bytes;
        }

        ///	<summary>
        ///	Read the status	byte of	a diagnostic trouble code.
        ///	</summary>
        ///	<param name="dtcHigh"></param>
        ///	<param name="dtcLow"></param>
        ///	<returns></returns>
        public byte[] ReadStatusOfDiagnosticTroubleCodes(byte dtcHigh, byte dtcLow)
        {
            byte[] command = {(byte) KwpCommands.ReadStatusOfDiagnosticTroubleCodes, dtcHigh, dtcLow};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("ReadStatusOfDiagnosticTroubleCodes failed.", (ResponseCode) response[2]);
            }

            return response.Bytes;
        }

        ///	<summary>
        ///	Clears all DTCs	in the system.
        ///	</summary>
        public void ClearDiagnosticInformation()
        {
            byte[] command = {(byte) KwpCommands.ClearDiagnosticInformation, 0xFF, 0x00};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("ClearDiagnosticInformation failed.", (ResponseCode) response[2]);
            }
        }

        ///	<summary>
        ///	Inputs the output control by local identifier.
        ///	</summary>
        ///	<param name="inputOutputLocalIdentifier">The input output local	identifier.</param>
        ///	<param name="controlParameter">The control parameter.</param>
        /// <exception cref="Kwp2000Exception"></exception>
        ///	<returns></returns>
        public byte[] InputOutputControlByLocalIdentifier(byte inputOutputLocalIdentifier,
                                                          InputOutputControlParameter controlParameter)
        {
            byte[] command = {
                                 (byte) KwpCommands.InputOutputControlByLocalIdentifier, inputOutputLocalIdentifier,
                                 (byte) controlParameter
                             };

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("InputOutputControlByLocalIdentifier failed.", (ResponseCode) response[2]);
            }

            return response.Bytes;
        }

        ///	<summary>
        ///	Inputs the output control by local identifier.
        ///	</summary>
        ///	<param name="inputOutputLocalIdentifier">The input output local	identifier.</param>
        ///	<param name="controlParameter">The control parameter.</param>
        ///	<param name="controlStateParameters">The control state parameters.</param>
        ///	<returns></returns>
        public byte[] InputOutputControlByLocalIdentifier(byte inputOutputLocalIdentifier,
                                                          InputOutputControlParameter controlParameter,
                                                          byte[] controlStateParameters)
        {
            if (controlStateParameters == null)
            {
                throw new ArgumentNullException("controlStateParameters");
            }

            var command = new byte[controlStateParameters.Length + 3];
            command[0] = (byte) KwpCommands.InputOutputControlByLocalIdentifier;
            command[1] = inputOutputLocalIdentifier;
            command[2] = (byte) controlParameter;

            for (int i = 0; i < controlStateParameters.Length; i++)
            {
                command[i + 3] = controlStateParameters[i];
            }

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("InputOutputControlByLocalIdentifier failed.", (ResponseCode) response[2]);
            }

            return response.Bytes;
        }

        ///	<summary>
        ///	Starts the routine by local	identifier.
        ///	</summary>
        ///	<param name="routineLocalIdentifier">The routine local identifier.</param>
        ///	<param name="routineEntryOptions">The routine entry	options.</param>
        public void StartRoutineByLocalIdentifier(byte routineLocalIdentifier, byte[] routineEntryOptions)
        {
            if (routineEntryOptions == null)
            {
                throw new ArgumentNullException("routineEntryOptions");
            }

            var command = new byte[routineEntryOptions.Length + 2];
            command[0] = (byte) KwpCommands.StartRoutinebyLocalIdentifier;
            command[1] = routineLocalIdentifier;

            for (int i = 0; i < routineEntryOptions.Length; i++)
            {
                command[i + 2] = routineEntryOptions[i];
            }

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("startRoutineByLocalIdentifier failed.", (ResponseCode) response[2]);
            }
        }

        ///	<summary>
        ///	Stops the routine by local identifier.
        ///	</summary>
        ///	<param name="routineLocalIdentifier">The routine local identifier.</param>
        public void StopRoutineByLocalIdentifier(byte routineLocalIdentifier)
        {
            byte[] command = {(byte) KwpCommands.StartRoutinebyLocalIdentifier, routineLocalIdentifier};

            Kwp2000Response response = SendKwpCommand(command);

            if (response[0] == 0x7F)
            {
                throw new Kwp2000Exception("StopRoutineByLocalIdentifier failed.", (ResponseCode) response[2]);
            }
        }

        ///	<summary>
        ///	Requests the routine results by	local identifier.
        ///	</summary>
        ///	<param name="routineLocalIdentifier">The routine local identifier.</param>
        ///	<returns>The result	bytes from the routine.</returns>
        public byte[] RequestRoutineResultsByLocalIdentifier(byte routineLocalIdentifier)
        {
            byte[] command = {(byte) KwpCommands.RequestRoutineResultByLocalIdentifier, routineLocalIdentifier};

            Kwp2000Response response = SendKwpCommand(command);

            return response.Bytes;
        }

        ///	<summary>
        ///	The	keep alive thread method this sends	a message to the
        ///	ecu	every 4	seconds.
        ///	</summary>
        private void KeepAliveMethod()
        {
            var keepAliveMessage = new[] {(byte) KwpCommands.TesterPresent};

            try
            {
                _keepAliveActive = true;
                _signalKeepAlive.Set();

                while (_keepAliveActive)
                {
                    int count = 0;
                    _signalKeepAlive.WaitOne();

                    while (count != KeepAliveInterval)
                    {
                        Thread.Sleep(1000);

                        // if any messages have	been sent then reset the counter.
                        if (_msgSentCount.Count > 0)
                        {
                            _msgSentCount.Count = 0;
                            count = 0;
                        }
                        else
                        {
                            count++;
                        }
                    }

                    lock (_lockObject)
                    {
                        if (_protocolConnected != true)
                        {
                            continue;
                        }

                        if (_msgSentCount.Count == 0)
                        {
                            var response = _interfaceDevice.SendCommand(_ecuId, _testerId, keepAliveMessage) as Kwp2000Response;

#if DEBUG
                            PrintRequestResponseArray(keepAliveMessage, response.Bytes);
#endif
                        }
                        else
                        {
                            _msgSentCount.Count = 0;
                        }
                    }
                }

                Trace.WriteLine("KeepAliveMethod exiting");
            }
            catch (Exception exp)
            {
                Trace.WriteLine("KeepAliveMethod exception: " + exp);

                if (!(exp is ThreadAbortException))
                {
                    HandleCommunicationsError(new Kwp2000Exception("KeepAliveThread Exception.", exp));
                }
            }
        }

        ///	<summary>
        ///	Send a command to the ecu.
        ///	</summary>
        ///	<param name="command">The command bytes	to send.</param>
        ///	<returns>Returns the response.</returns>
        private Kwp2000Response SendKwpCommand(byte[] command)
        {
            Kwp2000Response response = null;

            if (_protocolConnected == false)
            {
                throw new EcuNotConnectedException("Connection to Ecu is not open.");
            }

            if (_interfaceDevice == null)
            {
                throw new Kwp2000Exception("The Interface device has not be set.");
            }

            try
            {
                lock (_lockObject)
                {
                    _msgSentCount.Count++;
                    response = _interfaceDevice.SendCommand(_ecuId, _testerId, command) as Kwp2000Response;
                }
            }
            catch (DeviceTimeoutException exp)
            {
                HandleCommunicationsError(exp);
            }
            catch (Exception e)
            {
                _protocolConnected = false;
                throw new Kwp2000Exception("Error sending command to Ecu.\n Command: " + command, e);
            }

#if	DEBUG
            PrintRequestResponseArray(command, response.Bytes);
#endif
            return (response);
        }

        private void HandleCommunicationsError(Exception e)
        {
            if (_protocolConnected)
            {
                _protocolConnected = false;

                OnProtocolDisconnected(new ProtocolDisconnectedEventArgs(e));
            }
        }

        private void StopKeepAlive()
        {
            int timeout = 0;
            bool stopped = false;

            _keepAliveActive = false;

            if (_keepAliveThread != null)
            {
                while (!stopped)
                {
                    if ((_keepAliveThread.ThreadState & ThreadState.Stopped) == ThreadState.Stopped)
                    {
                        stopped = true;
                    }
                    else
                    {
                        if (timeout >= 500)
                        {
                            if (!_keepAliveThread.Join(1500))
                            {
                                _keepAliveThread.Abort();
                            }

                            stopped = true;
                        }
                    }

                    Thread.Sleep(timeout += 50);
                }
            }

            _keepAliveThread = null;
        }

        #endregion

        #region	enums

        #region BaudRate enum

        ///	<summary>
        ///	Enumeration	of supported ECU extened Baudrates.
        ///	</summary>
        public enum BaudRate
        {
            /// <summary>
            /// The default baud rate.
            /// </summary>
            BaudDefault = 0,
            ///	<summary>
            ///	9600 baud rate
            ///	</summary>
            Baud9600 = 1,
            ///	<summary>
            ///	19200 baud rate
            ///	</summary>
            Baud19200,
            ///	<summary>
            ///	38400 baud rate
            ///	</summary>
            Baud38400,
            ///	<summary>
            ///	57600 baud rate
            ///	</summary>
            Baud57600,
            ///	<summary>
            ///	115200 baud	rate
            ///	</summary>
            Baud115200
        } ;

        #endregion

        #region CommsParamIdentifier enum

        ///	<summary>
        ///	Enumeration	of supported 
        ///	</summary>
        public enum CommsParamIdentifier
        {
            ///	<summary>
            ///	Read the limits	of the timings.
            ///	</summary>
            ReadPossibleLimits = 0,
            ///	<summary>
            ///	Reset timings to defaults.
            ///	</summary>
            ResetToDefault,
            ///	<summary>
            ///	Read current value.
            ///	</summary>
            ReadCurrent,
            ///	<summary>
            ///	Set	the	value.
            ///	</summary>
            SetValues
        } ;

        #endregion

        #region InputOutputControlParameter enum

        ///	<summary>
        ///	Enumeration	of supported inputOutputControlParameters.
        ///	</summary>
        public enum InputOutputControlParameter
        {
            ///	<summary>
            ///	This parameter indicates that the ECU has to regain	complete control of	the	input
            ///	signal,	internal parameter or output signal	identified by InputOutputLocalIdentifier .
            ///	</summary>
            ReturnControlToEcu = 0,
            ///	<summary>
            ///	This parameter indicates that the ECU has to report	the	state of the input signal,
            ///	internal parameter or output signal	identified by InputOutputLocalIdentifier.
            ///	</summary>
            ReportCurrentState = 1,
            ///	<summary>
            ///	This parameter indicates that the ECU is requested to reset	the	input signal, internal
            ///	parameter or output	signal identified by InputOutputLocalIdentifier	to the default value.
            ///	</summary>
            ResetToDefault = 4,
            ///	<summary>
            ///	This value indicates that the ECU is requested to adjust the input signal,
            ///	internal parameter or output signal	referenced by the InputOutputLocalIdentifier in	RAM	to the value specified in
            ///	ControlOptionParameter (e.g. set idle actuator to a	specified step number, set a PWM signal	to a specified value).
            ///	</summary>
            ShortTermAdjustment = 7,
            ///	<summary>
            ///	This value indicates that the ECU is requested to adjust the input signal,
            ///	internal parameter or output signal	referenced by the InputOutputLocalIdentifier
            ///	in EEPROM/FLASH	EEPROM to the value	specified in ControlOptionParameter	
            ///	(e.g. set engine idle speed, set CO).
            ///	</summary>
            LongTermAdjustment = 8
        } ;

        #endregion

        #region InputOutputControlState enum

        ///	<summary>
        ///	The	enumeration	of supported inputOutputControlStates.
        ///	The	parameter inputOutputControlState indicates	the	state the component	has	to be brought to.
        ///	</summary>
        public enum InputOutputControlState : byte
        {
            ///	<summary>
            ///	Component is set to	ON state (ON/OFF components)
            ///	</summary>
            ComponentOn = 0xFF,
            ///	<summary>
            ///	Component is set to	OFF	state (ON/OFF components)
            ///	</summary>
            ComponentOff = 0x00
        } ;

        #endregion

        ///	<summary>
        ///	Enumeration	of supported serviceIdentifiers.
        ///	</summary>
        private enum KwpCommands : byte
        {
            StartCommunication = 0x81,
            StopCommunication = 0x82,
            AccessTimingParameter = 0x83,
            StartDiagnosticSession = 0x10,
            StopDiagnosticSession = 0x20,
            SecurityAccess = 0x27,
            TesterPresent = 0x3E,
            ReadEcuIdentification = 0x1A,
            ReadDataByLocalIdentifier = 0x21,
            WriteDataByLocalIdentifier = 0x3B,
            ReadDiagnosticTroubleCodesByStatus = 0x18,
            ReadStatusOfDiagnosticTroubleCodes = 0x17,
            ReadFreezeFrameData = 0x12,
            ClearDiagnosticInformation = 0x14,
            InputOutputControlByLocalIdentifier = 0x30,
            StartRoutinebyLocalIdentifier = 0x31,
            StopRoutineByLocalIdentifier = 0x32,
            RequestRoutineResultByLocalIdentifier = 0x33,
            RequestDownload = 0x34,
            TransferData = 0x36,
            RequestTransferExit = 0x37
        } ;

        #endregion
    }
}