#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Globalization;
using System.Threading;
using DnEcuDiag.Devices.Enums;
using DnEcuDiag.Devices.Exceptions;
using DnEcuDiag.Devices.Interfaces;
using DnEcuDiag.Devices.Responses;
using DnEcuDiag.Ecus.EventArgTypes;
using DnEcuDiag.Ecus.Exceptions;
using DnEcuDiag.Ecus.Generic;

namespace DnEcuDiag.Protocols
{
    ///	<summary>
    ///	The	Iso9141	protocol class.
    ///	</summary>
    public class Kwp81 : Protocol, IDisposable
    {
        #region	Fields

        private readonly byte _ecuId;
        private readonly IDevice _interfaceDevice;
        private readonly Object _lockObject = new Object();
        private readonly SyncCounter _msgSentCount = new SyncCounter(0);
        private readonly byte _testerId;
        private readonly ManualResetEvent _threadStarted = new ManualResetEvent(false);
        private Thread _keepAliveThread;
        private Boolean _keepAliveStop;
        private Boolean _protocolConnected;
        private ManualResetEvent _signalKeepAlive = new ManualResetEvent(false);

        #endregion

        #region	Dispose	Pattern

        private bool _disposed;

        public IDevice InterfaceDevice
        {
            get { return _interfaceDevice; }
        }


        ///	<summary>
        ///	Dispose	the	underlying resources.
        ///	</summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ///	<summary>
        ///	Finalizer.
        ///	</summary>
        ~Kwp81()
        {
            Dispose(false);
        }

        ///	<summary>
        ///	Close the device.
        ///	</summary>
        public void Close()
        {
            Dispose(true);
        }

        ///	<summary>
        ///	Dispose	of the objects resources.
        ///	</summary>
        ///	<param name="disposing">True if	explicity disposing	object else	GC finalizing object.</param>
        private void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    StopKeepAlive();

                    if (_threadStarted != null)
                    {
                        _threadStarted.Close();
                    }

                    if (_signalKeepAlive != null)
                    {
                        _signalKeepAlive.Close();
                        _signalKeepAlive = null;
                    }

                    _protocolConnected = false;
                }
            }

            _disposed = true;
        }

        #endregion Dispose Pattern

        #region	Constructors

        protected Kwp81()
        {
        }

        ///	<summary>
        ///	Iso9141	protocol class constructor.	
        ///	</summary>
        ///	<param name="interfaceDevice">The underlying physical device.</param>
        ///	<param name="ecuId">The	ID of the Ecu to connect too.</param>
        ///	<param name="testerId">The ID of the tester.</param>
        public Kwp81(IDevice interfaceDevice, byte ecuId, byte testerId)
        {
            _interfaceDevice = interfaceDevice;
            _ecuId = ecuId;
            _testerId = testerId;
        }

        #endregion

        #region	Methods

        ///	<summary>
        ///	Start the keep alove thread.
        ///	</summary>
        public void StartKeepAlive()
        {
            _signalKeepAlive.Set();
        }

        ///	<summary>
        ///	Start communication	with the Ecu.
        ///	</summary>
        ///	<returns>The key bytes.</returns>
        public byte[] StartCommunication()
        {
            if (_interfaceDevice == null)
            {
                throw new Kwp81Exception("The Interface device has not be set.");
            }

            // Start the keep alive	thread.
            try
            {
                _keepAliveThread = new Thread(KeepAliveMethod) {Name = "KeepAliveThread", IsBackground = true};
                _keepAliveThread.Start();

                // Wait	for	the	thread to start.
                _threadStarted.WaitOne();
            }
            catch (Exception e)
            {
                throw new Kwp81Exception("Unable to start the keep alive thread.", e);
            }

            try
            {
                _interfaceDevice.SendCommandWithInit(_ecuId, 0, new[] {(byte) IsoCommand.Acknowledge},
                                                    DeviceInitType.Standard);
            }
            catch (Exception e)
            {
                throw new Kwp81Exception("Error sending command to Ecu.", e);
            }

            _protocolConnected = true;
            _signalKeepAlive.Set();

            return _interfaceDevice.KeyBytes;
        }

        ///	<summary>
        ///	Read the ecu Identification	Data.
        ///	</summary>
        ///	<returns>An	aray of	bytes that contains	the	identification data.</returns>
        public virtual ReadOnlyCollection<Byte[]> ReadEcuIdentification()
        {
            KwpResponse response = SendIsoCommand(new[] {(byte) IsoCommand.IdentificationCodeReadRequest});

            CheckResponse(response.Responses[0], IsoCommand.IdentificationCodeReadRequestAck,
                          "Failed to read identification data.");

            return response.Responses;
        }

        ///	<summary>
        ///	Read a sequence	of ram cells from the device.
        ///	</summary>
        ///	<param name="startAddress">The start address of	the	ram	cell to	read</param>
        ///	<param name="numberCells">The number of	ram	cells to read.</param>
        ///	<returns>The ram cell data.</returns>
        public ReadOnlyCollection<Byte[]> ReadRamCell(UInt16 startAddress, Byte numberCells)
        {
            if (numberCells > 30)
            {
                throw new ArgumentOutOfRangeException("numberCells", "Cannot read more than 29 cells");
            }

            KwpResponse response = SendIsoCommand(new[]
                                                      {
                                                          (Byte) IsoCommand.RamCellsReadRequest,
                                                          numberCells,
                                                          (Byte) ((startAddress & 0xFF00) >> 8),
                                                          (Byte) (startAddress & 0x00FF)
                                                      });

            CheckResponse(response.Responses[0], IsoCommand.RamCellsReadRequestAck, "Failed to read ram data.");

            return response.Responses;
        }

        ///	<summary>
        ///	Read a sequence	of EEProm cells	from the device.
        ///	</summary>
        ///	<param name="startAddress">The start address of	the	EEProm cell	to read</param>
        ///	<param name="numberCells">The number of	EEProm cells to	read.</param>
        ///	<returns>The EEProm	cell data.</returns>
        public Byte[] ReadEePromCell(byte startAddress, byte numberCells)
        {
            if (numberCells > 30)
            {
                throw new ArgumentOutOfRangeException("numberCells", "Cannot read more than 29 cells");
            }

            KwpResponse response =
                SendIsoCommand(new[]
                                   {
                                       (byte) IsoCommand.EEepromReadRequest, numberCells, startAddress,
                                       (byte) (startAddress + numberCells)
                                   });

            CheckResponse(response.Responses[0], IsoCommand.EEepromReadRequestAck, "Failed to read eeprom data.");

            return response.Responses[0];
        }

        ///	<summary>
        ///	Reads the ADC channel.
        ///	</summary>
        ///	<param name="channel">The channel.</param>
        ///	<returns></returns>
        public ReadOnlyCollection<Byte[]> ReadAdcChannel(UInt16 channel)
        {
            KwpResponse response = SendIsoCommand(new[]
                                                      {
                                                          (Byte) IsoCommand.AdcChannelReadRequest, (byte) channel
                                                      });

            CheckResponse(response.Responses[0], IsoCommand.AdcChannelReadRequestAck,
                          "Failed to read ADC channel data.");

            return response.Responses;
        }

        ///	<summary>
        ///	Read the error memory data.
        ///	</summary>
        ///	<returns>Returns the error memory data.</returns>
        public ReadOnlyCollection<Byte[]> ReadErrorMemory()
        {
            KwpResponse response = SendIsoCommand(new[] {(byte) IsoCommand.ErrorMemoryReadRequest});

            if(response.Responses.Count != 0)
            {
                CheckResponse(response.Responses[0], IsoCommand.ErrorMemoryReadRequestAck, "Failed to read error data.");
            }

            Trace.WriteLine(response.ToString());

            return response.Responses;
        }

        ///	<summary>
        ///	End	the	diagnostics	session.
        ///	</summary>
        public void EndDiagnostics()
        {
            // Stop	the	keep alive thread.
            var sw = new Stopwatch();
            sw.Start();
            StopKeepAlive();
            sw.Stop();

            Trace.WriteLine(String.Format(CultureInfo.InvariantCulture, "KeepAlive stop time: {0}", sw.ElapsedMilliseconds));

            SendIsoCommand(new[] {(byte) IsoCommand.DiagnosisEnd});
            _protocolConnected = false;
        }

        ///	<summary>
        ///	Clear the error	memory.
        ///	</summary>
        public void ClearErrorMemory()
        {
            KwpResponse response = SendIsoCommand(new[] {(byte) IsoCommand.MemoryErrorDeleteRequest});

            if (response.Responses.Count != 0)
            {
                CheckResponse(response.Responses[0], IsoCommand.EcuNoAcknowledge, "Failed to clear error data.");
            }
        }

        ///	<summary>
        ///	Activate the Actuator.
        ///	</summary>
        ///	<param name="componentId">The component	to activate.</param>
        public void ActivateActuator(byte componentId)
        {
            KwpResponse response = SendIsoCommand(new[] {(byte) IsoCommand.ActuatorActivation, componentId});

            if (response.Responses.Count != 0)
            {
                CheckResponse(response.Responses[0], IsoCommand.ActuatorActivationAck, "Failed to Activate Actuator.");
            }
        }

        ///	<summary>
        ///	Requests the snap shot.
        ///	</summary>
        ///	<param name="snapshotRequestId">The	snapshot request id.</param>
        ///	<returns></returns>
        public virtual ReadOnlyCollection<Byte[]> RequestSnapshot(UInt16 snapshotRequestId)
        {
            KwpResponse response = SendIsoCommand(new[] {(Byte) IsoCommand.SnapshotRequest, (Byte) snapshotRequestId});

            CheckResponse(response.Responses[0], IsoCommand.SnapshotRequestAck, "Failed to Request Snap Shot.");

            return response.Responses;
        }

        public void StartDownloadRoutine(Byte[] parameters)
        {
            var request = new List<Byte> { (Byte)IsoCommand.StartDownloadRoutine};
            request.AddRange(parameters);

            var response = SendIsoCommand(request.ToArray());

            CheckResponse(response.Responses[0], IsoCommand.StartDownloadRoutineAck, "Failed to execute routine.");
        }

        internal void WriteRamCell(UInt16 startAddress, List<Byte> actuatorRoutine)
        {
            var request = new List<Byte> {(Byte) IsoCommand.RamCellsWriteRequest, (Byte) ((startAddress & 0xFF00) >> 8), (Byte) (startAddress & 0x00FF)};
            request.AddRange(actuatorRoutine.ToArray());

            var response = SendIsoCommand(request.ToArray());

            CheckResponse(response.Responses[0], IsoCommand.Acknowledge, "Failed to write bytes to Ram.");
        }

        #endregion Methods

        #region Virtual Methods

        protected virtual void CheckResponse(byte[] response, IsoCommand responseResult, String message)
        {
            if (response[1] != (byte) responseResult)
            {
                throw new Kwp81Exception(message, response[1]);
            }
        }

        #endregion

        #region	Private	Methods

        ///	<summary>
        ///	The	keep alive thread method this sends	a message to the ecu.
        ///	</summary>
        private void KeepAliveMethod()
        {
            try
            {
                // Signal to the other thread that we have started.
                _threadStarted.Set();

                while (!_keepAliveStop)
                {
                    int count = 0;

                    _signalKeepAlive.WaitOne();

                    while (count != 1)
                    {
                        Thread.Sleep(500);

                        // if any messages have	been sent then reset the counter.
                        if (_msgSentCount.Count > 0)
                        {
                            _msgSentCount.Count = 0;
                            count = 0;
                        }
                        else
                        {
                            count++;
                        }
                    }

                    if (!_keepAliveStop)
                    {
                        lock (_lockObject)
                        {
                            if (_protocolConnected != true)
                            {
                                continue;
                            }

                            if (_msgSentCount.Count == 0)
                            {
                                _interfaceDevice.SendCommand(_ecuId, _testerId, new[] {(byte) IsoCommand.Acknowledge});
                            }
                            else
                            {
                                _msgSentCount.Count = 0;
                            }
                        }
                    }
                }
            }
            catch (Exception exp)
            {
                Trace.WriteLine("KeepAliveMethod exception: " + exp);

                if (!(exp is ThreadAbortException))
                {
                    HandleCommunicationsError(new Kwp81Exception("KeepAliveThread Exception.", exp));
                }
            }
        }

        ///	<summary>
        ///	Send a command to the ecu.
        ///	</summary>
        ///	<param name="command">The command bytes	to send.</param>
        ///	<returns>Returns the response.</returns>
        internal KwpResponse SendIsoCommand(byte[] command)
        {
            KwpResponse response = null;

            if (_protocolConnected == false)
            {
                throw new EcuNotConnectedException("Connection to Ecu is not open.");
            }

            if (_interfaceDevice == null)
            {
                throw new Kwp81Exception("The Interface device has not been set.");
            }

            try
            {
                lock (_lockObject)
                {
                    _msgSentCount.Count++;
                    response = _interfaceDevice.SendCommand(_ecuId, _testerId, command) as KwpResponse;
                }
            }
            catch (DeviceTimeoutException exp)
            {
                HandleCommunicationsError(exp);
            }
            catch (Exception e)
            {
                _protocolConnected = false;
                throw new Kwp81Exception(String.Format(WinFormStrings.ErrorSendingCommandToEcu,_ecuId.ToString("x", CultureInfo.InvariantCulture), command), e);
            }

            return response;
        }

        ///	<summary>
        ///	Handles	the	communications error.
        ///	</summary>
        ///	<param name="e">The	exception to throw.</param>
        private void HandleCommunicationsError(Exception e)
        {
            _signalKeepAlive.Reset();

            if (_protocolConnected)
            {
                _protocolConnected = false;

                OnProtocolDisconnected(new ProtocolDisconnectedEventArgs(e));
            }
        }

        private void StopKeepAlive()
        {
            _keepAliveStop = true;

            if (_keepAliveThread != null)
            {
                if (!_keepAliveThread.Join(1000))
                {
                    try
                    {
                        _keepAliveThread.Abort();
                    }
                    catch (ThreadAbortException)
                    {
                        Trace.WriteLine("Keepalive Thread aborting");
                    }
                }
            }
        }

        #endregion Private Methods	
    }
}