#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.Generic;
using System.IO;
using DnEcuDiag.Configuration;
using DnEcuDiag.Ecus.Generic;

namespace DnEcuDiag.VehiclePerformance
{
    ///	<summary>
    ///	This class calculates the performance measurments from a 
    ///	set	of speed measurements and rpm measurements.
    ///	It calculates 0-60 etc
    ///	It also	calculates power (Kw) and torque (Nm).
    ///	</summary>
    ///	<remarks>
    ///	The	loss power is :	Pl = d * v^3 + f * v
    ///	with :
    ///	 - d = aerodynamic loss	factor
    ///	 - f = friction	loss factor
    ///	 d and f are the solutions of the equation : 0 = Pl	+ M	* v	* a
    ///	 where a is	the	(negative) acceleration	obtained when the output power
    ///	 of	the	engine is nul (i.e.	clutch pushed in).
    ///	 If	we call	y(i) = - M * a(i), we have :
    ///	  d	= average( (y(i+1) - y(i)) / (v(i+1)^2 - (vi)^2) )
    ///	  f	= average( y(i)	- d	* v(i)^2 )
    ///	</remarks>
    internal class VehiclePerformanceMeasure
    {
        #region	Fields

        // Horse power and Torque
        private readonly List<TimeStampedMeasurement> _dynoLossSpeedMeasurements;
        private readonly List<TimeStampedMeasurement> _rpmMeasurements;
        private readonly List<TimeStampedMeasurement> _speedMeasurements;

        // Imperial	measurements

        // Metric measurements.

        #endregion

        #region	Constructor

        ///	<summary>
        ///	Default	constructor
        ///	</summary>
        public VehiclePerformanceMeasure()
        {
            _speedMeasurements = new List<TimeStampedMeasurement>();
            _rpmMeasurements = new List<TimeStampedMeasurement>();
            _dynoLossSpeedMeasurements = new List<TimeStampedMeasurement>();
            DynoMeasurements = new DynoMeasurementValue[0];
            PerformanceRun = new VehiclePerformanceRun();
        }

        #endregion

        #region	Public Methods

        ///	<summary>
        ///	Convert	speed to m/s * 1000.
        ///	</summary>
        ///	<param name="speed">The	speed in KPH</param>
        ///	<returns>The speed in meters per second.</returns>
        public static int ConvertSpeedToMetersSecond(Int32 speed)
        {
            return (speed*10000/36);
        }

        ///	<summary>
        ///	Reset the Vechicle performance measure data.
        ///	</summary>
        public void Reset()
        {
            PerformanceRun.Clear();
            _dynoLossSpeedMeasurements.Clear();
            _dynoLossSpeedMeasurements.Clear();
            _speedMeasurements.Clear();
            _rpmMeasurements.Clear();
            DynoMeasurements = new DynoMeasurementValue[0];
            
            Complete = false;


        }

        ///	<summary>
        ///	Perform	the	power and torque calculations.
        ///	</summary>
        ///	<param name="correctForLoss">Include loss correction.</param>
        ///	<param name="filterCount">The numer	of filter iterations to	perform	on the dyno	results.</param>
        public void DynoCalculateResults(bool correctForLoss, int filterCount)
        {
            if (_rpmMeasurements.Count == 0) return;
            if (DynoMass == 0) throw new ApplicationException(WinFormStrings.TheDynoMassHasNotBeenEntered);
            if (DynoGear == 0) throw new ApplicationException(WinFormStrings.TheDynoGearHasNotBeenSet);

            int j = 0;
            DynoMeasurements = new DynoMeasurementValue[_rpmMeasurements.Count - 1];

#if	DEBUG
            var sw = new StreamWriter("c:\\performance_measures.csv", true);

            sw.WriteLine("Perfoemance run: " + DateTime.Now);

            for (int i = 0; i < _rpmMeasurements.Count - 1; i++)
            {
                var measure = _rpmMeasurements[i];
                sw.WriteLine(measure.Measurement + "," + measure.Timestamp.Ticks);
            }
            sw.Close();
#endif
            for (int i = 0; i < _rpmMeasurements.Count - 1; i++)
            {
                DynoMeasurements[j] = DynoCalculateResult(_rpmMeasurements[i], _rpmMeasurements[i + 1],correctForLoss);
                j++;
            }

            // Filter results.
            for (int i = 0; i < filterCount; i++)
                FilterDynoResults();

            // Fix a scaling problem.
            for (int i = 0; i < DynoMeasurements.Length; i++)
                DynoMeasurements[i].Power = DynoMeasurements[i].Power/1000;
        }

        ///	<summary>
        ///	Calculate the dyno loss	factors.
        ///	</summary>
        public void DynoLossCalculate()
        {
            if (DynoMass == 0) throw new ApplicationException("The dyno mass has not been set.");

#if	DEBUG
            var sw = new StreamWriter("c:\\speed_measures.csv", true);

            sw.WriteLine("Perfoemance run: " + DateTime.Now);

            for (int i = 0; i < _dynoLossSpeedMeasurements.Count - 1; i++)
            {
                var measure = _dynoLossSpeedMeasurements[i];
                sw.WriteLine(measure.Measurement + "," + measure.Timestamp.Ticks);
            }
            sw.Close();
#endif

            double sum = 0;
            TimeStampedMeasurement measure0;
            TimeStampedMeasurement measure1;

            for (int i = 0; i < _dynoLossSpeedMeasurements.Count; i++)
            {
                var measurement = _dynoLossSpeedMeasurements[i];
                measurement.Measurement = ConvertSpeedToMetersSecond((int) measurement.Measurement);
            }
            // Calculate aerodynamic loss factor.
            for (int i = 0; i < _dynoLossSpeedMeasurements.Count - 1; i++)
            {
                measure0 = _dynoLossSpeedMeasurements[i];
                measure1 = _dynoLossSpeedMeasurements[i + 1];
                Double d = (DynoLossY(i + 1) - DynoLossY(i))/
                           (Math.Pow((measure1.Measurement/1000.0), 2) - Math.Pow((measure0.Measurement/1000.0), 2));
                sum += d;
            }

            AerodynamicLossFactor = sum/_dynoLossSpeedMeasurements.Count;

            /* calculate f */
            sum = 0;
            for (int i = 0; i < _dynoLossSpeedMeasurements.Count; i++)
            {
                measure0 = _dynoLossSpeedMeasurements[i];

                double f = DynoLossY(i) - (AerodynamicLossFactor*Math.Pow((measure0.Measurement/1000.0), 2));
                sum += f;
            }
            FrictionLossFactor = sum/_dynoLossSpeedMeasurements.Count;
        }

        ///	<summary>
        ///	Add	a new speed	TimeStampedMeasurement.
        ///	</summary>
        ///	<param name="measurement">The measurement to add.</param>
        public void AddLossMeasurement(TimeStampedMeasurement measurement)
        {
            _dynoLossSpeedMeasurements.Add(measurement);
        }

        ///	<summary>
        ///	Add	a new speed	TimeStampedMeasurement.
        ///	</summary>
        ///	<param name="measurement">The measurement to add.</param>
        public void AddSpeedMeasurement(TimeStampedMeasurement measurement)
        {
            _speedMeasurements.Add(measurement);

            CalculatePeformanceMeasurements();
        }

        ///	<summary>
        ///	Add	a new rpm TimeStampedMeasurement.
        ///	</summary>
        ///	<param name="measurement">The measurement to add.</param>
        public void AddRpmMeasurement(TimeStampedMeasurement measurement)
        {
            _rpmMeasurements.Add(measurement);
        }

        #endregion Public Methods

        #region	Private	Methods

        #region	Gear

        private const int DynoGearAccuracy = 1000;

        ///	<summary>
        ///	Set	ratio from speed (m/s *	1000) and rpm
        ///	</summary>
        ///	<param name="speed"></param>
        ///	<param name="rpm"></param>
        public void DynoSetGear(int speed, int rpm)
        {
            DynoGear = speed*DynoGearAccuracy/rpm;
        }

        ///	<summary>
        ///	Get	speed (in m/s *	1000) from rpm
        ///	</summary>
        ///	<param name="rpm"></param>
        ///	<returns></returns>
        private int DynoGetSpeedFromRpm(int rpm)
        {
            int speed = rpm*DynoGear/DynoGearAccuracy;
            return speed;
        }

        #endregion Gear

        ///	<summary>
        ///	get	y(i)
        ///	</summary>
        ///	<param name="i"></param>
        ///	<returns></returns>
        private double DynoLossY(int i)
        {
            return (0 - DynoMass*DynoLossA(i));
        }

        ///	<summary>
        ///	Calculate loss power from speed	(m/s2 *	1000)
        ///	</summary>
        ///	<param name="speed"></param>
        ///	<returns></returns>
        private long DynoLossPower(long speed)
        {
            /* Pl =	d *	v^3	+ f	* v	*/
            double power = AerodynamicLossFactor*Math.Pow((speed/1000.0), 3) + FrictionLossFactor*(speed/1000.0);

            return (long) (power + .50);
        }

        ///	<summary>
        ///	Calculate the power	and	Torque curves between two points.
        ///	</summary>
        ///	<param name="measure0">First measurement.</param>
        ///	<param name="measure1">Second measurement.</param>
        ///	<param name="correctForLoss">If	loss corrections is	to be included in power	calculations.</param>
        ///	<returns>A dyno	measurment at the specified	rpm.</returns>
        private DynoMeasurementValue DynoCalculateResult(TimeStampedMeasurement measure0,
                                                           TimeStampedMeasurement measure1, bool correctForLoss)
        {
            var result = new DynoMeasurementValue();

            TimeSpan timeDiff = measure1.Timestamp.Subtract(measure0.Timestamp);

            /* calculate effective power from cinetic energy variation */
            double dyno = Math.Pow(DynoGetSpeedFromRpm((int) measure1.Measurement), 2) -
                            Math.Pow(DynoGetSpeedFromRpm((int) measure0.Measurement), 2);
            dyno = (dyno/timeDiff.TotalMilliseconds); /* m2/s3 */
            dyno = (dyno*DynoMass/2000);

            result.Rpm = (int) ((measure0.Measurement + measure1.Measurement)/2);

            if (correctForLoss)
            {
                /* calculate loss power	at given speed */
                Double loss = DynoLossPower(DynoGetSpeedFromRpm(result.Rpm));

                /* rpm,	power and torque */
                result.Power = dyno + loss;
            }
            else
            {
                result.Power = dyno;
            }

            result.Torque = (Torque(result.Power, result.Rpm) + .50);

            return result;
        }

        ///	<summary>
        ///	Calculate Torque.
        ///	</summary>
        ///	<param name="power"></param>
        ///	<param name="rpm"></param>
        ///	<returns></returns>
        private static double Torque(double power, int rpm)
        {
            return ((power*60)/(rpm*2*Math.PI));
        }

        ///	<summary>
        ///	a(i) (m/s^2)
        ///	</summary>
        ///	<param name="i"></param>
        ///	<returns></returns>
        private double DynoLossA(int i)
        {
            double a;

            if (i <= 0)
                a = DynoLossAInter(0, 1);
            else if (i >= _dynoLossSpeedMeasurements.Count - 1)
                a = DynoLossAInter(_dynoLossSpeedMeasurements.Count - 2, _dynoLossSpeedMeasurements.Count - 1);
            else
                a = (DynoLossAInter(i - 1, i) + DynoLossAInter(i, i + 1))/2;

            return a;
        }

        ///	<summary>
        ///	get	acceleration between 2 measures	(m/s^2)
        ///	</summary>
        ///	<param name="i"></param>
        ///	<param name="j"></param>
        ///	<returns></returns>
        private double DynoLossAInter(int i, int j)
        {
            /* avoid bad use */
            if ((i < 0) || (i >= j) || (j > _dynoLossSpeedMeasurements.Count))
            {
                return 0;
            }

            /* select measures */
            var measure0 = _dynoLossSpeedMeasurements[i];
            var measure1 = _dynoLossSpeedMeasurements[j];

            TimeSpan timeDiff = measure1.Timestamp.Subtract(measure0.Timestamp);

            /* get acceleration	*/
            double a = 1.0*(measure1.Measurement - measure0.Measurement)/timeDiff.TotalMilliseconds;

            return a;
        }

        ///	<summary>
        ///	This calculates	the	performance	measurements
        ///	from the speed data	in m_arrSpeedMeasurements.
        ///	</summary>
        private void CalculatePeformanceMeasurements()
        {
            TimeStampedMeasurement snapshotOne;
            TimeStampedMeasurement snapshotTwo;
            DateTime timeTwo;
            TimeSpan timeDiff;

            if (_speedMeasurements.Count < 2)
            {
                return;
            }

            double totalSecond = 0;
            double totaldistanceCovered = 0;
            for (int i = 1; i < _speedMeasurements.Count; i++)
            {
                snapshotOne = _speedMeasurements[(i - 1)];
                snapshotTwo = _speedMeasurements[(i)];
                // Calculate meters/second
                double metersPerSecond = ((snapshotTwo.Measurement + snapshotOne.Measurement)/2)*0.277777777777778;

                // Get the difference in seconds between the two points.
                DateTime timeOne = snapshotOne.Timestamp;
                timeTwo = snapshotTwo.Timestamp;
                timeDiff = timeTwo.Subtract(timeOne);
                double secondsDiff = timeDiff.TotalSeconds;
                // Record total	time difference.
                totalSecond += secondsDiff;
                double distanceCovered = secondsDiff*metersPerSecond;
                totaldistanceCovered += distanceCovered;

                // sixty foot measurement
                if (totaldistanceCovered >= 18.288 && PerformanceRun.SixtyFootTime.TotalSeconds == 0)
                {
                    PerformanceRun.SixtyFootTime =  TimeSpan.FromSeconds(totalSecond - ((totaldistanceCovered - 18.288) / metersPerSecond));
                }
                // 20 meter	measurement
                if (totaldistanceCovered >= 20 && PerformanceRun.TwentyMeterTime.TotalSeconds == 0)
                {
                    PerformanceRun.TwentyMeterTime = TimeSpan.FromSeconds(totalSecond - ((totaldistanceCovered - 20) / metersPerSecond));
                }


                // Sixty Mph Time
                if (snapshotTwo.Measurement >= 96.5606 && PerformanceRun.SixtyMphTime.TotalSeconds == 0)
                {
                    PerformanceRun.SixtyMphTime = TimeSpan.FromSeconds(totalSecond -
                                                                       (snapshotTwo.Measurement - 96.5606)/
                                                                       ((snapshotTwo.Measurement - snapshotOne.Measurement)/secondsDiff));
                }
                // 100 Kph Time
                if (snapshotTwo.Measurement >= 100 && PerformanceRun.OneHundredKphTime.TotalSeconds == 0)
                {
                    PerformanceRun.OneHundredKphTime = TimeSpan.FromSeconds(totalSecond -
                                                                            (snapshotTwo.Measurement - 100)/
                                                                            ((snapshotTwo.Measurement - snapshotOne.Measurement)/secondsDiff));
                }

                // 330 foot	measurement
                if (totaldistanceCovered >= 100.584 && PerformanceRun.ThreeThirtyTime.TotalSeconds == 0)
                {
                    PerformanceRun.ThreeThirtyTime = TimeSpan.FromSeconds(totalSecond - ((totaldistanceCovered - 100.584) / metersPerSecond));
                }
                // 100 meter measurement
                if (totaldistanceCovered >= 100 && PerformanceRun.OneHundredMeterTime.TotalSeconds == 0)
                {
                    PerformanceRun.OneHundredMeterTime = TimeSpan.FromSeconds(totalSecond - ((totaldistanceCovered - 100) / metersPerSecond));
                }

                // 200 meter time.
                double startDistance;
                if (totaldistanceCovered >= 200 && PerformanceRun.TwoHundredMeterTime.TotalSeconds == 0)
                {
                    startDistance = totaldistanceCovered - 200;
                    PerformanceRun.TwoHundredMeterTime = TimeSpan.FromSeconds(totalSecond - (startDistance / metersPerSecond));
                    PerformanceRun.TwoHundredMeterSpeed = (snapshotTwo.Measurement - snapshotOne.Measurement) *
                                                          (distanceCovered - startDistance)/distanceCovered +
                                                          snapshotOne.Measurement;
                }
                // Eight of	a mile time.
                if (totaldistanceCovered >= 201.168 && PerformanceRun.EighthMileTime.TotalSeconds == 0)
                {
                    startDistance = totaldistanceCovered - 201.168;
                    PerformanceRun.EighthMileTime = TimeSpan.FromSeconds(totalSecond - (startDistance / metersPerSecond));
                    PerformanceRun.EighthMileSpeed = (snapshotTwo.Measurement - snapshotOne.Measurement) *
                                                     (distanceCovered - startDistance)/distanceCovered +
                                                     snapshotOne.Measurement*0.621371;
                }
                // Three hundered meter	time.
                if (totaldistanceCovered >= 300 && PerformanceRun.ThreeHundredMeterTime.TotalSeconds == 0)
                {
                    PerformanceRun.ThreeHundredMeterTime = TimeSpan.FromSeconds(totalSecond - ((totaldistanceCovered - 300) / metersPerSecond));
                }
                // ThousandFootTime
                if (totaldistanceCovered >= 304.8 && PerformanceRun.ThousandFootTime.TotalSeconds == 0)
                {
                    PerformanceRun.ThousandFootTime = TimeSpan.FromSeconds(totalSecond - ((totaldistanceCovered - 304.8) / metersPerSecond));
                }

                // 400 meter time.
                if (totaldistanceCovered >= 400 && PerformanceRun.FourHundredMeterTime.TotalSeconds == 0)
                {
                    startDistance = totaldistanceCovered - 400;
                    PerformanceRun.FourHundredMeterTime = TimeSpan.FromSeconds(totalSecond - (startDistance / metersPerSecond));
                    PerformanceRun.FourHundredMeterSpeed = (snapshotTwo.Measurement - snapshotOne.Measurement) *
                                                           (distanceCovered - startDistance)/distanceCovered +
                                                           snapshotOne.Measurement;
                }
                // Quarter mile	time.
                if (totaldistanceCovered >= 402.336 && PerformanceRun.QuarterMileTime.TotalSeconds == 0)
                {
                    startDistance = totaldistanceCovered - 402.336;
                    PerformanceRun.QuarterMileTime = TimeSpan.FromSeconds(totalSecond - (startDistance / metersPerSecond));
                    PerformanceRun.QuarterMileSpeed = (snapshotTwo.Measurement - snapshotOne.Measurement) *
                                                      (distanceCovered - startDistance)/distanceCovered +
                                                      snapshotOne.Measurement*0.621371;
                    Complete = true;
                }
            }
        }

        ///	<summary>
        ///	Filter the calculated results to smooth	out	irregularities in the readings.
        ///	</summary>
        private void FilterDynoResults()
        {
            var filteredResults = new DynoMeasurementValue[DynoMeasurements.Length];

            DynoMeasurements.CopyTo(filteredResults, 0);

            for (int i = 1; i < filteredResults.Length - 1; i++)
            {
                filteredResults[i].Power = (DynoMeasurements[i - 1].Power + DynoMeasurements[i].Power +
                                            DynoMeasurements[i + 1].Power)/3;
                filteredResults[i].Torque = Torque(filteredResults[i].Power, filteredResults[i].Rpm); //	+ .50;
            }

            DynoMeasurements = filteredResults;
        }

        #endregion Private Methods

        #region	Properties

        ///	<summary>
        ///	The	dyno measuments	taken.
        ///	</summary>
        public DynoMeasurementValue[] DynoMeasurements { get; private set; }

        ///	<summary>
        ///	The	Gear that the dyno run is performed.
        ///	</summary>
        public int DynoGear { get; set; }

        ///	<summary>
        ///	The	frictional loss	factor.
        ///	This can be	calculated from	the	coast down data	or set manually.
        ///	</summary>
        public double FrictionLossFactor { get; set; }

        ///	<summary>
        ///	The	Aerodynamic	loss factor.
        ///	This can be	calculated from	the	coast down data	or set manually.
        ///	</summary>
        public double AerodynamicLossFactor { get; set; }

        ///	<summary>
        ///	The	weight of the vehicle expressed	in kilograms.
        ///	</summary>
        public double DynoMass { get; set; }

        ///	<summary>
        ///	If the performance run is complete.
        ///	</summary>
        public bool Complete { get; private set; }

        public VehiclePerformanceRun PerformanceRun { get; set; }

        #endregion Properties
    }
}