#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;

namespace InstrumentationControls
{
    public partial class DigitalClock : UserControl
    {
        private static Graphics _graphics;
        private readonly ArrayList _alarmTimes = new ArrayList();
        private readonly Object _lockObject = new Object();
        private char _amPm;
        private DigitalDisplay _amPmDisplay; // panel for AM/PM display
        private bool _blink; // toggle for blinking effect of colons

        private ClockType _clockType = ClockType.DigitalClock;

        private DigitalDisplay[] _colonDisplay; // panels for displaying colons
        // ColonTimer to blink the colons between the digits
        private Timer _colonTimer;
        // type of clock to display (a normal clock, stop watch or count down)
        // count down in milli seconds, default of 10 seconds
        private int _countDownMilliSeconds = 10000;
        // whenever count down starts this time is set to Now + countDownMilliSeconds
        private DateTime _countDownTo;
        private DigitalColor _digitColor;
        private DigitalDisplay[] _digitDisplay; // panels on which digits are displayed

        // currently displayed numbers on the clock, useful to freeze
        private int _hour;
        private DigitalDisplay _microSecDisplay; // panel for displaying 1/10 of a second
        private int _min;
        private int _ms;
        private int _sec;
        private Timer _showTimer;
        private DateTime _stopwatchBegin = DateTime.Now;

        public DigitalClock()
        {
            ClockDisplayFormat = ClockFormat.TwelveHourFormat;
            // This call is required by the Windows.Forms Form Designer.
            InitializeComponent();
        }

        /// <summary>
        /// Get or Set count down time in milli seconds
        /// </summary>
        public int CountdownTime
        {
            get { return _countDownMilliSeconds; }
            set
            {
                if (value < 1000)
                {
                    MessageBox.Show("Count down time cannot be less than 1000", "Error");
                }
                else
                {
                    _countDownMilliSeconds = value;
                }
            }
        }

        // set the alarm time
        public DateTime AlarmTime
        {
            set
            {
                if (value < DateTime.Now)
                {
                    MessageBox.Show("Alarm time cannot be earlier.", "Error");
                }
                else
                {
                    _alarmTimes.Add(value);
                }
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public ClockFormat ClockDisplayFormat { get; set; }

        // setting clock type
        // DigitalClock and StopWatch will automatically start the clock
        // For CountDown the number of seconds should be set before calling this property
        public ClockType SetClockType
        {
            get { return _clockType; }
            set
            {
                _clockType = value;
                switch (_clockType)
                {
                    case ClockType.Stopwatch:
                        _stopwatchBegin = DateTime.Now; // start stopwatch clock
                        break;
                    case ClockType.Countdown:
                        _countDownTo = DateTime.Now.AddMilliseconds(_countDownMilliSeconds);
                        break;
                }
            }
        }

        /// <summary>
        /// Get/Set the color in which the digits are displayed
        /// </summary>
        public DigitalColor SetDigitalColor
        {
            get { return _digitColor; }
            set
            {
                Invalidate();
                _digitColor = value;
                DigitalDisplay.SetPenColor(_digitColor);
            }
        }

        #region Component Designer generated code

        /// <summary>
        /// Required method for Designer support - do not modify 
        /// the contents of this method with the code editor.
        /// </summary>
        private void InitializeComponent()
        {
            this.components = new System.ComponentModel.Container();
            this._showTimer = new System.Windows.Forms.Timer(this.components);
            this._colonTimer = new System.Windows.Forms.Timer(this.components);
            // 
            // ShowTimer
            // 
            this._showTimer.Tick += new System.EventHandler(this.OnClockTimer);
            // 
            // ColonTimer
            // 
            this._colonTimer.Tick += new System.EventHandler(this.OnColonTimer);
            // 
            // DigitalClockCtrl
            // 
            this.BackColor = System.Drawing.Color.Black;
            this.Name = "DigitalClockCtrl";
            this.Size = new System.Drawing.Size(354, 70);
            this.Load += new System.EventHandler(this.OnLoad);
            this.Paint += new System.Windows.Forms.PaintEventHandler(this.OnPaint);
        }

        #endregion

        public event EventHandler CountdownDoneEvent;
        public event EventHandler RaiseAlarmEvent;

        public void Reset()
        {
            _hour = 0;
            _min = 0;
            _sec = 0;
            _ms = 0;
        }

        // OnPaint - called when regions of clock are invalidated
        private void OnPaint(object sender, PaintEventArgs e)
        {
            lock (_lockObject)
            {
                DisplayTime(e.Graphics);
            }
        }

        // This function is aware which panel should display what number
        // and also the colons and PM and AM displays

        private void DisplayTime(Graphics g)
        {
            DateTime dt = DateTime.Now;
            if (_clockType != ClockType.Freeze)
            {
                _hour = dt.Hour;
                _min = dt.Minute;
                _sec = dt.Second;
                _ms = dt.Millisecond;
                _amPm = ' ';
            }
            TimeSpan ts = TimeSpan.Zero;

            // check if alarms are set, raise them
            for (int i = 0; i < _alarmTimes.Count; i++)
            {
                if (dt > (DateTime) _alarmTimes[i] && RaiseAlarmEvent != null)
                {
                    _alarmTimes.RemoveAt(i);
                    RaiseAlarmEvent(this, new EventArgs());
                }
            }
            switch (_clockType)
            {
                case ClockType.DigitalClock:
                    if (ClockDisplayFormat == ClockFormat.TwelveHourFormat)
                    {
                        _hour = dt.Hour%12;
                    }

                    if (_hour == 0)
                    {
                        _hour = 12;
                    }

                    switch (ClockDisplayFormat)
                    {
                        case ClockFormat.TwentyFourHourFormat:
                            break;
                        case ClockFormat.TwelveHourFormat:
                            _amPm = (dt.Hour/12 > 0) ? 'P' : 'A';
                            break;
                    }
                    break;
                case ClockType.Countdown:
                    ts = _countDownTo.Subtract(dt);
                    if (ts < TimeSpan.Zero)
                    {
                        _clockType = ClockType.DigitalClock;
                        ts = TimeSpan.Zero;

                        if (CountdownDoneEvent != null)
                        {
                            CountdownDoneEvent(this, new EventArgs());
                        }
                    }
                    break;
                case ClockType.Stopwatch:
                    ts = dt.Subtract(_stopwatchBegin);
                    break;
                case ClockType.Freeze:
                    break;
            }
            if (_clockType != ClockType.DigitalClock &&
                _clockType != ClockType.Freeze) // ts used for stopwatch or countdown
            {
                _hour = ts.Hours;
                _min = ts.Minutes;
                _sec = ts.Seconds;
                _ms = ts.Milliseconds;
            }
            _digitDisplay[0].Draw(_hour/10, g);
            _digitDisplay[1].Draw(_hour%10, g);
            _digitDisplay[2].Draw(_min/10, g);
            _digitDisplay[3].Draw(_min%10, g);
            _digitDisplay[4].Draw(_sec/10, g);
            _digitDisplay[5].Draw(_sec%10, g);
            _microSecDisplay.Draw(_ms/100, g);
            if (_amPm == ' ')
            {
                _amPmDisplay.Draw(g);
            }
            else
            {
                _amPmDisplay.Draw(_amPm, g);
            }
        }

        // Timer used to refresh clock display
        private void OnClockTimer(object sender, EventArgs e)
        {
            DisplayTime(_graphics);
        }

        // Keeping the colon timer, gives a special effect of colon blinking
        // independent of the seconds or 1/10 seconds display
        private void OnColonTimer(object sender, EventArgs e)
        {
            //display the 2 colons between the hours-minutes and minutes-seconds
            _colonDisplay[0].DrawColon(_graphics, ColonType.Rectangular, _blink);
            _colonDisplay[1].DrawColon(_graphics, ColonType.Rectangular, _blink);
            if (_clockType == ClockType.Freeze)
            {
                _blink = false;
            }
            else
            {
                _blink = !_blink;
            }
        }

        private void OnLoad(object sender, EventArgs e)
        {
            _graphics = Graphics.FromHwnd(Handle);
            PreparePanels();
            _showTimer.Interval = 100;
            _showTimer.Start(); // digits are refreshed on timer count
            _colonTimer.Interval = 1000;
            _colonTimer.Start(); // this will blink the colon

            // adding the resize handler here so that it will be called
            // only after graphics variable is created
            Resize += OnResize;
        }

        // function to prepare the digital clock panels by dividing the rectangle
        // It is assumed that the height of each digit is double that of it's width
        // Spacing betweent the digits is 10% of the width
        // The colon characters occupy 50% of width of the digits
        private void PreparePanels()
        {
            // from the above assumptions for height and width
            // the height should be 2.4 units and width 8.8 units :-)
            // check height and width whichever is dominant and adjust the other
            // and set up margins
            var rect = new Rectangle(0, 0, Width, Height);


            // Calculate a digit width (which is our unit) from the above metrics
            // and settle for the least value
            var widthUnit = (int) (rect.Width/8.8F);
            var heightUnit = (int) (rect.Height/2.4F);
            int digitWidth = (widthUnit < heightUnit) ? widthUnit : heightUnit;

            int digitHeight = 2*digitWidth;
            int colonWidth = digitWidth/2;
            int spacing = digitWidth/10;
            if (spacing < 1) spacing = 1; // atleast a spacing of 1 is required
            // widths, spacings and margins
            // height of colon display is same as a digit
            float hMargin = (rect.Width - (8.8F*digitWidth))/2;
            // left and right margin top and bottom margin
            float vMargin = (rect.Height - digitHeight)/2;

            // This is the basic rectangle, offset it as required
            var basicRect = new Rectangle(0, 0, digitWidth, digitHeight);
            int xOffset;
            Rectangle calcRect; // calculated rectangle for a panel
            // Y offset is same for all elements, expcept 1/10 second and AM/PM display
            var yOffset = (int) (vMargin);

            // create digit panels.  6 digits
            if (_digitDisplay == null)
            {
                _digitDisplay = new DigitalDisplay[6];
            }
            for (int i = 0; i < 6; i++)
            {
                calcRect = basicRect;
                xOffset = (int) (hMargin + (spacing*(i + 2 + (i/2))) + (i*digitWidth) + ((i/2)*colonWidth));
                calcRect.Offset(xOffset, yOffset);
                if (_digitDisplay[i] == null)
                {
                    _digitDisplay[i] = new DigitalDisplay(calcRect);
                }
                else
                {
                    _digitDisplay[i].CalculateAllParameters(calcRect);
                }
            }

            if (_colonDisplay == null)
            {
                _colonDisplay = new DigitalDisplay[2];
            }
            // for first colon
            calcRect = basicRect;
            calcRect.Width = colonWidth;
            xOffset = (int) (hMargin + 3*spacing + 2*digitWidth);
            calcRect.Offset(xOffset, yOffset);
            if (_colonDisplay[0] == null)
            {
                _colonDisplay[0] = new DigitalDisplay(calcRect);
            }
            else
            {
                _colonDisplay[0].CalculateAllParameters(calcRect);
            }

            // for second colon
            calcRect = basicRect;
            calcRect.Width = colonWidth;
            xOffset = (int) ((6*spacing) + (4*digitWidth) + colonWidth + hMargin);
            calcRect.Offset(xOffset, yOffset);
            if (_colonDisplay[1] == null)
            {
                _colonDisplay[1] = new DigitalDisplay(calcRect);
            }
            else
            {
                _colonDisplay[1].CalculateAllParameters(calcRect);
            }

            // for displaying 'A'(AM) or 'P' (PM)
            calcRect = basicRect;
            calcRect.Width = colonWidth;
            calcRect.Height = calcRect.Height/2;
            xOffset = (int) ((10*spacing) + (6*digitWidth) + (2*colonWidth) + hMargin);
            calcRect.Offset(xOffset, (int) (yOffset + digitHeight/2.0 + 1));
            if (_amPmDisplay == null)
            {
                _amPmDisplay = new DigitalDisplay(calcRect);
            }
            else
            {
                _amPmDisplay.CalculateAllParameters(calcRect);
            }

            // for displaying 1/10 of a second
            // reuse AM/PM display panel rectangle
            // only change Y coordinate, shift upwards
            calcRect.Y = yOffset - 1; // just to keep it apart from AM/PM display
            if (_microSecDisplay == null)
            {
                _microSecDisplay = new DigitalDisplay(calcRect);
            }
            else
            {
                _microSecDisplay.CalculateAllParameters(calcRect);
            }
        }

        // On resize of control recalculate the rectangles for display
        // Also recreate the graphics so that the clipping region is updated
        private void OnResize(object sender, EventArgs e)
        {
            lock (_lockObject)
            {
                PreparePanels();
                _graphics.Dispose();
                _graphics = Graphics.FromHwnd(Handle);
                _graphics.Clear(BackColor);
            }
        }
    }
}