#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Globalization;

namespace InstrumentationControls
{
    // A digit class which can draw a number on a graphics surface
    internal sealed class DigitalDisplay
    {
        // pens used to draw the digits
        private static readonly Pen Pen = new Pen(Color.FromArgb(255, 0, 0));
        private static Pen _dimPen;

        // for each digit the display bits are set into an int
        // 'A' and 'P' are included for AM and PM display in the clock
        private readonly int[] _displayNum = new[]
                                                 {
                                                     63, 12, 118, 94, 77, 91, 123, 14, 127, 95,
                                                     111, // to display 'A'
                                                     103
                                                 }; // to display 'P'

        private readonly Point[] _points; // end point coordinates of lines in the digital display
        private float _linewidth = 20.0F;

        // Rectangles in which colons are displayed
        private Rectangle _colonRect1;
        private Rectangle _colonRect2;

        internal DigitalDisplay(Rectangle rect)
        {
            Pen.StartCap = LineCap.Triangle;
            Pen.EndCap = LineCap.Triangle;

            _points = new Point[14]; // there are 7 lines in a display
            for (int i = 0; i < 14; i++)
            {
                _points[i] = new Point(0, 0);
            }
            CalculateAllParameters(rect);
        }

        // This function is called by the paint method to display the numbers
        // A set of bits in the 'displayNum' variable define which of the
        // display legs to display
        // Based on this the ones with a '1' are in bright color and the rest
        // with '0's are in a dull color giving the effect of a digital clock
        internal void Draw(int num, // number to display
                           Graphics g) // graphics object for drawing
        {
            // although pens are global linewidths are specific to each instance
            Pen.Width = _dimPen.Width = _linewidth;

            for (int i = 0; i < 7; i++)
            {
                var check = (int) Math.Pow(2, i); // used to check if a leg of digit should be bright or dull
                if ((check & _displayNum[num]) == 0)
                {
                    g.DrawLine(_dimPen, _points[i*2], _points[i*2 + 1]);
                }
                else
                {
                    g.DrawLine(Pen, _points[i*2], _points[i*2 + 1]);
                }
            }
        }

        private static void SetDimPen()
        {
            _dimPen = (Pen) Pen.Clone();

            _dimPen.Width = Pen.Width;
            _dimPen.Color = Color.FromArgb(
                Pen.Color.R > 0 ? 60 : 0,
                Pen.Color.G > 0 ? 60 : 0,
                Pen.Color.B > 0 ? 60 : 0);
        }

        internal static void SetPenColor(DigitalColor dclr)
        {
            Pen.Color = Color.FromArgb(
                (dclr == DigitalColor.RedColor) ? 255 : 0,
                (dclr == DigitalColor.GreenColor) ? 255 : 0,
                (dclr == DigitalColor.BlueColor) ? 255 : 0);
            SetDimPen();
        }

        // function that draws a colon in the middle of the rectangular panel
        // possible modes are circular or rectangular points in the colon
        internal void DrawColon(Graphics g, ColonType type, bool dim)
        {
            Pen.Width = _dimPen.Width = _linewidth;
            Pen p = (dim) ? _dimPen : Pen; // choose a pen for blinking
            switch (type)
            {
                case ColonType.Circular:
                    g.DrawEllipse(p, _colonRect1);
                    g.DrawEllipse(p, _colonRect2);
                    break;
                case ColonType.Rectangular:
                    g.DrawRectangle(p, _colonRect1);
                    g.DrawRectangle(p, _colonRect2);
                    break;
            }
        }

        // Draws the complete rectangle in dim shade to give the digital effect :-)
        internal void Draw(Graphics g)
        {
            // althought pens are static, linewidths are specific to each instance
            _dimPen.Width = _linewidth;
            for (int i = 0; i < 7; i++)
                g.DrawLine(_dimPen, _points[i*2], _points[i*2 + 1]);
        }

        // Overloaded function to display characters 'A' and 'P' for AM and PM
        // Using the same algorithm used to display numbers above
        internal void Draw(char ch, // character to display
                           Graphics g) // graphics object for drawing
        {
            // 10 and 11 are indices of A and P in the displayNum array
            switch (Char.ToUpper(ch, CultureInfo.InvariantCulture))
            {
                case 'A':
                    Draw(10, g);
                    break;
                case 'P':
                    Draw(11, g);
                    break;
            }
        }

        // Constructor takes a rectangle and prepares the end points
        // of the lines to be drawn for the clock

        internal void CalculateAllParameters(Rectangle rect)
        {
            _linewidth = (rect.Width/5);
            if (_linewidth < 2) _linewidth = 2;
            if (_linewidth > 20) _linewidth = 20;
            Pen.Width = _linewidth;
            SetDimPen();

            CalculateLineEnds(rect);
            CalculateColonRectangles(rect);
        }

        // Function calculates end points of lines to display
        // The draw function will draw lines using this data
        private void CalculateLineEnds(Rectangle rect)
        {
            // 0,1,2,9,10,11,12 points share the same left edge X coordinate
            _points[0].X = _points[1].X = _points[2].X = _points[9].X =
                                                         _points[10].X = _points[11].X = _points[12].X = rect.Left;

            // points 3,4,5,6,7,8,13 the right edge X coordinate
            _points[3].X = _points[4].X = _points[5].X = _points[6].X =
                                                         _points[7].X = _points[8].X = _points[13].X = rect.Right - (int) _linewidth;

            // Points 1,2,3,4 are the top most points
            _points[1].Y = _points[2].Y = _points[3].Y = _points[4].Y = (rect.Top);

            // Points 0,11,12,13,5,6 are the middle points
            _points[0].Y = _points[11].Y = _points[12].Y = _points[13].Y =
                                                           _points[5].Y = _points[6].Y =
                                                                          rect.Top + (int) ((rect.Height - _linewidth)/2.0);
            // points 7,8,9,10 are on the bottom edge
            _points[7].Y = _points[8].Y = _points[9].Y = _points[10].Y
                                                         = rect.Top + (int) (rect.Height - _linewidth);

            // now adjust the coordinates that were computed, to get the digital look
            AdjustCoordinates();
        }

        // This function is necessary to give the lines a digital clock look
        // Push the coordinates a little away so that they look apart
        private void AdjustCoordinates()
        {
            for (int i = 0; i < 7; i++)
            {
                // Always draw from left to right and top to bottom
                // Adjust the end points accordingly
                if (_points[i*2].X > _points[(i*2) + 1].X || _points[i*2].Y > _points[(i*2) + 1].Y)
                {
                    Point swap = _points[i*2]; // required in case points have to be swapped
                    _points[i*2] = _points[(i*2) + 1];
                    _points[(i*2) + 1] = swap;
                }

                // for horizontal lines adjust the X coord
                if (_points[i*2].X != _points[(i*2) + 1].X)
                {
                    _points[i*2].X += (int) (_linewidth/1.6);
                    _points[(i*2) + 1].X -= (int) (_linewidth/1.6);
                }
                // for vertical lines adjust the y coord
                if (_points[i*2].Y != _points[(i*2) + 1].Y)
                {
                    _points[i*2].Y += (int) (_linewidth/1.6);
                    _points[(i*2) + 1].Y -= (int) (_linewidth/1.6);
                }
            }
        }

        // function to calculate the rectangles required to drawn colon dot inside
        private void CalculateColonRectangles(Rectangle rect)
        {
            _colonRect1 = _colonRect2 = rect;
            _colonRect1.X = _colonRect2.X = rect.X + (int) ((rect.Width - _linewidth)/2.0);
            _colonRect1.Y = rect.Y + rect.Height/3;
            _colonRect2.Y = rect.Y + (rect.Height*2)/3;
            _colonRect1.Width = _colonRect1.Height =
                               _colonRect2.Width = _colonRect2.Height = (int) _linewidth;
        }
    }
}