#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Drawing;
using System.Globalization;
using System.Windows.Forms;
using InstrumentationControls.SafeHandles;

namespace InstrumentationControls
{
    public class SingleChannelScope : ScopeBase
    {
        #region Fields

        private StringFormat _horzFormat;
        private StringFormat _vertFormat;
        private StringFormat _xAxisFormat;

        private readonly ScopeChannel _channel;

        private Color _crLegendColor;
        private Pen _plotPen;

        private double _currentPosition; // current position
        // since plotting is based on a LineTo for each new point
        // we need a starting point (i.e. a "previous" point)
        // use 0.0 as the default first point.
        // these are public member variables, and can be changed outside
        // (after construction).  Therefore _perviousPosition could be set to
        // a more appropriate value prior to the first call to SetPosition.
        private double _previousPosition; // previous position

        private SolidBrush _legendBrush;

        public SingleChannelScope()
        {
            _channel = new ScopeChannel(10, -10, Color.White) { ParentScope = this };
            
            _plotPen = new Pen(Color.White, 1);

            _horzFormat = new StringFormat { Alignment = StringAlignment.Far };

            _vertFormat = new StringFormat
            {
                FormatFlags = StringFormatFlags.DirectionVertical,
                Alignment = StringAlignment.Center
            };

            _xAxisFormat = new StringFormat { Alignment = StringAlignment.Center };

            _legendBrush = new SolidBrush(_crLegendColor);
        }

        #endregion

        public SingleChannelScope(String text)
            : base(text)
        {
        }

        public SingleChannelScope(String text, int left, int top, int width, int height)
            : base(text, left, top, width, height)
        {
        }

        public SingleChannelScope(Control parent, String text)
            : base(parent, text)
        {
        }

        public SingleChannelScope(Control parent, String text, int left, int top, int width, int height)
            : base(parent, text, left, top, width, height)
        {
        }

        #region Dispose

        private Boolean _disposed;

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        protected override void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    if (_horzFormat != null)
                    {
                        _horzFormat.Dispose();
                        _horzFormat = null;
                    }

                    if (_plotPen != null)
                    {
                        _plotPen.Dispose();
                        _plotPen = null;
                    }

                    if (_legendBrush != null)
                    {
                        _legendBrush.Dispose();
                        _legendBrush = null;
                    }

                    if (_vertFormat != null)
                    {
                        _vertFormat.Dispose();
                        _vertFormat = null;
                    }

                    if (_xAxisFormat != null)
                    {
                        _xAxisFormat.Dispose();
                        _xAxisFormat = null;
                    }

                    if (BitmapGrid != null)
                    {
                        BitmapGrid.Dispose();
                        BitmapGrid = null;
                    }

                    if (BitmapPlot != null)
                    {
                        BitmapPlot.Dispose();
                        BitmapPlot = null;
                    }
                }

                _disposed = true;
            }

            base.Dispose(disposing);
        }

        #endregion

        protected override void OnCreateControl()
        {
            ExecuteExceptionHandledDelegate(HandleControlCreated);
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            ExecuteExceptionHandledDelegate(() => HandleOnSizeChanged(e));
        }

        private void HandleOnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            InvalidateControl();
        }

        /// <summary>
        /// 
        /// </summary>
        private void DrawPoint()
        {
            // this does the work of "scrolling" the plot to the left
            // and appending a new data point all of the plotting is 
            // directed to the memory based bitmap associated with m_dcPlot
            // the will subsequently be BitBlt'd to the client in OnPaint

            if (BitmapPlot != null)
            {
                //// Bit Blit the plot onto itself
                using (Graphics graphicsPlot = Graphics.FromImage(BitmapPlot))
                {
                    using (var hBitMap = new GdiObjectSafeHandle(BitmapPlot.GetHbitmap()))
                    {
                        using (var hdcBitmap = new GdiPlusDeviceContextSafeHandle(graphicsPlot))
                        {
                            using (var hdcMem = NativeMethods.CreateCompatibleDC(hdcBitmap))
                            {
                                NativeMethods.SelectObject(hdcMem, hBitMap);

                                NativeMethods.BitBlt(hdcBitmap,
                                                     GridRectangle.Left,
                                                     GridRectangle.Top + 1,
                                                     GridRectangle.Width,
                                                     GridRectangle.Height,
                                                     hdcMem,
                                                     GridRectangle.Left +
                                                     ShiftPixels,
                                                     GridRectangle.Top + 1,
                                                     NativeMethods.RasterOp.
                                                         Srccopy);
                            }
                        }
                    }

                    var verticalFactor = GridRectangle.Height / _channel.Range;

                    int prevX = GridRectangle.Right - PlotShiftPixels;
                    int prevY = GridRectangle.Bottom - (int)((_previousPosition - _channel.LowerLimit) * verticalFactor);

                    int currX = GridRectangle.Right - HalfShiftPixels;
                    int currY = GridRectangle.Bottom - (int)((_currentPosition - _channel.LowerLimit) * verticalFactor);

                    if (_plotPen.Color != ForeColor)
                    {
                        _plotPen.Dispose();
                        _plotPen = new Pen(ForeColor);
                    }

                    graphicsPlot.DrawLine(_plotPen, prevX, prevY, currX, currY);

                    if ((prevY <= GridRectangle.Top) || (currY <= GridRectangle.Top))
                    {
                        graphicsPlot.FillRectangle(BackgroundBrush,
                                                   new Rectangle(prevX, ClientRectangle.Top, currX + 1, GridRectangle.Top + 1));
                    }

                    if ((prevY >= GridRectangle.Bottom) || (currY >= GridRectangle.Bottom))
                    {
                        graphicsPlot.FillRectangle(BackgroundBrush,
                                                   new Rectangle(prevX, GridRectangle.Bottom + 1, currX + 1, ClientRectangle.Bottom + 1));
                    }
                }


                // store the current point for connection to the next point
                _previousPosition = _currentPosition;
            }
        }

        internal override void AppendPoint(Double newPoint, ScopeChannel channelToAppend)
        {
            ExecuteExceptionHandledDelegate(delegate
            {
                _currentPosition = newPoint;
                DrawPoint();
                Invalidate();
            });
        }

        /// <summary>
        /// Redraw the grid bitmap.
        /// </summary>
        internal override void InvalidateControl()
        {
            ExecuteExceptionHandledDelegate(delegate
            {
                if ((ClientRectangle.Width != 0) & (ClientRectangle.Height != 0))
                {
                    var maxUnits = _channel.UpperLimit.ToString(CultureInfo.InvariantCulture).Length >
                                   _channel.LowerLimit.ToString(CultureInfo.InvariantCulture).Length
                                       ? _channel.UpperLimit.ToString(CultureInfo.InvariantCulture)
                                       : _channel.LowerLimit.ToString(CultureInfo.InvariantCulture);

                    using (var controlGraphics = CreateGraphics())
                    {
                        var labelWidth = (int)controlGraphics.MeasureString(maxUnits, Font).Width;
                        GridRectangle = new Rectangle(20 + labelWidth, 10, ClientRectangle.Width - 30 - labelWidth,
                                                      ClientRectangle.Height - 30);

                        BitmapGrid = CreateBitmapIfRequired(BitmapGrid, ClientSize, controlGraphics);

                        using (var gridGraphics = Graphics.FromImage(BitmapGrid))
                        {
                            DrawPlotGrid(gridGraphics);

                            DrawPlotLabels(gridGraphics);

                            BitmapPlot = CreateBitmapIfRequired(BitmapPlot, ClientSize, controlGraphics);

                            using (var graphicsPlot = Graphics.FromImage(BitmapPlot))
                            {
                                graphicsPlot.FillRectangle(BackgroundBrush, 0, 0, BitmapPlot.Width, BitmapPlot.Height);

                                BitmapDisplay = CreateBitmapIfRequired(BitmapDisplay, ClientSize, controlGraphics);
                            }
                        }
                    }
                }

                Invalidate();
            });
        }

        protected override void DrawPlotLabels(Graphics graphics)
        {
            var rect1 = new Rectangle(0, GridRectangle.Y, GridRectangle.Left - 4, base.Font.Height);
            var rect2 = new Rectangle(0, GridRectangle.Bottom - base.Font.Height, GridRectangle.Left - 4,
                                      base.Font.Height);

            graphics.DrawString(String.Format(CultureInfo.InvariantCulture, "{0}", _channel.UpperLimit), base.Font,
                                _legendBrush,
                                rect1, _horzFormat);

            graphics.DrawString(String.Format(CultureInfo.InvariantCulture, "{0}", _channel.LowerLimit), base.Font,
                                _legendBrush,
                                rect2, _horzFormat);

            graphics.DrawString(_channel.UnitsX, base.Font, _legendBrush,
                                new Point((GridRectangle.Left + GridRectangle.Right) / 2,
                                          GridRectangle.Bottom + 4),
                                _xAxisFormat);
            graphics.DrawString(_channel.UnitsY, base.Font, _legendBrush,
                                new Point((ClientRectangle.Left + GridRectangle.Left) / 2,
                                          (GridRectangle.Bottom + GridRectangle.Top) / 2), _vertFormat);
        }

        #region Public Methods

        public override Color BackColor
        {
            get { return base.BackColor; }
            set
            {
                base.BackColor = value;
                BackgroundBrush = new SolidBrush(value);
                InvalidateControl();
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public Color GridColor
        {
            get { return GridPen.Color; }
            set
            {
                GridPen = new Pen(value, 1);

                InvalidateGrid();
            }
        }

        /// <summary>
        /// The legend color
        /// </summary>
        public Color LegendColor
        {
            get { return _crLegendColor; }
            set
            {
                _crLegendColor = value;
                _legendBrush = new SolidBrush(_crLegendColor);
                InvalidateControl();
            }
        }

        #endregion

        #region Public Properties

        public ScopeChannel Channel
        {
            get { return _channel; }
        }

        #endregion
    }
}