#region License
// Copyright  2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Threading;
using SerialPort.Api;
using SerialPort.Api.Enums;
using SerialPort.Exceptions;
using SerialPort.Settings;

namespace SerialPort
{
    public class Ftd2XxPort : IPort, IDisposable
    {
        private readonly Ftd2XxSettings _portSettings;
        private readonly Queue<Byte> _rxFifo;
        private readonly Object _rxQueueLock;
        private readonly ManualResetEvent _threadStarted = new ManualResetEvent(false);
        private Boolean _brk;
        private Boolean _disposed;
        private Boolean _dtr;
        private Thread _eventThread;
        private IntPtr _hPort = IntPtr.Zero;
        private Int32 _inputLength;
        private Boolean _rts;
        private Boolean _stopComms;

        #region Constructor

        public Ftd2XxPort(String portName, Ftd2XxSettings portSettings)
        {
            _portSettings = portSettings;
            _rxFifo = new Queue<Byte>();
            _rxQueueLock = new Object();
        }

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        #region IPort Members

        public event EventHandler DataReceived;

        public Boolean Close()
        {
            Dispose(true);

            return true;
        }

        public Byte[] Output
        {
            set
            {
                UInt32 bytesWritten = 0;
                HandleApiError("Write",
                               FTD2xxNativeMethods.FT_Write(_hPort, value, (UInt32) value.Length, ref bytesWritten));
            }
        }

        public void SendImmediate(Byte toSend)
        {
            Output = new[] {toSend};
        }

        public Boolean Purge()
        {
            HandleApiError("Purge",
                           FTD2xxNativeMethods.FT_Purge(_hPort, FtPurgeFlags.FtPurgeRx | FtPurgeFlags.FtPurgeTx));

            return true;
        }

        public void SetBaudRate(Int32 baudRate)
        {
            _portSettings.BaudRate = baudRate;
            HandleApiError("Port FT_SetBaudRate",
                           FTD2xxNativeMethods.FT_SetBaudRate(_hPort, (uint) _portSettings.BaudRate));
        }

        public Boolean Open()
        {
            if (_hPort == IntPtr.Zero)
            {
                try
                {
                    HandleApiError("Port Open", FTD2xxNativeMethods.FT_Open(0, ref _hPort));

                    HandleApiError("Port FT_SetBaudRate",
                                   FTD2xxNativeMethods.FT_SetBaudRate(_hPort, (uint) _portSettings.BaudRate));

                    HandleApiError("Port FT_SetDataCharacteristics",
                                   FTD2xxNativeMethods.FT_SetDataCharacteristics(_hPort, _portSettings.DataSize,
                                                                                 (byte) _portSettings.StopBits,
                                                                                 (byte) _portSettings.Parity));
                    HandleApiError("Port FT_SetFlowControl",
                                   FTD2xxNativeMethods.FT_SetFlowControl(_hPort, (UInt16) _portSettings.FlowControl,
                                                                         _portSettings.Xon, _portSettings.XoffChar));

                    HandleApiError("Port FT_SetTimeouts",
                                   FTD2xxNativeMethods.FT_SetTimeouts(_hPort, 0, 0));

                    if (_portSettings.BusSettings.HasValue)
                    {
                        ConfigureUsbParameters();
                    }



                    _stopComms = false;
                    _eventThread = new Thread(CommEventThread) {Priority = ThreadPriority.Highest};
                    _eventThread.Start();

                    // wait for the thread to actually get spun up
                    _threadStarted.WaitOne();

                    return true;
                }
                catch (Exception)
                {
                    Close();
                    throw;
                }
            }

            return true;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Send or check for a communications BREAK event
        /// </summary>
        public Boolean Break
        {
            get
            {
                if (_hPort == IntPtr.Zero)
                {
                    return false;
                }

                return _brk;
            }
            set
            {
                if (_hPort == IntPtr.Zero)
                {
                    return;
                }

                if (value)
                {
                    HandleApiError("Break On", FTD2xxNativeMethods.FT_SetBreakOn(_hPort));
                    _brk = true;
                }
                else
                {
                    HandleApiError("Break Off", FTD2xxNativeMethods.FT_SetBreakOff(_hPort));
                    _brk = false;
                }
            }
        }

        public int InputLen
        {
            get { return _inputLength; }
            set { _inputLength = value; }
        }

        /// <summary>
        /// Gets or sets the current DTR line state (true = 1, false = 0)
        /// </summary>
        public Boolean DtrEnable
        {
            get { return _dtr; }
            set
            {
                if (_hPort == IntPtr.Zero)
                {
                    return;
                }

                if (value)
                {
                    HandleApiError("DTR On", FTD2xxNativeMethods.FT_SetDtr(_hPort));
                    _dtr = true;
                }
                else
                {
                    HandleApiError("DTR Off", FTD2xxNativeMethods.FT_ClrDtr(_hPort));
                    _dtr = false;
                }
            }
        }

        /// <summary>
        /// The actual amount of data in the input buffer
        /// </summary>
        public Int32 InBufferCount
        {
            get
            {
                if (_hPort == IntPtr.Zero)
                {
                    return 0;
                }

                return _rxFifo.Count;
            }
        }

        public byte[] Input
        {
            get
            {
                if (_hPort == IntPtr.Zero)
                {
                    return null;
                }

                Byte[] data;

                lock (_rxQueueLock)
                {
                    Int32 dequeueLength;
                    if (_inputLength == 0)
                    {
                        dequeueLength = _rxFifo.Count;
                    }
                    else
                    {
                        dequeueLength = (_inputLength < _rxFifo.Count) ? _inputLength : _rxFifo.Count;
                    }

                    data = new byte[dequeueLength];

                    for (Int32 p = 0; p < dequeueLength; p++)
                    {
                        data[p] = _rxFifo.Dequeue();
                    }
                }

                return data;
            }
        }

        /// <summary>
        /// Gets or sets the current RTS line state (true = 1, false = 0)
        /// </summary>
        public Boolean RtsEnable
        {
            get { return _rts; }
            set
            {
                if (_hPort == IntPtr.Zero)
                {
                    return;
                }

                if (value)
                {
                    HandleApiError("RTS On", FTD2xxNativeMethods.FT_SetRts(_hPort));
                    _dtr = true;
                }
                else
                {
                    HandleApiError("RTS Off", FTD2xxNativeMethods.FT_ClrRts(_hPort));
                    _dtr = false;
                }

                _rts = value;
            }
        }

        #endregion

        #region Static Methods

        public static ReadOnlyCollection<UsbPortDescriptor> GetUsbPorts()
        {
            var usbPortDescriptors = new List<UsbPortDescriptor>();
            var deviceName = new byte[64];
            var serialNumber = new byte[16];

            unsafe
            {
                UInt32 numberOfDevices = 0u;

                HandleApiError("Get Device Count", FTD2xxNativeMethods.FT_CreateDeviceInfoList(ref numberOfDevices));

                for (uint i = 0; i < numberOfDevices; i++)
                {
                    fixed (byte* deviceNamePtr = deviceName)
                    {
                        fixed (byte* serialNumberPtr = serialNumber)
                        {
                            var handle = new IntPtr();
                            UInt32 deviceFlags = 0;
                            FtDevice deviceType = 0;
                            UInt32 deviceId = 0u;
                            UInt32 locId = 0u;

                            HandleApiError("Get Device Info Detail",
                                           FTD2xxNativeMethods.FT_GetDeviceInfoDetail(i, ref deviceFlags, ref deviceType,
                                                                                      ref deviceId, ref locId,
                                                                                      new IntPtr(serialNumberPtr),
                                                                                      new IntPtr(deviceNamePtr),
                                                                                      ref handle));

                            usbPortDescriptors.Add(new UsbPortDescriptor(ConvertBytesToString(deviceName), (Int32) locId));
                        }
                    }
                }
            }

            return usbPortDescriptors.AsReadOnly();
        }

        #endregion

        #region Private Methods

        private static String ConvertBytesToString(Byte[] stringBytes)
        {
            var enc = new ASCIIEncoding();
            Int32 i = 0;

            foreach (byte b in stringBytes)
            {
                if (b == 0)
                {
                    break;
                }

                i++;
            }

            return enc.GetString(stringBytes, 0, i);
        }

        private void CommEventThread()
        {
            FtEventNotificationFlags eventFlags = FtEventNotificationFlags.FtEventRxchar;

            var rxEvent = new AutoResetEvent(false);

            try
            {
                HandleApiError("SetEventNotification",
                               FTD2xxNativeMethods.FT_SetEventNotification(_hPort,
                                                                           FtEventNotificationFlags.
                                                                               FtEventModemStatus |
                                                                           FtEventNotificationFlags.FtEventRxchar,
                                                                           rxEvent.SafeWaitHandle));
                _threadStarted.Set();

                while (!_stopComms)
                {
                    rxEvent.WaitOne();

                    UInt32 txBytes = 0;
                    UInt32 rxBytes = 0;
                    HandleApiError("GetStatus",
                                   FTD2xxNativeMethods.FT_GetStatus(_hPort, ref rxBytes, ref txBytes, ref eventFlags));

                    if ((eventFlags & FtEventNotificationFlags.FtEventModemStatus) ==
                        FtEventNotificationFlags.FtEventModemStatus)
                    {
                        FtModemStatus status = FtModemStatus.Cts;
                        HandleApiError("GetModemStatus", FTD2xxNativeMethods.FT_GetModemStatus(_hPort, ref status));

                        if ((status & FtModemStatus.Cts) == FtModemStatus.Cts)
                        {
                            // Cts is high
                        }
                        else
                        {
                            // Cts is low
                        }

                        if ((status & FtModemStatus.Dsr) == FtModemStatus.Dsr)
                        {
                            // DSR is high
                        }
                        else
                        {
                            // DSR is low
                        }

                        if ((status & FtModemStatus.Dcd) == FtModemStatus.Dcd)
                        {
                            // DSR is high
                        }
                        else
                        {
                            // DSR is low
                        }

                        if ((status & FtModemStatus.Ri) == FtModemStatus.Ri)
                        {
                            // DSR is high
                        }
                        else
                        {
                            // DSR is low
                        }
                    }

                    if (rxBytes > 0)
                    {
                        HandleRecieve(rxBytes);
                    }
                }
            }
            finally
            {
                rxEvent.Close();
            }
        }

        private void HandleRecieve(uint rxBytes)
        {
            var buffer = new byte[rxBytes];

            uint bytesRead = 0;
            HandleApiError("Read", FTD2xxNativeMethods.FT_Read(_hPort, buffer, rxBytes, ref bytesRead));

            lock (_rxQueueLock)
            {
                foreach (byte b in buffer)
                {
                    _rxFifo.Enqueue(b);
                }
            }

            if (DataReceived != null)
            {
                DataReceived(this, EventArgs.Empty);
            }
        }

        private void ConfigureUsbParameters()
        {
            HandleApiError("Set LatencyTimer",
                           FTD2xxNativeMethods.FT_SetLatencyTimer(_hPort,
                                                                  _portSettings.BusSettings.Value.
                                                                      LatencyTimer));
            HandleApiError("Set USB Parameters",
                           FTD2xxNativeMethods.FT_SetUSBParameters(_hPort,
                                                                   _portSettings.BusSettings.Value.
                                                                       InTransferSize,
                                                                   _portSettings.BusSettings.Value.
                                                                       OutTransferSize));
        }

        private static void HandleApiError(String message, FtStatus status)
        {
            if (status != FtStatus.FtOk)
            {
                throw new UsbPortException(String.Format("Message: {0} Reason: {1}", message, status));
            }
        }

        #endregion

        ~Ftd2XxPort()
        {
            Dispose(false);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    _stopComms = true;

                    if (_eventThread != null)
                    {
                        Boolean joined = _eventThread.Join(1000);

                        if (!joined)
                        {
                            _eventThread.Abort();
                        }
                    }

                    if (_hPort != IntPtr.Zero)
                    {
                        HandleApiError("Close", FTD2xxNativeMethods.FT_Close(_hPort));
                    }

                    if (_threadStarted != null)
                    {
                        _threadStarted.Close();
                    }
                }

                _disposed = true;
            }
        }
    }
}