﻿#region License
// Copyright © 2004 - 2010 All Right Reserved, Derek Goslin
// 
// 1. Definitions
// The terms "reproduce," "reproduction," "derivative works," and "distribution" have the same meaning here as under U.S. copyright law.
// A "contribution" is the original software, or any additions or changes to the software.
// A "contributor" is any person that distributes its contribution under this license.
// "Licensed patents" are a contributor's patent claims that read directly on its contribution.
// 
// 2. Grant of Rights
// (A) Copyright Grant- Subject to the terms of this license, including the license
//     conditions and limitations in section 3, each contributor grants you a non-exclusive,
//     worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative 
//     works of its contribution, and distribute its contribution or any derivative works that you create.
// (B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations
//     in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its
//     licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of
//     its contribution in the software or derivative works of the contribution in the software.
// 
// 3. Conditions and Limitations
// (A) Reciprocal Grants- For any file you distribute that contains code from the software (in source code or binary format), 
//     you must provide recipients the source code to that file along with a copy of this license, which license will govern that file. 
//     You may license other files that are entirely your own work and do not contain code from the software under any terms you choose.
// (B) No Trademark License- This license does not grant you rights to use any contributors' name, logo, or trademarks.
// (C) If you bring a patent claim against any contributor over patents that you claim are infringed by the software,
//     your patent license from such contributor to the software ends automatically.
// (D) If you distribute any portion of the software, you must retain all copyright, patent, trademark,
//     and attribution notices that are present in the software.
// (E) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy 
//     of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do 
//     so under a license that complies with this license.
// (F) The software is licensed "as-is." You bear the risk of using it. The contributors give no express warranties, guarantees or conditions.
//    You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your
//    local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
#endregion

using System;
using CommStudio.Connections;

namespace SerialPort
{
    public class SaxPort : IPort, IDisposable
    {
        private const int InputQueueSize = 4096;
        private const int OutputQueueSize = 2048;
        private readonly SerialConnection _port;
        private Boolean _disposed;
        private Int32 _inputLenght;
        private SerialOptions _serialOptions;

        public SaxPort(String portName, Int32 baudRate, Parity parity, Int32 dataBits,
                       CommStopBits stopBits)
        {
            _port = new SerialConnection();

            _serialOptions = new SerialOptions(portName, baudRate, parity, dataBits, stopBits, false, false,
                                              false, false, false, false);

            _port.SetOptions(_serialOptions);
            _port.ReadTimeOut = Int32.MaxValue;
            _port.WriteTimeOut = 0;

            _port.DataAvailable += PortDataAvailable;
        }

        #region IDisposable Members

        public void Dispose()
        {
            Dispose(true);
        }

        #endregion

        #region IPort Members

        public event EventHandler DataReceived;

        public bool DtrEnable
        {
            get { return _serialOptions.EnableDtr; }
            set { _port.Dtr = _serialOptions.EnableDtr = value; }
        }

        public bool RtsEnable
        {
            get { return _serialOptions.EnableRts; }
            set { _port.Rts = _serialOptions.EnableRts = value; }
        }

        public bool Break
        {
            get { return _port.Break; }
            set { _port.Break = value; }
        }

        public int InputLen
        {
            get { return _inputLenght; }
            set { _inputLenght = value; }
        }

        public int InBufferCount
        {
            get { return _port.Available; }
        }

        public byte[] Input
        {
            get
            {
                var buffer = new byte[_inputLenght];

                _port.Read(buffer, 0, _inputLenght);

                return buffer;
            }
        }

        public byte[] Output
        {
            set { _port.Write(value, 0, value.Length); }
        }

        public bool Open()
        {
            _port.Open(InputQueueSize, OutputQueueSize);

            return true;
        }

        public bool Close()
        {
            if (_port.IsOpen)
            {
                _port.Close();
            }

            return true;
        }

        public bool Purge()
        {
            _port.Flush(true);
            return true;
        }

        public void SendImmediate(Byte toSend)
        {
            _port.Write(toSend);
        }

        public void SetBaudRate(Int32 baudRate)
        {
            _serialOptions.BaudRate = baudRate;
            _port.SetOptions(_serialOptions);
        }

        #endregion

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    if (_port.IsOpen)
                    {
                        _port.Close();
                    }
                    _port.Dispose();
                }

                _disposed = true;
            }
        }

        private void PortDataAvailable(object sender, EventArgs e)
        {
            if (DataReceived != null)
            {
                DataReceived(null, EventArgs.Empty);
            }
        }
    }
}